/***************************************************************************
 *   Copyright (C) 2005 by Sean Harmer                                     *
 *   sh@rama.homelinux.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
 ***************************************************************************/
#include "kfileaclpropsplugin.h"

#include <qcheckbox.h>
#include <qframe.h>
#include <qgroupbox.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qptrvector.h>
#include <qpushbutton.h>
#include <qtooltip.h>

#include <kcombobox.h>
#include <klocale.h>
#include <kmessagebox.h>

#include <kdebug.h>

#include <errno.h>
#ifdef HAVE_ACL_LIBACL_H
# include <acl/libacl.h>
#endif
#include <pwd.h>
#include <grp.h>

#include "images.h"
#include "kacllistview.h"

static int getACLQualifier( acl_entry_t aclEntry, acl_tag_t aclType,
                     KFileItem* pItem, QString& qualifier )
{
    if ( aclType == ACL_USER )
    {
        uid_t* uid = (uid_t*) acl_get_qualifier( aclEntry );
        if ( uid != 0 )
        {
            passwd* user = getpwuid( *uid );
            qualifier = QString( user->pw_name );
            acl_free( uid );
        }
        else
        {
            kdDebug() << "Could not retrieve uid for ACL entry" << endl;
            qualifier = QString::null;
            return errno;
        }
    }
    else if ( aclType == ACL_GROUP )
    {
        gid_t* gid = (gid_t*) acl_get_qualifier( aclEntry );
        if ( gid != 0 )
        {
            group* group = getgrgid( *gid );
            qualifier = QString( group->gr_name );
            acl_free( gid );
        }
        else
        {
            kdDebug() << "Could not retrieve uid for ACL entry" << endl;
            qualifier = QString::null;
            return errno;
        }
    }
    else if ( aclType == ACL_USER_OBJ )
    {
        qualifier = pItem->user();
    }
    else if ( aclType == ACL_GROUP_OBJ )
    {
        qualifier = pItem->group();
    }
    else
    {
        qualifier = QString::null;
    }

    return 0;
}


static int getACLPerms( acl_entry_t aclEntry, acl_perm_t& perms )
{
    perms = 0;

    acl_permset_t permset;
    int ret = acl_get_permset( aclEntry, &permset );
    if ( ret != 0 )
    {
        kdDebug() << "KACLEntry::KACLEntry(): Could not retrieve the permission set" << endl;
        return errno;
    }

#ifdef HAVE_ACL_LIBACL_H
    if ( acl_get_perm( permset, ACL_READ ) )
        perms |= ACL_READ;
    if ( acl_get_perm( permset, ACL_WRITE ) )
        perms |= ACL_WRITE;
    if ( acl_get_perm( permset, ACL_EXECUTE ) )
        perms |= ACL_EXECUTE;
#else
    /** \todo Add non-linux support by converting to text format. */
#endif // HAVE_ACL_LIBACL_H
    return 0;
}


/*
 * KACLEntry Class
 */
KACLEntry::KACLEntry( acl_entry_t aclEntry, KFileItem* pItem )
 : m_type( ACL_UNDEFINED_TAG ),
   m_perms( 0 ),
   m_partialPerms( 0 ),
   m_qualifier( QString::null ),
   m_bPartialEntry( false )
{
    int ret;
    if ( aclEntry && pItem )
    {
        // Retrieve the entry type
        ret = acl_get_tag_type( aclEntry, &m_type );
        if ( ret != 0 )
        {
            kdDebug() << "KACLEntry::KACLEntry(): Could not query ACL entry tag type" << endl;
            return;
        }

        ret = getACLQualifier( aclEntry, m_type, pItem, m_qualifier );
        if ( ret != 0 )
        {
            kdDebug() << "KACLEntry::KACLEntry(): Could not query ACL qualifier" << endl;
            return;
        }

        // Which permissions are set?
        getACLPerms( aclEntry, m_perms );
    }
}


KACLEntry::~KACLEntry()
{

}


void KACLEntry::mergePermissions( acl_entry_t aclEntry )
{
    acl_perm_t perms;
    getACLPerms( aclEntry, perms );

    acl_perm_t tmpPerms = m_perms;
    m_perms &= perms;
    m_partialPerms = ( tmpPerms ^ perms ) | m_partialPerms;
}


const QString& KACLEntry::mergeQualifier( acl_entry_t aclEntry, KFileItem* pItem )
{
    int ret;
    if ( aclEntry && pItem )
    {
        QString qualifier;
        ret = getACLQualifier( aclEntry, m_type, pItem, qualifier );
        if ( ret != 0 )
        {
            kdDebug() << "KACLEntry::mergeQualifier(): Could not query ACL qualifier" << endl;
            return m_qualifier;
        }

        if ( qualifier != m_qualifier )
            m_qualifier = QString( "" );
    }

    return m_qualifier;
}


QString KACLEntry::debug() const
{
    QString s;

    // Type
    switch ( m_type )
    {
    case ACL_USER_OBJ:
    case ACL_USER:
        s = "user:";
        break;
    case ACL_GROUP_OBJ:
    case ACL_GROUP:
        s = "group:";
        break;
    case ACL_OTHER:
        s = "other:";
        break;
    case ACL_MASK:
        s = "mask:";
        break;
    default:
        s = "unknown:";
    }

    // Qualifier
    s += m_qualifier + ":";

    // Perms
    if ( m_partialPerms & ACL_READ )
        s += "R";
    else if ( m_perms & ACL_READ )
        s += "r";
    else
        s += "-";

    if ( m_partialPerms & ACL_WRITE )
        s += "W";
    else if ( m_perms & ACL_WRITE )
        s += "w";
    else
        s += "-";

    if ( m_partialPerms & ACL_EXECUTE )
        s += "X";
    else if ( m_perms & ACL_EXECUTE )
        s += "x";
    else
        s += "-";

    if ( m_bPartialEntry )
        s += QString( "\t# Partial Entry" );

    return s;
}


/*
 * KACL Class
 */
KACL::KACL( acl_type_t aclType )
 : m_aclType( aclType ),
   m_bHasEmptyACL( false )
{

}


KACL::~KACL()
{

}


int KACL::addEntries( KFileItemList items )
{
    // Set our list of KACLEntrys so that it represents the ACL on the first file
    KFileItemListIterator it( items );
    KFileItem* pItem = it.current();
    if ( !pItem )
    {
        // Nothing to add
        return 0;
    }

    // If we are interested in the Default ACL - make sure that we have a directory!
    if ( m_aclType == ACL_TYPE_DEFAULT )
    {
        while ( pItem && pItem->isDir() == false )
        {
            ++it;
            pItem = it.current();
        }
    }

    bool foundFirstACL = false;
    while ( !foundFirstACL && pItem )
    {
        QString filename = pItem->url().path( -1 );
        kdDebug() << "Adding ACL entries from " << filename << endl;
        acl_t aclFile = acl_get_file( filename.local8Bit(), m_aclType );
        if ( aclFile == (acl_t) NULL )
        {
            kdDebug() << "There was an error retrieving the ACL for " << filename << endl;
            return 0;
        }

        acl_entry_t aclEntry;
        int ret = acl_get_entry( aclFile, ACL_FIRST_ENTRY, &aclEntry );
        if ( ret == -1 )
        {
            kdDebug() << "There was an error reading the ACL for the file " << filename << endl;
            return 0;
        }
        else if ( ret == 0 )
        {
            kdDebug() << "The file " << filename << " does not have an ACL of this type." << endl;
            kdDebug() << "Flagging that all other ACLs should consist of partial ACL Entries." << endl;
            m_bHasEmptyACL = true;
            foundFirstACL = true;
            /*
            kdDebug() << "The file " << filename << " does not have an ACL of this type. Trying next file/dir (if any)" << endl;
            ++it;
            pItem = it.current();
            continue;
            */
        }
        else
        {
            foundFirstACL = true;
        }

        // Don't try to add any entries from an empty ACL!
        if ( !m_bHasEmptyACL )
        {
            do
            {
                // Create a KACLEntry object and add it to the list
                m_entries.append( new KACLEntry( aclEntry, pItem ) );

            } while ( acl_get_entry( aclFile, ACL_NEXT_ENTRY, &aclEntry ) != 0 );
        }
    }

    // Now iterate through the rest of the files (if any) and merge their
    // ACL entries with our existing list.
    ++it;
    while ( ( pItem = it.current() ) != 0 )
    {
        ++it;
        if ( m_aclType == ACL_TYPE_ACCESS ||
           ( m_aclType == ACL_TYPE_DEFAULT && pItem->isDir() == true ) )
        {
            int ret = mergeEntries( pItem );
            if ( ret != 0 )
                kdDebug() << "Error merging ACL entries for " << pItem->url().path( -1 ) << endl;
        }
    }

    // If we found an empty ACL somewhere along the line, then we must mark
    // all of the entries as partial. This should only happen for default
    // ACLs on directories, as the usual POSIX perms map onto the first
    // 3 entries for access ACLs.
    if ( m_bHasEmptyACL )
    {
        KACLEntry* pEntry;
        QPtrListIterator<KACLEntry> itEntries( m_entries );
        while ( ( pEntry = itEntries.current() ) != 0 )
        {
            ++itEntries;
            pEntry->setPartialEntry( true );
        }
    }

    // Output the ACL for debugging
    QPtrListIterator<KACLEntry> it2( m_entries );
    KACLEntry* pEntry;
    while ( ( pEntry = it2.current() ) != 0 )
    {
        ++it2;
        kdDebug() << pEntry->debug() << endl;
    }

    return m_entries.count();
}


int KACL::mergeEntries( KFileItem* pItem )
{
    QString filename = pItem->url().path( -1 );
    kdDebug() << "Merging ACL entries from " << filename << endl;
    acl_t aclFile = acl_get_file( filename.local8Bit(), m_aclType );
    if ( aclFile == (acl_t) NULL )
    {
        kdDebug() << "There was an error retrieving the ACL for " << filename << endl;
        return 0;
    }

    acl_entry_t aclEntry;
    int ret = acl_get_entry( aclFile, ACL_FIRST_ENTRY, &aclEntry );
    if ( ret == -1 )
    {
        kdDebug() << "There was an error reading the ACL for the file " << filename << endl;
        return 0;
    }
    else if ( ret == 0 )
    {
        kdDebug() << "The file " << filename << " does not have an ACL of this type." << endl;
        kdDebug() << "Flagging that all other ACLs should consist of partial ACL Entries." << endl;
        m_bHasEmptyACL = true;
    }

    // Keep a list of pointers to entries that are common to the current list and the
    // file being processed at the moment. This is so that we can mark any pre-existing
    // entries as partial if they belong to a previous file but not this file.
    QPtrList<KACLEntry> commonEntries;
    do
    {
        // Do we have an entry of this type and maybe qualifier?
        acl_tag_t aclType;
        ret = acl_get_tag_type( aclEntry, &aclType );
        if ( ret != 0 )
        {
            kdDebug() << "Could not query ACL entry tag type" << endl;
            return errno;
        }

        QString qualifier;
        ret = getACLQualifier( aclEntry, aclType, pItem, qualifier );
        if ( ret != 0 )
        {
            kdDebug() << "KACL::mergeEntries( " << filename << " ): Could not query ACL qualifier" << endl;
            return ret;
        }

        bool bFoundEntry = false;
        QPtrListIterator<KACLEntry> it( m_entries );
        KACLEntry* pEntry;
        while ( ( pEntry = it.current() ) != 0 )
        {
            ++it;
            if ( aclType == ACL_USER_OBJ || aclType == ACL_GROUP_OBJ ||
                 aclType == ACL_OTHER || aclType == ACL_MASK )
            {
                // We only need to match the entry type as there can be only one of each of
                // these entry types per ACL.
                if ( pEntry->m_type == aclType )
                {
                    bFoundEntry = true;
                    break;
                }
            }
            else
            {
                // We need to match the type and qualifier for the case of an ACL_USER or ACL_GROUP entry.
                if ( pEntry->m_type == aclType && pEntry->m_qualifier == qualifier )
                {
                    bFoundEntry = true;
                    break;
                }
            }
        }

        if ( bFoundEntry == true )
        {
            // OK we have found an entry with the same type and qualifier.
            // Tell the entry to merge in the permissions from the entry attached
            // to the file being processed.
            pEntry->mergePermissions( aclEntry );

            // If we have an ACL_USER_OBJ or ACL_GROUP_OBJ entry, then we also need to merge the
            // qualifiers. If the qualifiers do not match then we do not display a qualifier.
            if ( aclType == ACL_USER_OBJ || aclType == ACL_GROUP_OBJ )
                pEntry->mergeQualifier( aclEntry, pItem );

            // Add this entry to the list of common entries if it is not already there.
            bool bCommonEntry = false;
            QPtrListIterator<KACLEntry> it2( commonEntries );
            KACLEntry* pCommonEntry;
            while ( ( pCommonEntry = it2.current() ) != 0 )
            {
                ++it2;
                if ( pCommonEntry == pEntry )
                {
                    // This is entry is already marked as being a common entry
                    // no need to add it again.
                    bCommonEntry = true;
                    break;
                }
            }

            if ( bCommonEntry == false )
                commonEntries.append( pEntry );
        }
        else
        {
            // There are no matching entries already in the list.
            // Create a KACLEntry object and add it to the list as a partial entry.
            pEntry = new KACLEntry( aclEntry, pItem );
            pEntry->setPartialEntry( true );
            m_entries.append( pEntry );
        }
    } while ( acl_get_entry( aclFile, ACL_NEXT_ENTRY, &aclEntry ) != 0 );


    // Test to see if any existing ACL entries need to be marked as partial
    // after merging in the ACL for this file.
    // NB. We only need to do this if m_bHasEmptyACL is false, since if it
    // is true, then the addEntries function will set all entries to be partial
    // regardless. This is because no existing entry can be the same as the
    // absence of an entry.
    if ( !m_bHasEmptyACL )
    {
        KACLEntry* pEntry;
        KACLEntry* pCommonEntry;

        // Iterate through the list of entries
        QPtrListIterator<KACLEntry> itEntries( m_entries );
        while ( ( pEntry = itEntries.current() ) != 0 )
        {
            ++itEntries;
            // Compare this entry with each entry in the commonEntries list
            // If it is not there, then mark it as a partial entry.
            bool bCommonEntry = false;
            QPtrListIterator<KACLEntry> itCommonEntries( commonEntries );
            while ( ( pCommonEntry = itCommonEntries.current() ) != 0 )
            {
                ++itCommonEntries;
                if ( pCommonEntry == pEntry )
                {
                    bCommonEntry = true;
                    break;
                }
            }

            if ( bCommonEntry == false )
                pEntry->setPartialEntry( true );
        }
    }

    return 0;
}


bool KACL::hasEntryOfType( acl_tag_t entryType ) const
{
    QPtrListIterator<KACLEntry> itEntries( m_entries );
    KACLEntry* pEntry;
    while ( ( pEntry = itEntries.current() ) != 0 )
    {
        ++itEntries;
        if ( pEntry->entryType() == entryType )
            return true;
    }
    return false;
}


void KACL::removeExtendedEntries()
{
    kdDebug() << "KACL::removeExtendedEntries()" << endl;
    QPtrListIterator<KACLEntry> itEntries( m_entries );
    KACLEntry* pEntry;
    while ( ( pEntry = itEntries.current() ) != 0 )
    {
        ++itEntries;
        if ( pEntry->entryType() == ACL_USER ||
             pEntry->entryType() == ACL_GROUP ||
             pEntry->entryType() == ACL_MASK )
        {
            kdDebug() << "Removing extended entry" << endl;
            m_entries.removeRef( pEntry );
        }
    }
}


KACLEntryDlg::KACLEntryDlg( QWidget* parent, const char* name, bool modal, WFlags fl )
 : QDialog( parent, name, modal, fl )
{
    QPixmap pmUser = QPixmap( qembed_findImage( "user" ) );
    QPixmap pmGroup = QPixmap( qembed_findImage( "group" ) );
    QPixmap pmMask = QPixmap( qembed_findImage( "mask" ) );

    setSizeGripEnabled( false );
    QVBoxLayout* KACLEntryDlgLayout = new QVBoxLayout( this, 11, 6, "KACLEntryDlgLayout" );
    KACLEntryDlgLayout->setResizeMode( QLayout::Minimum );

    QGridLayout* layout1 = new QGridLayout( 0, 1, 1, 0, 6, "layout1");
    textLabel1 = new QLabel( this, "textLabel1" );
    textLabel1->setText( i18n( "Entry &Type:" ) );
    layout1->addWidget( textLabel1, 0, 0 );
    textLabel2 = new QLabel( this, "textLabel2" );
    textLabel2->setText( i18n( "&Qualifier:" ) );
    layout1->addWidget( textLabel2, 1, 0 );
    QHBoxLayout* layout2 = new QHBoxLayout( 0, 0, 6, "layout2");

    m_comboType = new QComboBox( FALSE, this, "m_comboType" );
    textLabel1->setBuddy( m_comboType );
    m_comboType->insertItem( pmUser, i18n( "User" ) );
    m_comboType->insertItem( pmGroup, i18n( "Group" ) );
    m_comboType->insertItem( pmMask, i18n( "Mask" ) );
    QToolTip::add( m_comboType, i18n( "Select the entry type" ) );
    layout2->addWidget( m_comboType );

    QSpacerItem* spacer2 = new QSpacerItem( 80, 20, QSizePolicy::Expanding, QSizePolicy::Minimum );

    layout2->addItem( spacer2 );
    layout1->addLayout( layout2, 0, 1 );

    m_comboQualifier = new QComboBox( FALSE, this, "m_comboQualifier" );
    textLabel2->setBuddy( m_comboQualifier );
    m_comboQualifier->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)7, (QSizePolicy::SizeType)0, 0, 0, m_comboQualifier->sizePolicy().hasHeightForWidth() ) );
    QToolTip::add( m_comboQualifier, i18n( "Select the user/group that this entry applies to" ) );
    layout1->addWidget( m_comboQualifier, 1, 1 );

    KACLEntryDlgLayout->addLayout( layout1 );

    groupBox1 = new QGroupBox( this, "groupBox1" );
    groupBox1->setTitle( i18n( "Permissions" ) );
    groupBox1->setColumnLayout(0, Qt::Vertical );
    groupBox1->layout()->setSpacing( 6 );
    groupBox1->layout()->setMargin( 11 );
    QHBoxLayout* groupBox1Layout = new QHBoxLayout( groupBox1->layout() );
    groupBox1Layout->setAlignment( Qt::AlignTop );
    m_cbRead = new QCheckBox( groupBox1, "m_cbRead" );
    m_cbRead->setAccel( QKeySequence( tr( "Alt+R" ) ) );
    m_cbRead->setText( i18n( "&Read" ) );
    groupBox1Layout->addWidget( m_cbRead );
    m_cbWrite = new QCheckBox( groupBox1, "m_cbWrite" );
    m_cbWrite->setAccel( QKeySequence( tr( "Alt+W" ) ) );
    m_cbWrite->setText( i18n( "&Write" ) );
    groupBox1Layout->addWidget( m_cbWrite );
    m_cbExecute = new QCheckBox( groupBox1, "m_cbExecute" );
    m_cbExecute->setAccel( QKeySequence( tr( "Alt+X" ) ) );
    m_cbExecute->setText( i18n( "E&xecute" ) );
    groupBox1Layout->addWidget( m_cbExecute );
    KACLEntryDlgLayout->addWidget( groupBox1 );

    QHBoxLayout* Layout1 = new QHBoxLayout( 0, 0, 6, "Layout1");
    m_pbHelp = new QPushButton( this, "m_pbHelp" );
    m_pbHelp->setText( i18n( "&Help" ) );
    m_pbHelp->setAutoDefault( true );
    Layout1->addWidget( m_pbHelp );
    QSpacerItem* Horizontal_Spacing2 = new QSpacerItem( 20, 20, QSizePolicy::Expanding, QSizePolicy::Minimum );
    Layout1->addItem( Horizontal_Spacing2 );
    m_pbOK = new QPushButton( this, "m_pbOK" );
    m_pbOK->setText( i18n( "&OK" ) );
    m_pbOK->setAutoDefault( true );
    m_pbOK->setDefault( true );
    Layout1->addWidget( m_pbOK );
    m_pbCancel = new QPushButton( this, "m_pbCancel" );
    m_pbCancel->setText( i18n( "&Cancel" ) );
    m_pbCancel->setAutoDefault( true );
    Layout1->addWidget( m_pbCancel );
    KACLEntryDlgLayout->addLayout( Layout1 );

    resize( QSize(235, 172).expandedTo(minimumSizeHint()) );
    clearWState( WState_Polished );

    // signals and slots connections
    connect( m_pbOK, SIGNAL( clicked() ), this, SLOT( accept() ) );
    connect( m_pbCancel, SIGNAL( clicked() ), this, SLOT( reject() ) );
}


KACLEntryDlg::~KACLEntryDlg()
{

}



/*
 * KFileACLPropsPlugin Class
 */
class KFileACLPropsPlugin::KFileACLPropsPluginPrivate
{
public:
    KFileACLPropsPluginPrivate()
    {
    }

    ~KFileACLPropsPluginPrivate()
    {
    }
};


KFileACLPropsPlugin::KFileACLPropsPlugin( KFileItemList items, QWidget* parent, const char* name )
 : QWidget( parent, name ),
   m_aclAccess(),
   m_aclDefault( ACL_TYPE_DEFAULT ),
   m_pACL( 0 )
{
    kdDebug() << "Entering KFileACLPropsPlugin::KFileACLPropsPlugin() with " << items.count()
              << " file(s) to process" << endl;
    // Private d pointer for future expansion
    d = new KFileACLPropsPluginPrivate;

    // We operate on the access control list by default
    m_pACL = &m_aclAccess;

    // Fake the fact that we have directories
    //bool hasDir = true;
    //bool isDir = true;
    KFileItemListIterator it( items );
    KFileItem* pItem = it.current();

    bool isDir = pItem->isDir(); // all dirs
    bool hasDir = pItem->isDir(); // at least one dir

    while ( ( pItem = it.current() ) != 0 )
    {
        ++it;
        if ( pItem->isDir() != isDir )
            isDir = false;
        hasDir |= pItem->isDir();
    }

    // Create an ACL that contains the union of ACL entries associated with each file in the list.
    // We always have an access ACL.
    int entryCount = m_aclAccess.addEntries( items );
    kdDebug() << "KACL object m_aclAccess now has " << entryCount << " entries" << endl;

    if ( isDir || hasDir )
    {
        entryCount = m_aclDefault.addEntries( items );
        kdDebug() << "KACL object m_aclDefault now has " << entryCount << " entries" << endl;
    }

    //
    // Create the GUI
    //
    QVBoxLayout* layoutMain = new QVBoxLayout( this, 11, 6, "KFileACLPropsDlgLayout");

    m_gbACL = new QGroupBox( this, "m_gbACL" );
    m_gbACL->setTitle( i18n( "Control Lists" ) );
    m_gbACL->setColumnLayout(0, Qt::Vertical );
    m_gbACL->layout()->setSpacing( 6 );
    m_gbACL->layout()->setMargin( 11 );
    QVBoxLayout* m_gbACLLayout = new QVBoxLayout( m_gbACL->layout() );
    m_gbACLLayout->setAlignment( Qt::AlignTop );

    // Do we have any directories?
    if ( isDir || hasDir )
    {
        // Yes, we need to be able to get at the default acl too.
        QHBoxLayout* layout2 = new QHBoxLayout( 0, 0, 6, "layout2");

        m_lblControlList = new QLabel( m_gbACL, "m_lblControlList" );
        m_lblControlList->setText( i18n( "Control List:" ) );
        layout2->addWidget( m_lblControlList );

        m_comboControlList = new KComboBox( FALSE, m_gbACL, "m_comboControlList" );
        m_comboControlList->insertItem( i18n( "Access" ) );
        m_comboControlList->insertItem( i18n( "Default" ) );
        QToolTip::add( m_comboControlList, i18n( "Select the type of control list" ) );
        layout2->addWidget( m_comboControlList );
        QSpacerItem* spacer2 = new QSpacerItem( 40, 20, QSizePolicy::Expanding, QSizePolicy::Minimum );
        layout2->addItem( spacer2 );
        m_gbACLLayout->addLayout( layout2 );
    }

    QHBoxLayout* layout6 = new QHBoxLayout( 0, 0, 6, "layout6");

    m_lvACL = new KACLListView( &m_aclAccess, m_gbACL, "m_lvACL" );
    layout6->addWidget( m_lvACL );

    QVBoxLayout* layout5 = new QVBoxLayout( 0, 0, 6, "layout5");

    m_pbAdd = new QPushButton( m_gbACL, "m_pbAdd" );
    m_pbAdd->setText( i18n( "&Add..." ) );
    m_pbAdd->setAccel( QKeySequence( i18n( "Alt+A" ) ) );
    QToolTip::add( m_pbAdd, i18n( "Add a new entry to the control list" ) );
    layout5->addWidget( m_pbAdd );

    m_pbModify = new QPushButton( m_gbACL, "m_pbModify" );
    m_pbModify->setText( i18n( "&Modify..." ) );
    m_pbModify->setAccel( QKeySequence( i18n( "Alt+M" ) ) );
    QToolTip::add( m_pbModify, i18n( "Modify the selected entries" ) );
    layout5->addWidget( m_pbModify );

    m_pbRemove = new QPushButton( m_gbACL, "m_pbRemove" );
    m_pbRemove->setText( i18n( "&Remove" ) );
    m_pbRemove->setAccel( QKeySequence( i18n( "Alt+R" ) ) );
    QToolTip::add( m_pbRemove, i18n( "Remove the selected entries" ) );
    layout5->addWidget( m_pbRemove );

    m_pbDeleteACL = new QPushButton( m_gbACL, "m_pbDeleteACL" );
    m_pbDeleteACL->setText( i18n( "&Delete ACL" ) );
    m_pbDeleteACL->setAccel( QKeySequence( i18n( "Alt+D" ) ) );
    QToolTip::add( m_pbRemove, i18n( "Delete the access control list" ) );
    layout5->addWidget( m_pbDeleteACL );

    QSpacerItem* spacer3 = new QSpacerItem( 20, 30, QSizePolicy::Minimum, QSizePolicy::Expanding );
    layout5->addItem( spacer3 );
    layout6->addLayout( layout5 );
    m_gbACLLayout->addLayout( layout6 );
    layoutMain->addWidget( m_gbACL );

    m_gbSpecial = new QGroupBox( this, "m_gbSpecial" );
    m_gbSpecial->setTitle( i18n( "Special Flags" ) );
    m_gbSpecial->setColumnLayout(0, Qt::Vertical );
    m_gbSpecial->layout()->setSpacing( 6 );
    m_gbSpecial->layout()->setMargin( 11 );
    QHBoxLayout* m_gbSpecialLayout = new QHBoxLayout( m_gbSpecial->layout() );
    m_gbSpecialLayout->setAlignment( Qt::AlignTop );

    m_cbSetUID = new QCheckBox( m_gbSpecial, "m_cbSetUID" );
    m_cbSetUID->setText( i18n( "Set &UID" ) );
    m_cbSetUID->setAccel( QKeySequence( i18n( "Alt+U" ) ) );
    m_gbSpecialLayout->addWidget( m_cbSetUID );

    m_cbSetGID = new QCheckBox( m_gbSpecial, "m_cbSetGID" );
    m_cbSetGID->setText( i18n( "Set &GID" ) );
    m_cbSetGID->setAccel( QKeySequence( i18n( "Alt+G" ) ) );
    m_gbSpecialLayout->addWidget( m_cbSetGID );

    m_cbSticky = new QCheckBox( m_gbSpecial, "m_cbSticky" );
    m_cbSticky->setText( i18n( "&Sticky" ) );
    m_cbSticky->setAccel( QKeySequence( i18n( "Alt+S" ) ) );
    m_gbSpecialLayout->addWidget( m_cbSticky );
    QSpacerItem* spacer4 = new QSpacerItem( 40, 20, QSizePolicy::Expanding, QSizePolicy::Minimum );
    m_gbSpecialLayout->addItem( spacer4 );
    layoutMain->addWidget( m_gbSpecial );

    // Can we recurse?
    if ( isDir || hasDir )
    {
        m_cbRecursive = new QCheckBox( this, "m_cbRecursive" );
        m_cbRecursive->setText( i18n( "A&pply all changes to subfolders and their contents" ) );
        m_cbRecursive->setAccel( QKeySequence( i18n( "Alt+P" ) ) );
        layoutMain->addWidget( m_cbRecursive );
        connect( m_comboControlList, SIGNAL( activated( int ) ), this, SLOT( slotAccessListChanged( int ) ) );
    }

    connect( m_lvACL, SIGNAL( selectionChanged() ), this, SLOT( slotSelectedEntriesChanged() ) );
    connect( m_pbRemove, SIGNAL( clicked() ), this, SLOT( removeSelectedEntries() ) );
    connect( m_pbAdd, SIGNAL( clicked() ), this, SLOT( showAddEntryDlg() ) );
    connect( m_pbModify, SIGNAL( clicked() ), this, SLOT( showModifyEntryDlg() ) );
    connect( m_pbDeleteACL, SIGNAL( clicked() ), this, SLOT( deleteACL() ) );
}


KFileACLPropsPlugin::~KFileACLPropsPlugin()
{

}


void KFileACLPropsPlugin::slotAccessListChanged( int index )
{
    //kdDebug() << "KFileACLPropsPlugin::slotAccessListChanged( " << index << " ) " << endl;
    m_pACL = ( index == 0 ) ? &m_aclAccess : &m_aclDefault;
    m_lvACL->setACL( m_pACL );
    updateDeleteACLButton();
}


void KFileACLPropsPlugin::slotSelectedEntriesChanged()
{
    // Enable/Disable buttons depending upon which ACL entries are selected

    // Iterate through the items to see which are selected
    QListViewItemIterator it( m_lvACL );
    KACLListViewItem* pItem;
    int selectedCount = 0;
    bool bUndeleteableSelected = false;
    bool bMaskSelected = false;
    while ( ( pItem = dynamic_cast<KACLListViewItem*>( it.current() ) ) != 0 )
    {
        ++it;
        if ( pItem->isSelected() )
        {
            if ( pItem->entryType() == ACL_USER_OBJ ||
                 pItem->entryType() == ACL_GROUP_OBJ ||
                 pItem->entryType() == ACL_OTHER )
            {
                bUndeleteableSelected = true;
            }
            else if ( pItem->entryType() == ACL_MASK )
            {
                bMaskSelected = true;
            }

            ++selectedCount;
        }
    }

    bool bEnable = selectedCount > 0 &&
                   ( ( !bUndeleteableSelected && !bMaskSelected ) ||
                   ( !bUndeleteableSelected && bMaskSelected && ( selectedCount == m_pACL->entryCount() - 3 ) ) );
    m_pbRemove->setEnabled( bEnable );
}


void KFileACLPropsPlugin::removeSelectedEntries()
{
    // Iterate through the items to see which are selected
    QListViewItemIterator it( m_lvACL );
    KACLListViewItem* pItem;
    while ( ( pItem = dynamic_cast<KACLListViewItem*>( it.current() ) ) != 0 )
    {
        ++it;
        if ( pItem->isSelected() )
        {
            // Delete the listviewitem and the entry itself
            KACLEntry* pEntry = pItem->entry();
            m_pACL->removeEntry( pEntry );
            delete pEntry;
            delete pItem;
        }
    }

    updateDeleteACLButton();
}


void KFileACLPropsPlugin::showAddEntryDlg()
{
    KACLEntryDlg dlg( this );
    dlg.setCaption( i18n( "Add an ACL Entry" ) );
    dlg.exec();
}


void KFileACLPropsPlugin::showModifyEntryDlg()
{

}


void KFileACLPropsPlugin::deleteACL()
{
    if ( m_pACL->aclType() == ACL_TYPE_DEFAULT )
    {
        // We can legally delete the entire ACL
        m_pACL->clearEntries();
    }
    else
    {
        // We must leave the ACL_USER_OBJ, ACL_GROUP_OBJ, and
        // ACL_OTHER entries in place as these are required for
        // a valid ACL
        m_pACL->removeExtendedEntries();
    }

    m_lvACL->setACL( m_pACL );
    m_lvACL->update();
    updateDeleteACLButton();
    slotSelectedEntriesChanged();
}


void KFileACLPropsPlugin::updateDeleteACLButton()
{
    if ( m_pACL->aclType() == ACL_TYPE_DEFAULT )
    {
        // It is always possible to delete a default ACL as long as it has
        // at least one entry.
        m_pbDeleteACL->setEnabled( m_pACL->entryCount() > 0 );
    }
    else
    {
        // With an access ACL we can only delete entries of the ACL_USER,
        // ACL_GROUP, and ACL_MASK types.
        bool bEnable = m_pACL->hasEntryOfType( ACL_MASK );
        bEnable |= m_pACL->hasEntryOfType( ACL_USER );
        bEnable |= m_pACL->hasEntryOfType( ACL_GROUP );
        m_pbDeleteACL->setEnabled( bEnable );
    }
}

#include "kfileaclpropsplugin.moc"
