/**************************************************************************
*   Copyright (C) 2009 by Ian Reinhart Geiser                             *
*   geiseri@kde.org                                                       *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "launcher.h"
#include "ui_launcher.h"
#include "applicationmenu.h"

#include <kworkspace/kworkspace.h>

#include <KDE/KGlobal>
#include <KDE/KStandardDirs>
#include <KDE/KDesktopFile>
#include <KDE/KConfigGroup>
#include <KDE/KToolInvocation>
#include <KDE/KRun>
#include <KDE/KEMailSettings>
#include <KDE/KStyle>
#include <KDE/KIcon>
#include <KDE/KPageWidgetItem>
#include <KDE/KDirWatch>
#include <KDE/KMimeType>
#include <KDE/KApplication>
#include <KDE/KSycoca>
#include <KDE/KMessageBox>

#include <QDir>
#include <QFileInfo>
#include <QFile>
#include <QVBoxLayout>
#include <QTimer>
#include <QDesktopServices>
#include <QStyleFactory>
#include <QTimer>

#include <kdebug.h>

ApplicationCommandLinkButton::ApplicationCommandLinkButton( QWidget *parent )
{
    connect( this, SIGNAL(clicked()), this, SLOT(executeCommandLink()));
}

void ApplicationCommandLinkButton::setDesktopFilePath( const QString &path )
{
    m_desktopFilePath = path;
}

QString ApplicationCommandLinkButton::desktopFilePath( ) const
{
    return m_desktopFilePath;
}

void ApplicationCommandLinkButton::setDesktopStartUrl( const QString &url )
{
    m_desktopStartUrl = url;
}

QString ApplicationCommandLinkButton::desktopStartUrl( ) const
{
    return m_desktopStartUrl;
}

void ApplicationCommandLinkButton::executeCommandLink()
{
    QString errorString;
    QStringList urls;
    if( !m_desktopStartUrl.isEmpty() )
        urls << m_desktopStartUrl;

    if( m_desktopFilePath.isEmpty() )
    {
        kDebug() << "Prepare to launch" << m_desktopStartUrl;
         new KRun( KUrl( m_desktopStartUrl ), this);
    }
    else
    {
        QString errorString;
        if( KToolInvocation::startServiceByDesktopPath( m_desktopFilePath, urls, &errorString ) > 0)
        {
            KMessageBox::detailedSorry( this, i18n("Failed to execute command"),  errorString );
            kDebug() << "Error:"  << errorString;
        }
    }

    emit activated(1);
}

Launcher::Launcher(QWidget *parent) :
    QFrame(parent),
    m_ui(new Ui::Launcher)
{
    setObjectName("Launcher");
    ApplicationMenu *menu = new ApplicationMenu(this);
    connect( menu, SIGNAL(aboutToHide()), this, SIGNAL(popupDone()));

    m_ui->setupUi(this);
    m_ui->logout->setIcon( KIcon("system-log-out.png"));

    m_ui->applications->setIcon( KIcon("start-here.png") );
    m_ui->applications->setMenu(menu);

    m_watcher = new KDirWatch(this);
    foreach( QString profiles, KGlobal::dirs()->findDirs("data", "blazer/profiles/"))
    {
        m_watcher->addDir( profiles );
    }
    connect(m_watcher, SIGNAL(created(QString)), this, SLOT(reloadTasks()));
    connect(m_watcher, SIGNAL(deleted(QString)), this, SLOT(reloadTasks()));
    connect(m_watcher, SIGNAL(dirty(QString)), this, SLOT(reloadTasks()));
    connect( KSycoca::self(), SIGNAL(databaseChanged()), this, SLOT(reloadTasks()));

    QTimer::singleShot( 0, this, SLOT(setupTasks()));
    QTimer::singleShot( 0, this, SLOT(setupShortcuts()));
}

Launcher::~Launcher()
{
    delete m_ui;
}

void Launcher::changeEvent(QEvent *e)
{
    switch (e->type()) {
    case QEvent::LanguageChange:
        m_ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void Launcher::configureButtonFromDesktopFile( ApplicationCommandLinkButton *button, const QString &desktopFileName )
{
    QString desktopFilePath;
    KService::Ptr service = KService::serviceByDesktopName(desktopFileName);
    if( service )
    {
        desktopFilePath = service->entryPath();
    }
    if( desktopFilePath.isEmpty() )
    {
        button->hide();
    }
    else
    {
        KDesktopFile desktopFile( desktopFilePath );
        button->setText( desktopFile.readGenericName() );
        button->setDescription( desktopFile.readName() );
        button->setIcon( KIcon( desktopFile.readIcon() ));
        button->setDesktopFilePath( desktopFilePath );

        connect( button, SIGNAL(activated(int)), this, SIGNAL(popupDone(int)));

        button->show();
    }
}

void Launcher::setupShortcuts()
{
    // Read emaildefaults
    // Read kdeglobals
    //
    configureButtonFromDesktopFile( m_ui->webSelection, QLatin1String("konqbrowser") );
    configureButtonFromDesktopFile( m_ui->emailSelection, QLatin1String("Kontact") );
    configureButtonFromDesktopFile( m_ui->imSelection, QLatin1String("kopete") );
    configureButtonFromDesktopFile( m_ui->settings, QLatin1String("systemsettings") );

    KConfigGroup paths = KGlobal::config()->group(QLatin1String("Paths"));
    QString documentsDirectory = paths.readPathEntry( QLatin1String("Documents"),
                                                  QDesktopServices::storageLocation(QDesktopServices::DocumentsLocation) );
    m_ui->documents->setIcon( KIcon("folder.png") );
    m_ui->documents->setDesktopStartUrl( documentsDirectory );
    connect( m_ui->documents, SIGNAL(activated(int)), this, SIGNAL(popupDone(int)));
}


void Launcher::setupTasks()
{
    bool isEmpty = true;
    foreach( QString profiles, KGlobal::dirs()->findDirs("data", "blazer/profiles/") )
    {
        QDir profilesDir( profiles );
        foreach( QFileInfo profileInfo, profilesDir.entryInfoList( QStringList() << "*.profile", QDir::Files, QDir::Name ) )
        {
            kDebug() << "Reading for" << profileInfo.absoluteFilePath();
            m_watcher->addFile( profileInfo.absoluteFilePath() );

            KConfig taskProfile( profileInfo.absoluteFilePath(), KConfig::SimpleConfig );
            QWidget *taskWidget = new QWidget(this);
            QVBoxLayout *taskLayout = new QVBoxLayout( taskWidget );
            taskLayout->setSpacing(0);
            taskLayout->setMargin(0);

            bool noApps = true;
            KConfigGroup taskProfileDetails( &taskProfile, "Profile");
            QStringList taskProfileEntryNameList = taskProfile.groupList();
            taskProfileEntryNameList.removeAll( "Profile" );

            foreach( QString taskProfileEntryName, taskProfileEntryNameList )
            {                
                KConfigGroup taskProfileEntry( &taskProfile, taskProfileEntryName );
                QString taskProfileEntryType = taskProfileEntry.readEntry("Type", QString("Application"));
                QString taskProfileEntryFile = taskProfileEntry.readEntry("Desktop File", QString() );
                QString taskProfileEntryIcon = taskProfileEntry.readEntry("Icon", QString() );
                QString taskProfileEntryTitle = taskProfileEntry.readEntry("Title", QString() );
                QString taskProfileEntryDescription = taskProfileEntry.readEntry("Description", QString() );
                QString taskProfileEntryUrl = taskProfileEntry.readEntry("Url", QString() );
                QString desktopFilePath;

                if( !(taskProfileEntryUrl.isEmpty() && taskProfileEntryFile.isEmpty()) )
                {
                    isEmpty = false;
                    noApps = false;
                    if( taskProfileEntryType == QLatin1String( "Application" ) )
                    {
                        KService::Ptr service = KService::serviceByDesktopName(taskProfileEntryFile);
                        if( service )
                        {
                            desktopFilePath = service->entryPath();
                        }
                        KDesktopFile desktopFile( desktopFilePath );
                        if( !desktopFile.readType().isEmpty() )
                        {
                            if( taskProfileEntryIcon.isEmpty() ) taskProfileEntryIcon = desktopFile.readIcon();
                            if( taskProfileEntryDescription.isEmpty() ) taskProfileEntryDescription = desktopFile.readName();
                            if( taskProfileEntryTitle.isEmpty() ) taskProfileEntryTitle = desktopFile.readGenericName( );
                            if( taskProfileEntryTitle.isEmpty() )
                            {
                                taskProfileEntryTitle = desktopFile.readName( );
                                taskProfileEntryDescription.clear();
                            }
                        }
                    }
                    else if ( taskProfileEntryType == QLatin1String( "Url" ) )
                    {
                        if( taskProfileEntryDescription.isEmpty() )
                            taskProfileEntryDescription = taskProfileEntryUrl;
                        if( taskProfileEntryTitle.isEmpty() )
                            taskProfileEntryTitle = taskProfileEntryName;
                        if( taskProfileEntryIcon.isEmpty() )
                            taskProfileEntryIcon = KMimeType::favIconForUrl(  taskProfileEntryUrl );
                        if( taskProfileEntryIcon.isEmpty() )
                            taskProfileEntryIcon = KMimeType::iconNameForUrl( taskProfileEntryUrl );
                    }

                    if( !desktopFilePath.isEmpty() || !taskProfileEntryUrl.isEmpty() )
                    {
                        ApplicationCommandLinkButton *profileButton = new ApplicationCommandLinkButton( taskWidget );
                        taskLayout->addWidget( profileButton );
                        profileButton->setText( taskProfileEntryTitle );
                        profileButton->setDescription( taskProfileEntryDescription );
                        profileButton->setIcon( KIcon( taskProfileEntryIcon));
                        profileButton->setDesktopStartUrl( taskProfileEntryUrl );
                        profileButton->setDesktopFilePath( desktopFilePath );
                        profileButton->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);

                        connect( profileButton, SIGNAL(activated(int)), this, SIGNAL(popupDone(int)));
                    }
                }
            }
            taskLayout->addStretch(1);

            if( noApps )
                delete taskWidget;
            else
            {
                KPageWidgetItem *item = new KPageWidgetItem( taskWidget, taskProfileDetails.readEntry("Name", QString() ));
                item->setHeader( taskProfileDetails.readEntry("Description", QString() ) );
                item->setIcon( KIcon( taskProfileDetails.readEntry("Icon", QString() ) ) );
                m_ui->taskPages->addPage( item );
            }

        }
    }

    if( isEmpty )
    {
        m_ui->taskPages->hide();
    }
    else
    {

    }
}

void Launcher::on_logout_clicked()
{
    QTimer::singleShot(0, this, SLOT(slotLogout()));
    emit popupDone(1);
}

void Launcher::slotLogout()
{
    KWorkSpace::requestShutDown( KWorkSpace::ShutdownConfirmDefault,
                                 KWorkSpace::ShutdownTypeDefault,
                                 KWorkSpace::ShutdownModeDefault );
}

void Launcher::execByServiceType( const QString &app )
{
    KService::Ptr service = KService::serviceByStorageId("kde4-" + app + ".desktop");
    if( service )
    {
        QString errorString;
        if( KToolInvocation::startServiceByDesktopPath( service->entryPath(), QStringList(), &errorString ) > 0)
        {
            kDebug() << "Error:"  << errorString;
        }
    }
}

void Launcher::reloadTasks()
{
    kDebug() << "Reload task pages.";
    while( m_ui->taskPages->currentPage() )
        m_ui->taskPages->removePage( m_ui->taskPages->currentPage() );
    setupTasks();
}
