/************************************************************************
*  This file is part of the KDE project, module kcontrol.               *
*  Copyright (C) 2007 Ian Ventura-Whiting <fizz@titania.co.uk>          *
*                                                                       *
*  You can Freely distribute this program under the GNU LGPL version 2  *
*************************************************************************/

// Qt Includes...
#include <QLabel>
#include <QTreeWidget>
#include <QString>
#include <QStringList>
#include <QLabel>
#include <QWidget>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QGridLayout>
#include <QRadioButton>
#include <QSpacerItem>

// KDE Includes...
#include <KPageWidgetModel>
#include <KComboBox>
#include <KLineEdit>
#include <KPushButton>
#include <KUrlRequester>

// ODBC/SQL/DB Includes...
#include <sqlext.h>
#include <sqltypes.h>

// Other Includes...
#include "dsn_dialog.h"

// Defines...
#define ODBC_BUFFER_SIZE 9600
#define ODBC_FILEDSN_SIZE 256


// ----------------------------------------------------------------------------------------
// Constructor and Destructor

DSNDialog::DSNDialog(QWidget *parent) : KPageDialog( parent )
{
	// Set defaults...
	dialogMode = DSNDialog::Add;
	odbcType = DSNDialog::User;
	dsnName.clear();
	dsnDir.clear();
	odbcProperties = 0;
	dsnPage = 0;
	driverPage = 0;
	basicPage = 0;
	advancedPage = 0;
	loggingPage = 0;
	finishPage = 0;

	// Setup Buttons...
	setButtons(KDialog::Cancel | KDialog::User1 | KDialog::User2);
	setButtonText(KDialog::User1, tr("Next"));
	setButtonText(KDialog::User2, tr("Previous"));
	enableButton(KDialog::User2, FALSE);

	// SIGNALS / SLOTS...
	connect(this, SIGNAL(user1Clicked()), this, SLOT(nextClicked()));
	connect(this, SIGNAL(user2Clicked()), this, SLOT(previousClicked()));
}

DSNDialog::~DSNDialog()
{
	// If structure exists, destroy it!
	if (odbcProperties != 0)
		ODBCINSTDestructProperties(&odbcProperties);
}


// ----------------------------------------------------------------------------------------
// Setup Dialog

bool DSNDialog::setup()
{
	// Declare variables...
	char odbcDrivers[ODBC_BUFFER_SIZE + 1];
	char odbcResults[ODBC_BUFFER_SIZE + 1];
	QString odbcName;
	QTreeWidgetItem *currentItem = 0;
	int odbcDriversPos = 0;
	QStringList header;

	// Set Dialog Config
	setFaceType(Plain);
	setCaption(tr("Add Data Source"));

	// Type of Data Source Group...
	QGroupBox *dsnGroup = new QGroupBox(this);
	dsnGroup->setTitle(tr("Data Source Types"));
	QVBoxLayout *layout = new QVBoxLayout(this);
	dsnGroup->setLayout(layout);

	// User Data Source...
	userDSNRadio = new QRadioButton(dsnGroup);
	userDSNRadio->setText(tr("User data source"));
	if (odbcType == DSNDialog::User)
		userDSNRadio->setChecked(TRUE);
	layout->addWidget(userDSNRadio);
	QLabel *label = new QLabel(dsnGroup);
	label->setText(tr("A user data source will only be accessable by the user that created it."));
	layout->addWidget(label);
	connect(userDSNRadio, SIGNAL(toggled(bool)), this, SLOT(dsnSelectionChanged(bool)));

	// System Data Source...
	systemDSNRadio = new QRadioButton(dsnGroup);
	systemDSNRadio->setText(tr("System data source"));
	if (odbcType == DSNDialog::System)
		systemDSNRadio->setChecked(TRUE);
	layout->addWidget(systemDSNRadio);
	label = new QLabel(dsnGroup);
	label->setText(tr("A system data source is available to all the users of the system."));
	layout->addWidget(label);
	connect(systemDSNRadio, SIGNAL(toggled(bool)), this, SLOT(dsnSelectionChanged(bool)));

	// File Data Source...
	fileDSNRadio = new QRadioButton(dsnGroup);
	fileDSNRadio->setText(tr("File data source"));
	if (odbcType == DSNDialog::File)
		fileDSNRadio->setChecked(TRUE);
	layout->addWidget(fileDSNRadio);
	label = new QLabel(dsnGroup);
	label->setText(tr("A file data source is available to all users who can access the file."));
	layout->addWidget(label);
	connect(fileDSNRadio, SIGNAL(toggled(bool)), this, SLOT(dsnSelectionChanged(bool)));

	// Data Source Spacer...
	QSpacerItem *spacer = new QSpacerItem(0, 0, QSizePolicy::Fixed, QSizePolicy::Expanding);
	layout->addItem(spacer);

	// Data Source Page...
	dsnPage = new KPageWidgetItem(dsnGroup, tr("Select a data source type"));
	addPage(dsnPage);

	// Driver List...
	driverList = new QTreeWidget(this);
	driverList->setColumnCount(2);
	header.clear();
	header << tr("Name");
	header << tr("Description");
	driverList->setHeaderLabels(header);
	driverList->setRootIsDecorated(FALSE);
	driverList->setItemsExpandable(FALSE);
	driverList->setSelectionBehavior(QAbstractItemView::SelectRows);
	driverList->setSelectionMode(QAbstractItemView::SingleSelection);
	driverList->clear();

	// Populate the driver list...
	if (SQLGetInstalledDrivers(odbcDrivers, ODBC_BUFFER_SIZE, 0) > 0)
	{
		odbcName = QString(odbcDrivers);
		while (!odbcName.isEmpty())
		{
			currentItem = new QTreeWidgetItem(driverList);
			currentItem->setText(0, odbcName.toLatin1());
			if (SQLGetPrivateProfileString(odbcName.toLatin1(), "Description", "", odbcResults, ODBC_BUFFER_SIZE, "odbcinst") > 0)
				currentItem->setText(1, QString(odbcResults));
			
			// Get the next driver name...
			odbcDriversPos = odbcDriversPos + odbcName.length() + 1;
			if (odbcDriversPos < ODBC_BUFFER_SIZE)
				odbcName = QString(odbcDrivers + odbcDriversPos);
			else
				odbcName = "";
		}
	}

	// Set default driver selection...
	if (driverList->topLevelItemCount() > 0)
		driverList->setCurrentItem(driverList->topLevelItem(0));
	else
		return FALSE;

	// Driver Page...
	driverPage = new KPageWidgetItem(driverList, tr("Select a database driver"));
	addPage(driverPage);

	// Set first page...
	if (getuid() == 0)
		setCurrentPage(dsnPage);
	else
		setCurrentPage(driverPage);

	return TRUE;
}


// ----------------------------------------------------------------------------------------
// Set Property Methods...delete

// Set the dialog Mode
void DSNDialog::setDialogMode(DSNMode mode)
{
	dialogMode = mode;
}

// Set the ODBC Type
void DSNDialog::setODBCType(DSNType type)
{
	odbcType = type;
}

// Set the DSN Name
void DSNDialog::setDSNName(QString name)
{
	dsnName = name;
}

// Set the DSN Dir Name
void DSNDialog::setDSNDir(QString dirName)
{
	dsnDir = dirName;
}


// ----------------------------------------------------------------------------------------
// Next and Previous Buttons...

// The Next Button has been clicked...
void DSNDialog::nextClicked()
{
	// Variables...
	QLabel *label;
	QGridLayout *gridLayout;
	QSpacerItem *spacer;
	bool finish = FALSE;

	// If on the data source type page...
	if (currentPage() == dsnPage)
	{
		setCurrentPage(driverPage);
		enableButton(KDialog::User2, TRUE);
	}

	// If on the select driver page...
	else if (currentPage() == driverPage)
	{
		getProperties();
		buildInterface();
		setCurrentPage(basicPage);
		enableButton(KDialog::User2, TRUE);
	}

	// If on the basic configuration page...
	else if (currentPage() == basicPage)
	{
		// If there is a advanced page...
		if (advancedPage != 0)
			setCurrentPage(advancedPage);
		// If there is a logging page...
		else if (loggingPage != 0)
			setCurrentPage(loggingPage);
		// If finish page...
		else
			finish = TRUE;
	}

	// If on the advanced configuration page...
	else if (currentPage() == advancedPage)
	{
		// If there is a logging page...
		if (loggingPage != 0)
			setCurrentPage(loggingPage);
		// If finish page...
		else
			finish = TRUE;
	}

	// If on the logging configuration page...
	else if (currentPage() == loggingPage)
		finish = TRUE;

	// If finish was clicked...
	else if (currentPage() == finishPage)
		accept();

	// If Finish page needs creating...
	if (finish == TRUE)
	{
		setButtonText(KDialog::User1, tr("Finish"));
		finishWidget = new QWidget(this);
		label = new QLabel(finishWidget);
		gridLayout = new QGridLayout(finishWidget);
		((QGridLayout *)finishWidget->layout())->addWidget(label, 0, 0);
		if (saveProperties() == TRUE)
		{
			label->setText(tr("Data source successfully configured and is now ready for use. You can test the data source configuration using the test button."));
			label->setWordWrap(TRUE);
			if (odbcType != DSNDialog::File)
			{
				testDSNButton = new KPushButton(finishWidget);
				testDSNButton->setText(tr("Test Data Source"));
				((QGridLayout *)finishWidget->layout())->addWidget(testDSNButton, 1, 0);
				//connect(testDSNButton, SIGNAL(clicked(bool)), this, SLOT(testDSN(void)));
				spacer = new QSpacerItem(0, 0, QSizePolicy::Fixed, QSizePolicy::Expanding);
				((QGridLayout *)finishWidget->layout())->addItem(spacer, 2, 0);
			}
			else
			{
				spacer = new QSpacerItem(0, 0, QSizePolicy::Fixed, QSizePolicy::Expanding);
				((QGridLayout *)finishWidget->layout())->addItem(spacer, 1, 0);
			}
		}
		else
		{
			label->setText(tr("It was not possible to configure the data source. This is probably because it was not possibe to write to odbc.ini or the disk was full."));
			label->setWordWrap(TRUE);
			spacer = new QSpacerItem(0, 0, QSizePolicy::Fixed, QSizePolicy::Expanding);
			((QGridLayout *)finishWidget->layout())->addItem(spacer, 1, 0);
		}
		finishPage = new KPageWidgetItem(finishWidget, tr("Data source configuration complete"));
		addPage(finishPage);
		setCurrentPage(finishPage);
	}
}


// The Previous Button has been clicked...
void DSNDialog::previousClicked()
{
	// If on the driver page...
	if (currentPage() == driverPage)
		setCurrentPage(dsnPage);

	// If on the basic page...
	else if (currentPage() == basicPage)
	{
		if (getuid() != 0)
			enableButton(KDialog::User2, FALSE);
		setCurrentPage(driverPage);
	}

	// If on the advanced page...
	else if (currentPage() == advancedPage)
		setCurrentPage(basicPage);

	// If on the logging page...
	else if (currentPage() == loggingPage)
	{
		if (advancedPage != 0)
			setCurrentPage(advancedPage);
		else
			setCurrentPage(basicPage);
	}

	// If on the finish page...
	else if (currentPage() == finishPage)
	{
		// Set Page...
		if (loggingPage != 0)
			setCurrentPage(loggingPage);
		else if (advancedPage != 0)
			setCurrentPage(advancedPage);
		else
			setCurrentPage(basicPage);

		//Set Button Text...
		setButtonText(KDialog::User1, tr("Next"));

		// Set config mode...
		if (odbcType == DSNDialog::User)
			SQLSetConfigMode(ODBC_USER_DSN);
		else if (odbcType == DSNDialog::System)
			SQLSetConfigMode(ODBC_SYSTEM_DSN);

		// Remove DSN...
		if (odbcType != DSNDialog::File)
			SQLRemoveDSNFromIni(((KLineEdit *)odbcProperties->pWidget)->text().toLatin1());
		else
			unlink(QString("%1%2").arg(dsnDir, ((KLineEdit *)odbcProperties->pWidget)->text()).toLatin1());

		// Reset config mode...
		SQLSetConfigMode(ODBC_BOTH_DSN);

		// Delete finish page...
		removePage(finishPage);
		finishPage->deleteLater();
		finishPage = 0;
	}
}


// ----------------------------------------------------------------------------------------
// Manage ODBC Properties...

// Create ODBC properties...
void DSNDialog::getProperties()
{
	// Variables...
	char odbcResults[ODBC_BUFFER_SIZE];
	HODBCINSTPROPERTY lastDataSourcePointer = NULL;
	bool propertyExists = FALSE;
	QString odbcName = "";
	QString odbcTemp = "";
	QString odbcTemp2 = "";
	char driverName[ODBC_BUFFER_SIZE + 1];
	int odbcResultsPos = 0;
	int getResult = 0;

	// Init
	memset(odbcResults, 0, ODBC_BUFFER_SIZE);

	// If structure exists, destroy it!
	if (odbcProperties != 0)
		ODBCINSTDestructProperties(&odbcProperties);

	// Set config mode...
	if (odbcType == DSNDialog::User)
		SQLSetConfigMode(ODBC_USER_DSN);
	else if (odbcType == DSNDialog::System)
		SQLSetConfigMode(ODBC_SYSTEM_DSN);

	// Build Data Source Properties from Setup Library...
	// Get driver name...
	if (dialogMode == DSNDialog::Add)
		strncpy(driverName, driverList->currentItem()->text(0).toAscii(), sizeof(driverName));
	else
	{
		// if it is a user or system dsn...
		if (odbcType != DSNDialog::File)
			SQLGetPrivateProfileString(dsnName.toLatin1(), "Driver", "", driverName, ODBC_BUFFER_SIZE, 0);
		// file dsn...
		else
			SQLReadFileDSN(dsnName.toLatin1(), "ODBC", "DRIVER", driverName, ODBC_BUFFER_SIZE, NULL);
	}

	if (ODBCINSTConstructProperties(driverName, &odbcProperties) != ODBCINST_SUCCESS)
	{
		// If it did not work, the setup library is probably missing!

		// If this is not for adding a data source, we can start the structure and
		// use the "add missing properties" and pick up the other properties.
		if ((odbcType != DSNDialog::File) && (dialogMode != DSNDialog::Add))
		{
			odbcProperties  = (HODBCINSTPROPERTY)malloc(sizeof(ODBCINSTPROPERTY));
			memset(odbcProperties, 0, sizeof(ODBCINSTPROPERTY));
			odbcProperties->nPromptType = ODBCINST_PROMPTTYPE_TEXTEDIT;
			odbcProperties->pNext = NULL;
			odbcProperties->bRefresh = 0;
			odbcProperties->hDLL = NULL;
			odbcProperties->pWidget = NULL;
			odbcProperties->pszHelp = strdup(tr("<b>Data Source Name</b><br><br>The name of the data source (DSN).").toAscii().constData());
			odbcProperties->aPromptData = NULL;
			strncpy(odbcProperties->szName, "Name", INI_MAX_PROPERTY_NAME);
			strncpy(odbcProperties->szValue, dsnName.toAscii(), INI_MAX_PROPERTY_VALUE);
		}

		// Not much else we can do, but create some generic entries. The user
		// will probably have to manually edit them in order to get them
		// working
		else
		{
			odbcProperties = (HODBCINSTPROPERTY)malloc(sizeof(ODBCINSTPROPERTY));
			odbcPointer = odbcProperties;
			memset(odbcPointer, 0, sizeof(ODBCINSTPROPERTY));
			odbcPointer->nPromptType = ODBCINST_PROMPTTYPE_TEXTEDIT;
			odbcPointer->bRefresh = 0;
			odbcPointer->hDLL = NULL;
			odbcPointer->pWidget = NULL;
			odbcPointer->pszHelp = strdup(tr("<b>Data Source Name</b><br><br>The name of the data source.").toAscii().constData());
			odbcPointer->aPromptData = NULL;
			strncpy(odbcPointer->szName, "Name", INI_MAX_PROPERTY_NAME);
			strncpy(odbcPointer->szValue, "", INI_MAX_PROPERTY_VALUE);

			odbcPointer->pNext = (HODBCINSTPROPERTY)malloc(sizeof(ODBCINSTPROPERTY));
			odbcPointer = odbcPointer->pNext;
			memset(odbcPointer, 0, sizeof(ODBCINSTPROPERTY));
			odbcPointer->nPromptType = ODBCINST_PROMPTTYPE_TEXTEDIT;
			odbcPointer->bRefresh = 0;
			odbcPointer->hDLL = NULL;
			odbcPointer->pWidget = NULL;
			odbcPointer->pszHelp = strdup(tr("<b>Data Source Description</b><br><br>A short description of the data source.").toAscii().constData());
			odbcPointer->aPromptData = NULL;
			strncpy(odbcPointer->szName, "Description", INI_MAX_PROPERTY_NAME);
			strncpy(odbcPointer->szValue, "", INI_MAX_PROPERTY_VALUE);

			odbcPointer->pNext = (HODBCINSTPROPERTY)malloc(sizeof(ODBCINSTPROPERTY));
			odbcPointer = odbcPointer->pNext;
			memset(odbcPointer, 0, sizeof(ODBCINSTPROPERTY));
			odbcPointer->nPromptType = ODBCINST_PROMPTTYPE_LABEL;
			odbcPointer->bRefresh = 0;
			odbcPointer->hDLL = NULL;
			odbcPointer->pWidget = NULL;
			odbcPointer->pszHelp = strdup(tr("<b>Data Source ODBC Driver</b><br><br>The ODBC driver used by the data source.").toAscii().constData());
			odbcPointer->aPromptData = NULL;
			strncpy(odbcPointer->szName, "Driver", INI_MAX_PROPERTY_NAME);
			strncpy(odbcPointer->szValue, "", INI_MAX_PROPERTY_VALUE);

			odbcPointer->pNext = (HODBCINSTPROPERTY)malloc(sizeof(ODBCINSTPROPERTY));
			odbcPointer = odbcPointer->pNext;
			memset(odbcPointer, 0, sizeof(ODBCINSTPROPERTY));
			odbcPointer->nPromptType = ODBCINST_PROMPTTYPE_TEXTEDIT;
			odbcPointer->bRefresh = 0;
			odbcPointer->hDLL = NULL;
			odbcPointer->pWidget = NULL;
			odbcPointer->pszHelp = strdup(tr("<b>Database Server</b><br><br>The database server name or IP address.").toAscii().constData());
			odbcPointer->aPromptData = NULL;
			strncpy(odbcPointer->szName, "Server", INI_MAX_PROPERTY_NAME);
			strncpy(odbcPointer->szValue, "", INI_MAX_PROPERTY_VALUE);

			odbcPointer->pNext = (HODBCINSTPROPERTY)malloc(sizeof(ODBCINSTPROPERTY));
			odbcPointer = odbcPointer->pNext;
			memset(odbcPointer, 0, sizeof(ODBCINSTPROPERTY));
			odbcPointer->nPromptType = ODBCINST_PROMPTTYPE_TEXTEDIT;
			odbcPointer->bRefresh = 0;
			odbcPointer->hDLL = NULL;
			odbcPointer->pWidget = NULL;
			odbcPointer->pszHelp = strdup(tr("<b>Database User</b><br><br>The database user for the data source to use.").toAscii().constData());
			odbcPointer->aPromptData = NULL;
			strncpy(odbcPointer->szName, "Username", INI_MAX_PROPERTY_NAME);
			strncpy(odbcPointer->szValue, "", INI_MAX_PROPERTY_VALUE);

			odbcPointer->pNext = (HODBCINSTPROPERTY)malloc(sizeof(ODBCINSTPROPERTY));
			odbcPointer = odbcPointer->pNext;
			memset(odbcPointer, 0, sizeof(ODBCINSTPROPERTY));
			odbcPointer->nPromptType = ODBCINST_PROMPTTYPE_TEXTEDIT;
			odbcPointer->bRefresh = 0;
			odbcPointer->hDLL = NULL;
			odbcPointer->pWidget = NULL;
			odbcPointer->pszHelp = strdup(tr("<b>Database User Password</b><br><br>The database user password for the data source to use.").toAscii().constData());
			odbcPointer->aPromptData = NULL;
			strncpy(odbcPointer->szName, "Password", INI_MAX_PROPERTY_NAME);
			strncpy(odbcPointer->szValue, "", INI_MAX_PROPERTY_VALUE);
			odbcPointer->pNext = NULL;
		}
	}

	// Get additional properties...
	if ((odbcType != DSNDialog::File) && (dialogMode != DSNDialog::Add))
	{
		// Get a full list of configured properties...
		getResult = SQLGetPrivateProfileString(dsnName.toLatin1(), 0, 0, odbcResults, ODBC_BUFFER_SIZE, "odbcinst");

		if (getResult > 0)
		{
			// Get a pointer to the last set of properties...
			for (odbcPointer = odbcProperties; odbcPointer != NULL; odbcPointer = odbcPointer->pNext)
				lastDataSourcePointer = odbcPointer;

			// Check to see it property exists...
			odbcName = QString(odbcResults);
			while (!odbcName.isEmpty())
			{
				propertyExists = FALSE;
				odbcTemp = odbcName.toUpper();
				for (odbcPointer = odbcProperties; ((odbcPointer != NULL) && (propertyExists == FALSE)); odbcPointer = odbcPointer->pNext)
				{
					odbcTemp2 = odbcPointer->szName;
					if (odbcTemp.compare(odbcTemp2.toUpper()) == 0)
						propertyExists = TRUE;
				}

				// Add property...
				if ((propertyExists == FALSE) && (odbcName.compare("UsageCount") != 0))
				{
					lastDataSourcePointer->pNext = (HODBCINSTPROPERTY)malloc(sizeof(ODBCINSTPROPERTY));
					lastDataSourcePointer = lastDataSourcePointer->pNext;
					memset(lastDataSourcePointer, 0, sizeof(ODBCINSTPROPERTY));
					lastDataSourcePointer->nPromptType = ODBCINST_PROMPTTYPE_TEXTEDIT;
					lastDataSourcePointer->pNext = NULL;
					lastDataSourcePointer->bRefresh = 0;
					lastDataSourcePointer->hDLL = NULL;
					lastDataSourcePointer->pWidget = NULL;
					lastDataSourcePointer->pszHelp = strdup(tr("No help for this option, read the driver documentation").toAscii().constData());
					lastDataSourcePointer->aPromptData = NULL;
					strncpy(lastDataSourcePointer->szName, odbcName.toLatin1(), INI_MAX_PROPERTY_NAME);
					strcpy(lastDataSourcePointer->szValue, "");
				}

				// Get the next property name...
				odbcResultsPos = odbcResultsPos + odbcName.length() + 1;
				if (odbcResultsPos < ODBC_BUFFER_SIZE)
				{
					odbcName = QString(odbcResults + odbcResultsPos);
				}
				else
				{
					odbcName = "";
				}
			}
		}
	}

	if (odbcType != DSNDialog::File)
		SQLSetConfigMode(ODBC_BOTH_DSN);
}


// Save the Properties...
bool DSNDialog::saveProperties()
{
	// Declare vaviables...
	bool noError = TRUE;
	QString saveName;
	QString fileDSN;

	// Set config mode...
	if (odbcType == DSNDialog::User)
		SQLSetConfigMode(ODBC_USER_DSN);
	else if (odbcType == DSNDialog::System)
		SQLSetConfigMode(ODBC_SYSTEM_DSN);

	if (odbcType != DSNDialog::File)										// User / System DSN...
	{
		saveName = ((QLineEdit *)odbcProperties->pWidget)->text();
		if (SQLWritePrivateProfileString(saveName.toLatin1(), NULL, NULL, "odbc.ini") == FALSE)
			noError = FALSE;
	}
	else																	// File DSN...
	{
		fileDSN = QString("%1%2").arg(dsnDir, ((QLineEdit *)odbcProperties->pWidget)->text());
		if (SQLWriteFileDSN(fileDSN.toLatin1(), "ODBC", NULL, NULL) == FALSE)
			noError = FALSE;
	}

	// Save all DSN settings...
	if (noError == TRUE)
	{
		for (odbcPointer = odbcProperties->pNext; odbcPointer != NULL; odbcPointer = odbcPointer->pNext)
		{
			switch (odbcPointer->nPromptType)
			{
				case ODBCINST_PROMPTTYPE_LABEL:
				case ODBCINST_PROMPTTYPE_TEXTEDIT:
				case ODBCINST_PROMPTTYPE_TEXTEDIT_PASSWORD:
					if (odbcType != DSNDialog::File)
						SQLWritePrivateProfileString(saveName.toLatin1(), odbcPointer->szName, ((KLineEdit *)odbcPointer->pWidget)->text().toLatin1(), "odbc.ini");
					else
						SQLWriteFileDSN(fileDSN.toLatin1(), "ODBC", odbcPointer->szName, ((KLineEdit *)odbcPointer->pWidget)->text().toLatin1());
					break;

				case ODBCINST_PROMPTTYPE_FILENAME:
					if (odbcType != DSNDialog::File)
						SQLWritePrivateProfileString(saveName.toLatin1(), odbcPointer->szName, ((KUrlRequester *)odbcPointer->pWidget)->url().path().toLatin1(), "odbc.ini");
					else
						SQLWriteFileDSN(fileDSN.toLatin1(), "ODBC", odbcPointer->szName, ((KUrlRequester *)odbcPointer->pWidget)->url().path().toLatin1());
					break;

				case ODBCINST_PROMPTTYPE_LISTBOX:
				case ODBCINST_PROMPTTYPE_COMBOBOX:
					if (odbcType != DSNDialog::File)
						SQLWritePrivateProfileString(saveName.toLatin1(), odbcPointer->szName, ((KComboBox *)odbcPointer->pWidget)->currentText().toLatin1(), "odbc.ini");
					else
						SQLWriteFileDSN(fileDSN.toLatin1(), "ODBC", odbcPointer->szName, ((KComboBox *)odbcPointer->pWidget)->currentText().toLatin1());
					break;
			}
		}
	}

	// Set config mode...
	if (odbcType != DSNDialog::File)
		SQLSetConfigMode(ODBC_BOTH_DSN);

	return noError;
}


// ----------------------------------------------------------------------------------------
// Construct Interface from ODBC properties...

// Build UI from ODBC structure...
void DSNDialog::buildInterface()
{
	// Variables...
	QWidget *widgetPointer;
	QGridLayout *gridLayout;
	KLineEdit *lineEdit;
	KUrlRequester *fileWidget;
	KComboBox *comboBox;
	QLabel *label;
	QSpacerItem *spacer;
	QString settingName;
	char dataSourceResults[ODBC_BUFFER_SIZE + 1];
	
	// Init...
	QWidget *basicWidget = 0;
	QWidget *advancedWidget = 0;
	QWidget *loggingWidget = 0;
	int basicLoop = 0;
	int advancedLoop = 0;
	int loggingLoop = 0;
	int loop = 0;

	// Delete old Widgets/Pages...
	if (basicPage != 0)
	{
		removePage(basicPage);
		basicPage->deleteLater();
		basicPage = 0;
	}
	if (advancedPage != 0)
	{
		removePage(advancedPage);
		advancedPage->deleteLater();
		advancedPage = 0;
	}
	if (loggingPage != 0)
	{
		removePage(loggingPage);
		loggingPage->deleteLater();
		loggingPage = 0;
	}

	// Create new Widgets
	basicWidget = new QWidget();
	advancedWidget = new QWidget();
	loggingWidget = new QWidget();

	// Add layouts...
	gridLayout = new QGridLayout(basicWidget);
	gridLayout = new QGridLayout(advancedWidget);
	gridLayout = new QGridLayout(loggingWidget);

	// Add widgets to UI...
	for (odbcPointer = odbcProperties; odbcPointer != NULL; odbcPointer = odbcPointer->pNext)
	{
		// Place property on correct tab...
		settingName = odbcPointer->szName;

		if (settingName.toUpper() == "TRACE")			// Logging options
		{
			settingName = "Trace:";
			widgetPointer = loggingWidget;
			loop = loggingLoop;
		}

		else if (settingName.toUpper() == "TRACEFILE")
		{
			settingName = "Trace File:";
			widgetPointer = loggingWidget;
			loop = loggingLoop;
		}
														// Settings
		else if (settingName.toUpper() == "NAME" || settingName.toUpper() == "DESCRIPTION" || settingName.toUpper() == "DRIVER" || settingName.toUpper() == "SETUP" || settingName.toUpper() == "DIRECTORY")
		{
			settingName.append(":");
			widgetPointer = basicWidget;
			loop = basicLoop;
		}

		else if (settingName.toUpper() == "SERVER" || settingName.toUpper() == "SERVERNAME" || settingName.toUpper() == "HOST")
		{
			settingName = "Server:";
			widgetPointer = basicWidget;
			loop = basicLoop;
		}

		else if (settingName.toUpper() == "DB" || settingName.toUpper() == "DATABASE")
		{
			settingName = "Database:";
			widgetPointer = basicWidget;
			loop = basicLoop;
		}

		else if (settingName.toUpper() == "USER" || settingName.toUpper() == "UID" || settingName.toUpper() == "USERNAME")
		{
			settingName = "Username:";
			widgetPointer = basicWidget;
			loop = basicLoop;
		}

		else if (settingName.toUpper() == "PASS" || settingName.toUpper() == "PWD" || settingName.toUpper() == "PASSWORD")
		{
			settingName = "Password:";
			widgetPointer = basicWidget;
			loop = basicLoop;
		}
													// Advanced settings for everything else
		else
		{
			if (settingName.toUpper() == "CPTIMEOUT")
				settingName = "CP Timeout:";
			else if (settingName.toUpper() == "CPREUSE")
				settingName = "CP Reuse:";
			else if (settingName.toUpper() == "DRIVER64")
				settingName = "Driver 64:";
			else if (settingName.toUpper() == "SETUP64")
				settingName = "Setup 64:";
			else
				settingName.append(":");
			widgetPointer = advancedWidget;
			loop = advancedLoop;
		}

		// If in view / edit mode, get configured settings. Else use driver default...
		if (dialogMode != DSNDialog::Add)
		{
			if (basicLoop != 0)
			{
				if (odbcType != DSNDialog::File)  	// if it is a user or system dsn...
					SQLGetPrivateProfileString(dsnName.toLatin1(), odbcPointer->szName, "", dataSourceResults, ODBC_BUFFER_SIZE, 0);
				else								// if it is a file dsn
					SQLReadFileDSN(dsnName.toLatin1(), "ODBC", odbcPointer->szName, dataSourceResults, ODBC_BUFFER_SIZE, 0);
			}
			else
				strncpy(dataSourceResults, dsnName.toLatin1(), ODBC_BUFFER_SIZE);
		}
		else
			strncpy(dataSourceResults, odbcPointer->szValue, ODBC_BUFFER_SIZE);


		// Layout label Widgets...
		if (odbcPointer->nPromptType != ODBCINST_PROMPTTYPE_HIDDEN)
		{
			label = new QLabel(widgetPointer);
			label->setText(tr(settingName.toAscii()));
			label->setWhatsThis(tr(odbcPointer->pszHelp));
			((QGridLayout *)widgetPointer->layout())->addWidget(label, loop, 0);
		}

		// Layout Input Widgets...
		switch (odbcPointer->nPromptType)
		{
			case ODBCINST_PROMPTTYPE_TEXTEDIT:
				lineEdit = new KLineEdit(widgetPointer);
				lineEdit->setText(dataSourceResults);
				if (dialogMode == DSNDialog::View)
				{
					lineEdit->setClearButtonShown(FALSE);
					lineEdit->setReadOnly(TRUE);
				}
				else
				{
					lineEdit->setReadOnly(FALSE);
					lineEdit->setClearButtonShown(TRUE);
				}
				lineEdit->setFocus();
				lineEdit->setWhatsThis(tr(odbcPointer->pszHelp));
				label->setBuddy(lineEdit);
				((QGridLayout *)widgetPointer->layout())->addWidget(lineEdit, loop, 1);
				odbcPointer->pWidget = lineEdit;
				break;

			case ODBCINST_PROMPTTYPE_LISTBOX:
			case ODBCINST_PROMPTTYPE_COMBOBOX:
				if (dialogMode == DSNDialog::View)
				{
					lineEdit = new KLineEdit(widgetPointer);
					lineEdit->setText(dataSourceResults);
					lineEdit->setReadOnly(TRUE);
					lineEdit->setClearButtonShown(FALSE);
					lineEdit->setWhatsThis(tr(odbcPointer->pszHelp));
					label->setBuddy(lineEdit);
					((QGridLayout *)widgetPointer->layout())->addWidget(lineEdit, loop, 1);
					odbcPointer->pWidget = lineEdit;
				}
				else
				{
					comboBox = new KComboBox(widgetPointer);
					int stringLoop = 0;
					QStringList tempList;
					while ((const char*)odbcPointer->aPromptData[stringLoop] != 0)
					{
						tempList << odbcPointer->aPromptData[stringLoop];
						stringLoop++;
					}
					comboBox->addItems(tempList);
					comboBox->setEditText(dataSourceResults);
					comboBox->setEditable(TRUE);
					comboBox->setWhatsThis(tr(odbcPointer->pszHelp));
					label->setBuddy(comboBox);
					((QGridLayout *)widgetPointer->layout())->addWidget(comboBox, loop, 1);
					odbcPointer->pWidget = comboBox;
				}
				break;

			case ODBCINST_PROMPTTYPE_FILENAME:
				fileWidget = new KUrlRequester(widgetPointer);
				fileWidget->url().fromPath(dataSourceResults);
				if (dialogMode == DSNDialog::View)
					fileWidget->setEnabled(FALSE);
				else
					fileWidget->setEnabled(TRUE);
				fileWidget->setWhatsThis(tr(odbcPointer->pszHelp));
				label->setBuddy(fileWidget);
				((QGridLayout *)widgetPointer->layout())->addWidget(fileWidget, loop, 1);
				odbcPointer->pWidget = fileWidget;
				break;

			case ODBCINST_PROMPTTYPE_TEXTEDIT_PASSWORD:
				lineEdit = new KLineEdit(widgetPointer);
				lineEdit->setText(dataSourceResults);
				if (dialogMode == DSNDialog::View)
				{
					lineEdit->setClearButtonShown(FALSE);
					lineEdit->setReadOnly(TRUE);
				}
				else
				{
					lineEdit->setReadOnly(FALSE);
					lineEdit->setClearButtonShown(TRUE);
				}
				lineEdit->setEchoMode(QLineEdit::Password);
				lineEdit->setWhatsThis(tr(odbcPointer->pszHelp));
				label->setBuddy(lineEdit);
				((QGridLayout *)widgetPointer->layout())->addWidget(lineEdit, loop, 1);
				odbcPointer->pWidget = lineEdit;
				break;

			default: //ODBCINST_PROMPTTYPE_LABEL:
				lineEdit = new KLineEdit(widgetPointer);
				lineEdit->setReadOnly(TRUE);
				lineEdit->setClearButtonShown(FALSE);
				lineEdit->setText(dataSourceResults);
				lineEdit->setWhatsThis(tr(odbcPointer->pszHelp));
				label->setBuddy(lineEdit);
				((QGridLayout *)widgetPointer->layout())->addWidget(lineEdit, loop, 1);
				odbcPointer->pWidget = lineEdit;
				break;
		}

		// Increment loop counters...
		if (widgetPointer == basicWidget)
			basicLoop++;
		else if (widgetPointer == advancedWidget)
			advancedLoop++;
		else
			loggingLoop++;
	}

	// Add Spacers...
	spacer = new QSpacerItem(0, 0, QSizePolicy::Fixed, QSizePolicy::Expanding);
	((QGridLayout *)basicWidget->layout())->addItem(spacer, basicLoop, 0, 1, 2);
	if (advancedLoop > 0)
	{
		spacer = new QSpacerItem(0, 0, QSizePolicy::Fixed, QSizePolicy::Expanding);
		((QGridLayout *)advancedWidget->layout())->addItem(spacer, advancedLoop, 0, 1, 2);
	}
	if (loggingLoop > 0)
	{
		spacer = new QSpacerItem(0, 0, QSizePolicy::Fixed, QSizePolicy::Expanding);
		((QGridLayout *)loggingWidget->layout())->addItem(spacer, loggingLoop, 0, 1, 2);
	}

	// Data Source Basic Page...
	basicPage = new KPageWidgetItem(basicWidget, tr("Data source configuration settings"));
	addPage(basicPage);

	// Data Source Advanced Page...
	if (advancedLoop > 0)
	{
		advancedPage = new KPageWidgetItem(advancedWidget, tr("Advanced data source settings"));
		addPage(advancedPage);
	}

	// Data Source Logging Page...
	if (loggingLoop > 0)
	{
		loggingPage = new KPageWidgetItem(loggingWidget, tr("Data source logging settings"));
		addPage(loggingPage);
	}
}


// ----------------------------------------------------------------------------------------
// Misc Slots...

// The DSN type has chenged
void DSNDialog::dsnSelectionChanged(bool state)
{
	if (state == TRUE)
	{
		if (QObject::sender() == fileDSNRadio)
			odbcType = DSNDialog::File;
		else if (QObject::sender() == systemDSNRadio)
			odbcType = DSNDialog::System;
		else
			odbcType = DSNDialog::User;
	}
}

