/*
    This file is part of KDE.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "view.h"

#include "engine.h"
#include "appstuff.h"
#include "appview.h"
#include "installersuse.h"
#include "installerrpm.h"
#include "installerdummy.h"
#include "appmanager.h"
#include "prefs.h"
#include "kupdateiface_stub.h"

#include <klistview.h>
#include <klocale.h>
#include <kdebug.h>
#include <kdialog.h>
#include <kmessagebox.h>

#include <qlayout.h>
#include <qtextedit.h>
#include <qsplitter.h>
#include <qdatetime.h>
#include <qlabel.h>
#include <qpushbutton.h>

class AppItem : public KListViewItem
{
  public:
    AppItem( KListView *listView, KNS::Entry *entry )
      : KListViewItem( listView ), mEntry( entry )
    {
      setText( 0, mEntry->name() );
      setText( 1, mEntry->version() );
    }
    
    KNS::Entry *entry() const
    {
      return mEntry;
    }

  private:
    KNS::Entry *mEntry;
};

View::View( QWidget *parent )
  : QWidget( parent )
{
  QBoxLayout *topLayout = new QHBoxLayout( this );

  QSplitter *splitter = new QSplitter( this );
  topLayout->addWidget( splitter );

  QWidget *entryList = new QWidget( splitter );
  QBoxLayout *listLayout = new QVBoxLayout( entryList );

  mListView = new KListView( entryList );
  mListView->addColumn( i18n("Application") );
  mListView->addColumn( i18n("Version") );
  connect( mListView, SIGNAL( selectionChanged( QListViewItem * ) ),
    SLOT( showEntry( QListViewItem * ) ) );
  listLayout->addWidget( mListView );
  
  QBoxLayout *listButtonLayout = new QHBoxLayout( listLayout );
  listButtonLayout->setMargin( KDialog::marginHint() );
  listButtonLayout->setSpacing( KDialog::spacingHint() );
  
  mLastUpdateLabel = new QLabel( entryList );
  listButtonLayout->addWidget( mLastUpdateLabel );
  listButtonLayout->addStretch( 1 );
  
  QPushButton *button = new QPushButton( i18n("Refresh"), entryList );
  listButtonLayout->addWidget( button );
  connect( button, SIGNAL( clicked() ), SLOT( updateListing() ) );

  mAppStuff = new AppStuff();
  mEngine = new Engine( mAppStuff, "apps", this );
  connect( mEngine, SIGNAL( entryRetrieved( KNS::Entry * ) ),
    SLOT( addEntry( KNS::Entry * ) ) );
  connect( mEngine, SIGNAL( retrievalFinished() ),
    SLOT( slotRetrievalFinished() ) );
  connect( mEngine, SIGNAL( retrievalFailed( const QString & ) ),
    SLOT( slotRetrievalError( const QString & ) ) );

  if ( Prefs::installer() == Prefs::InstallerSuse ) {
    mInstaller = new InstallerSuse( mEngine, this );
  } else if ( Prefs::installer() == Prefs::InstallerRpm ) {
    mInstaller = new InstallerRpm( mEngine, this );
  } else {
    mInstaller = new InstallerDummy( mEngine, this );
  }
  connect( mInstaller, SIGNAL( installationFinished() ),
    SLOT( slotInstallationFinished() ) );
  connect( mInstaller, SIGNAL( installationFailed( const QString & ) ),
    SLOT( slotInstallationFailed( const QString & ) ) );

  mAppManager = new AppManager( mEngine, mInstaller );

  mAppView = new AppView( splitter, mAppManager );

  connect( mAppView, SIGNAL( install( KNS::Entry * ) ),
    mInstaller, SLOT( install( KNS::Entry * ) ) );

  updateListing();

  connect( &mCheckTimer, SIGNAL( timeout() ), SLOT( periodicCheck() ) );
  mCheckTimer.start( Prefs::checkInterval() * 60 * 1000 );
}

View::~View()
{
  delete mAppManager;
  delete mInstaller;
  delete mEngine;
  delete mAppStuff;
}

void View::periodicCheck()
{
  kdDebug() << "View::periodicCheck()" << endl;

  updateListing();
}

void View::updateListing()
{
  mListView->clear();
  mEngine->retrieveListing();
  mLastUpdateLabel->setText( i18n("Retrieving list...") );
}

void View::slotRetrievalFinished()
{
  updateLastUpdateLabel();

  KUpdateIface_stub kupdate( "*", "KUpdate" );
  kupdate.checkForUpdate();
}

void View::updateLastUpdateLabel()
{
  Prefs::setLastCheck( QDateTime::currentDateTime() );

  mLastUpdateLabel->setText( i18n("Last Update: %1")
    .arg( KGlobal::locale()->formatDateTime( Prefs::lastCheck() ) ) );
}

void View::slotRetrievalError( const QString & )
{
  mLastUpdateLabel->setText( i18n("Retrieval failed.") );
}

void View::addEntry( KNS::Entry *entry )
{
  AppItem *item = new AppItem( mListView, entry );
  if ( !mListView->selectedItem() ) {
    mListView->setSelected( item, true );
  }
}

void View::showEntry( QListViewItem *i )
{
  AppItem *item = static_cast<AppItem *>( i );
  if ( !item ) return;
  
  mAppView->showEntry( item->entry() );  
}

void View::slotInstallationFinished()
{
  mAppView->setStatusMessage( i18n("Installation finished.") );
  kdDebug() << "Installation finished." << endl;
}

void View::slotInstallationFailed( const QString &msg )
{
  mAppView->setStatusMessage( i18n("Installation failed.") );
  KMessageBox::sorry( this, msg,
    i18n("Installation failed.") );
}

#include "view.moc"
