/*
 *   Copyright © 2009 Dmitry Suzdalev (dimsuz@gmail.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include "historyfiletest.h"
#include "historyfile.h"

#include "clipboardtypes.h"
#include "testutils.h"

#include <QFile>
#include <QUuid>

#include <KConfigGroup>

void tst_HistoryFile::initTestCase()
{
    removeTempFile();
}

void tst_HistoryFile::cleanupTestCase()
{
    removeTempFile();
}

void tst_HistoryFile::createFileTest()
{
    {
        // opening not existing file
        HistoryFile historyFile(tempFileName);
        QVERIFY(QFile::exists(tempFileName) == false);
        QVERIFY(historyFile.entryCount() == 0);
    }

    removeTempFile();

    {
        HistoryFile historyFile;
        // check that loading non existing file in OpenExistingMode fails
        bool res = historyFile.setFile(tempFileName, HistoryFile::OpenExistingMode);
        QVERIFY(res == false);
        QVERIFY(historyFile.entryCount() == 0);
        QVERIFY(QFile::exists(tempFileName) == false);
    }

    removeTempFile();

    {
        // check that setting empty file name fails
        HistoryFile file;
        bool res = file.setFile(QString(), HistoryFile::CreateIfAbsentMode);
        QVERIFY(res == false);
    }

    removeTempFile();

    {
        // check that setting new file cleans older entries

        ClipboardEntry entry;
        entry.data = "TestData";
        entry.uuid = "uuid";

        HistoryFile file("/tmp/klipperhistorytesttemp");
        file.addEntry(entry);
        QVERIFY(file.entryCount() == 1);
        file.setFile(tempFileName, HistoryFile::CreateIfAbsentMode);
        QVERIFY(file.entryCount() == 0);
        QFile::remove("/tmp/klipperhistorytesttemp");
    }

    removeTempFile();

    {
        // check that trying to save to location with no permissions fails
        HistoryFile file("/root/configfile");
        ClipboardEntry entry;
        bool res = file.addEntry(entry);
        QVERIFY(res == false);
        QVERIFY(file.entryCount() == 0);
        QVERIFY(QFile::exists("/root/configfile") == false);
        QFile::remove("/root/configfile");// just in case it got created :)

        res = file.setFile("/root/configfile1", HistoryFile::CreateIfAbsentMode);
        QVERIFY(res == false);
        res = file.addEntry(entry);
        QVERIFY(res == false);
        QVERIFY(file.entryCount() == 0);
        QVERIFY(QFile::exists("/root/configfile1") == false);
        QFile::remove("/root/configfile1");// just in case it got created :)
    }
}

void tst_HistoryFile::addEntryTest()
{
    {
        // adding with no file set should fail
        ClipboardEntry entry;
        HistoryFile historyFile;
        bool res = historyFile.addEntry(entry);
        QVERIFY(res == false);
        QVERIFY(historyFile.entryCount() == 0);
    }

    {
        // test adding a single entry
        ClipboardEntry entry;
        entry.data = "TestData";
        entry.uuid = "testuuid";
        entry.isText = true;
        entry.isSelection = true;

        HistoryFile historyFile(tempFileName);
        bool res = historyFile.addEntry(entry);
        QVERIFY(res == true);
        QCOMPARE(historyFile.entryCount(), 1);
        ClipboardEntry resEntry = historyFile.entry(0);
        QCOMPARE(resEntry.data.toString(), entry.data.toString());
        QCOMPARE(resEntry.uuid, entry.uuid);
        QCOMPARE(resEntry.isText, entry.isText);
        QCOMPARE(resEntry.isSelection, entry.isSelection);

    }

    removeTempFile();

    {
        // check that adding an entry creates file
        ClipboardEntry entry;
        entry.data = "Test";
        entry.uuid = "uuid";
        HistoryFile historyFile(tempFileName);
        bool res = historyFile.addEntry(entry);
        QVERIFY(res == true);
        QVERIFY(QFile::exists(tempFileName) == true);
    }

    removeTempFile();

    {
        // check that adding entry with an empty uuid fails
        ClipboardEntry entry;
        HistoryFile historyFile(tempFileName);
        bool res = historyFile.addEntry(entry);
        QVERIFY(res == false);
    }
}

void tst_HistoryFile::addEntriesTest()
{
    {
        // adding with no file set should fail
        ClipboardEntry entry;
        HistoryFile historyFile;
        ClipboardEntryList list = makeEntryList(10);
        bool res = historyFile.addEntries(list);
        QVERIFY(res == false);
        QVERIFY(historyFile.entryCount() == 0);
    }

    removeTempFile();

    {
        // adding a list
        ClipboardEntryList list = makeEntryList(10);

        HistoryFile file(tempFileName);
        bool res = file.addEntries(list);
        QVERIFY(res == true);
        QCOMPARE(file.entryCount(), list.count());

        for (int i=0; i<file.entryCount(); ++i)
        {
            ClipboardEntry entry = file.entry(i);
            QCOMPARE(entry.data.toString(), list.at(i).data.toString());
            QCOMPARE(entry.uuid, list.at(i).uuid);
            QCOMPARE(entry.isText, list.at(i).isText);
            QCOMPARE(entry.isSelection, list.at(i).isSelection);
        }
    }

    removeTempFile();

    {
        // appending entries to existing ones
        ClipboardEntryList list = makeEntryList(5);

        HistoryFile file(tempFileName);
        bool res = file.addEntries(list);
        QCOMPARE(file.entryCount(), 5);
        QVERIFY(res == true);

        ClipboardEntryList listNew = makeEntryList(10);
        file.addEntries(listNew);
        QCOMPARE(file.entryCount(), 15);

        for (int i=0; i<file.entryCount(); ++i)
        {
            ClipboardEntryList& listToUse = i < list.count() ? list : listNew;
            int idxShift = i < list.count() ? 0 : -list.count();

            ClipboardEntry entry = file.entry(i);
            QCOMPARE(entry.data.toString(), listToUse.at(i+idxShift).data.toString());
            QCOMPARE(entry.uuid, listToUse.at(i+idxShift).uuid);
            QCOMPARE(entry.isText, listToUse.at(i+idxShift).isText);
            QCOMPARE(entry.isSelection, listToUse.at(i+idxShift).isSelection);
        }
    }
}

void tst_HistoryFile::clearEntriesTest()
{
    {
        // clearing nothing should work and leave us with nothing
        HistoryFile file;
        file.clear();
        QVERIFY(file.entryCount() == 0);

        file.setFile(tempFileName, HistoryFile::CreateIfAbsentMode);
        file.clear();
        QVERIFY(file.entryCount() == 0);
    }

    removeTempFile();

    {
        HistoryFile file(tempFileName);
        ClipboardEntryList list = makeEntryList(10);
        file.addEntries(list);
        QCOMPARE(file.entryCount(), list.count());
        file.clear();
        QCOMPARE(file.entryCount(), 0);
    }

    removeTempFile();

    {
        // clear should clear the file too
        HistoryFile file(tempFileName);
        ClipboardEntryList list = makeEntryList(10);
        file.addEntries(list);
        QCOMPARE(file.entryCount(), list.count());
        file.clear();
        file.close();

        HistoryFile file1(tempFileName);
        QCOMPARE(file1.entryCount(), 0);
    }
}

void tst_HistoryFile::removeEntryTest()
{
    {
        // removing entry on empty file should do nothing and return an error
        HistoryFile file;
        bool res = file.removeEntry("someuuid");
        QVERIFY(res == false);
        QVERIFY(file.entryCount() == 0);
        res = file.removeEntry(0);
        QVERIFY(res == false);
        QVERIFY(file.entryCount() == 0);
    }

    removeTempFile();

    {
        // removing entry with undefined uuid should fail
        HistoryFile file(tempFileName);
        bool res = file.removeEntry(QString());
        QVERIFY(res == false);
        QCOMPARE(file.entryCount(), 0);
    }

    {
        ClipboardEntry entry;
        entry.uuid = "uuid";
        HistoryFile file(tempFileName);
        // deleting by index
        file.addEntry(entry);
        bool res = file.removeEntry(0);
        QVERIFY(res == true);
        QVERIFY(file.entryCount() == 0);

    }

    removeTempFile();

    {
        // removing from middle by index
        ClipboardEntryList list = makeEntryList(3);
        HistoryFile file(tempFileName);
        file.addEntries(list);
        bool res = file.removeEntry(1);
        QVERIFY(res == true);
        QVERIFY(file.entryCount() == 2);

        QCOMPARE(file.entry(0).data, list.at(0).data);
        QCOMPARE(file.entry(0).isSelection, list.at(0).isSelection);
        QCOMPARE(file.entry(0).isText, list.at(0).isText);

        QCOMPARE(file.entry(1).data, list.at(2).data);
        QCOMPARE(file.entry(1).isSelection, list.at(0).isSelection);
        QCOMPARE(file.entry(1).isText, list.at(2).isText);
    }

    removeTempFile();

    {
        // removing from middle by uuid
        ClipboardEntryList list = makeEntryList(3);

        HistoryFile file(tempFileName);
        file.addEntries(list);
        bool res = file.removeEntry(list[1].uuid);
        QVERIFY(res == true);
        QVERIFY(file.entryCount() == 2);

        QCOMPARE(file.entry(0).data, list.at(0).data);
        QCOMPARE(file.entry(0).isSelection, list.at(0).isSelection);
        QCOMPARE(file.entry(0).isText, list.at(0).isText);

        QCOMPARE(file.entry(1).data, list.at(2).data);
        QCOMPARE(file.entry(1).isSelection, list.at(0).isSelection);
        QCOMPARE(file.entry(1).isText, list.at(2).isText);
    }

    removeTempFile();

    {
        // when removing it should be removed from file too.
        // Also removing from middle shouldn't mess up indexes

        HistoryFile file(tempFileName);
        ClipboardEntryList list = makeEntryList(6);
        file.addEntries(list);
        bool res = file.removeEntry(list.at(1).uuid);
        QVERIFY(res);
        res = file.removeEntry(list.at(3).uuid);
        QVERIFY(res);
        res = file.removeEntry(list.at(5).uuid);
        QVERIFY(res);
        file.close();
        // so file now should have content: "0", "2", "4"
        // and list is 0, 1, 2, 3, 4, 5

        HistoryFile file1(tempFileName);
        QCOMPARE(file1.entryCount(), 3);
        QCOMPARE(file1.entry(0).data.toString(), list.at(0).data.toString());
        QCOMPARE(file1.entry(1).data.toString(), list.at(2).data.toString());
        QCOMPARE(file1.entry(3).data.toString(), list.at(4).data.toString());
    }
}

void tst_HistoryFile::fileCloseTest()
{
    {
        ClipboardEntryList list = makeEntryList(10);
        // write some file and try to read from it
        HistoryFile file(tempFileName);
        file.addEntries(list);
        file.close();
        QVERIFY(QFile::exists(tempFileName));
        QVERIFY(file.entryCount() == 0);
        // add entry should fail - no new file name specified
        bool res = file.addEntries(list);
        QVERIFY(res == false);
    }
}

void tst_HistoryFile::fileWritingTest()
{
    removeTempFile();

    {
        ClipboardEntryList list = makeEntryList(10);
        // write some file and try to read from it
        HistoryFile file(tempFileName);
        file.addEntries(list);
        file.close();
        QVERIFY(QFile::exists(tempFileName));

        // using constructor to open
        HistoryFile newFile(tempFileName);
        QCOMPARE(newFile.entryCount(), 10);
        for (int i=0; i<10; ++i)
            QCOMPARE(newFile.entry(i), list.at(i));
        newFile.close();

        // using setFile to open
        newFile.setFile(tempFileName, HistoryFile::OpenExistingMode);
        QCOMPARE(newFile.entryCount(), 10);
        for (int i=0; i<10; ++i)
        {
            QCOMPARE(newFile.entry(i).data, list.at(i).data);
            QCOMPARE(newFile.entry(i).uuid, list.at(i).uuid);
            QCOMPARE(newFile.entry(i).isText, list.at(i).isText);
            QCOMPARE(newFile.entry(i).isSelection, list.at(i).isSelection);
        }
        newFile.close();
    }

    removeTempFile();
}

void tst_HistoryFile::entriesTest()
{
    removeTempFile();
    {
        ClipboardEntryList list = makeEntryList(10);
        // write some file and try to read from it
        HistoryFile file(tempFileName);
        file.addEntries(list);
        ClipboardEntryList entriesList = file.entries();
        QCOMPARE(entriesList.count(), 10);
    }
}

void tst_HistoryFile::isWritableTest()
{
    removeTempFile();
    {
        HistoryFile file;
        QVERIFY(file.isWritable() == false);

        file.setFile(QString(), HistoryFile::CreateIfAbsentMode);
        QVERIFY(file.isWritable() == false);

        file.setFile("/root/somefile", HistoryFile::CreateIfAbsentMode);
        QVERIFY(file.isWritable() == false);

        file.setFile(tempFileName, HistoryFile::CreateIfAbsentMode);
        QVERIFY(file.isWritable() == true);

        file.close();
        QVERIFY(file.isWritable() == false);
    }
}

void tst_HistoryFile::setFileWithMainComponentSet()
{
    removeTempFile();
    {
        // TODO: find a way to ensure that a call to KSharedConfig::openConfig() in
        // HistoryFile::setFile() will produce a config object which has some global settings merged
        // Then comment KConfig::SimpleConfig in setFile(), check that test fails, then return it
        // back and check that it works :)
        // code below can't work in the way i want, investigate
        KConfigGroup gr(KGlobal::config(), "TestGroup");
        gr.writeEntry("SomeEntry", "entry");
        KGlobal::config()->sync();
        HistoryFile file;
        bool res = file.setFile(tempFileName, HistoryFile::CreateIfAbsentMode);
        QVERIFY(res = true);
        QVERIFY(file.entryCount() == 0);
    }
    removeTempFile();
}

QTEST_KDEMAIN_KLIPPER(tst_HistoryFile, GUI)
