/* This file is part of the KDE Project
   Copyright (c) 2009 Daniel Winter <dw@danielwinter.de>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QList>
#include <QVariant>

#include <KDebug>
#include <KPluginFactory>

#include <Nepomuk/ResourceManager>
#include <Nepomuk/Variant>

#include "filesystem.h"
#include "filesystemmanagementservice.h"
#include "kmountpointhandler.h"
#include "mountsignalfilter.h"
#include "nsvo.h"
#include "solidfilesystemhandler.h"


using namespace Soprano;


Nepomuk::FilesystemManagementService::FilesystemManagementService( QObject* parent, const QList<QVariant>& )
    : Service( parent )
{
    Nepomuk::ResourceManager::instance()->initialized();

    AbstractFilesystemHandler* handler;

    m_signalFilter = new MountSignalFilter( this, 3000, 10 );
    connect( m_signalFilter, SIGNAL( filesystemMounted( Filesystem ) ),
         this, SLOT( slotFilesystemMounted( Filesystem ) ) );
    connect( m_signalFilter, SIGNAL( filesystemUnmounted( Filesystem ) ),
         this, SLOT( slotFilesystemUnmounted( Filesystem ) ) );

    // solid is prefered so it gets 10 as priority to remove any delay for its signals
    handler = new SolidFilesystemHandler( this, 10 );
    m_handlerList.append( handler );

    // kmountpointhandler's second argument is the maxpriority it should use when it is "sure"
    // (currently for nfs and smbfss)
    handler = new KMountPointHandler( this, 2000, 5, 10 );
    m_handlerList.append( handler );

    foreach( const AbstractFilesystemHandler* handler, m_handlerList ) {
        connect( handler, SIGNAL( filesystemMounted( Filesystem ) ),
                 m_signalFilter, SLOT( slotFilesystemMounted( Filesystem ) ) );
        connect( handler, SIGNAL( filesystemUnmounted( Filesystem ) ),
                 m_signalFilter, SLOT( slotFilesystemUnmounted( Filesystem ) ) );
    }

    resetFilesystemsInNepomuk();
    initFilesystemList();
    writeAllFilesystemsToNepomuk();

}

void Nepomuk::FilesystemManagementService::initFilesystemList()
{
    foreach( const AbstractFilesystemHandler* handler, m_handlerList ) {
        foreach( const Filesystem& fs, handler->mountedFilesystems() ) {
            addFilesystemToList( fs );
        }
    }

}

// returns true if the filesystem got replaced or added
bool Nepomuk::FilesystemManagementService::addFilesystemToList( const Filesystem &fs )
{
    // check if we already have a filesystem for that mountpoint in our hash
    // if so put the one with the higher priority in it

    QString mountpoint = fs.mountpoint();

    if ( m_filesystemList.contains( mountpoint ) ) {
        if ( m_filesystemList.value( mountpoint ).priority() < fs.priority() ) {
            m_filesystemList[mountpoint] = fs;
            return true;
        } else {
            return false;
        }
    } else {
        m_filesystemList.insert( mountpoint, fs );
        return true;
    }
}

void Nepomuk::FilesystemManagementService::writeAllFilesystemsToNepomuk()
{
    foreach( Filesystem fs, m_filesystemList ) {
        fs.writeToNepomuk();
    }
}


void Nepomuk::FilesystemManagementService::resetFilesystemsInNepomuk() const
{
    QList< Nepomuk::Resource > resources = Nepomuk::ResourceManager::instance()->allResourcesWithProperty( Nepomuk::Vocabulary::NSVO::isMounted() , Variant( true ) );
    foreach( Nepomuk::Resource res, resources) {
        res.setProperty( Nepomuk::Vocabulary::NSVO::isMounted(), Variant( false ) );
    }
}



Nepomuk::FilesystemManagementService::~FilesystemManagementService()
{

}


void Nepomuk::FilesystemManagementService::slotFilesystemMounted( Filesystem fs )
{
    m_filesystemList.insert( fs.mountpoint(), fs );
    fs.writeToNepomuk();
}


void Nepomuk::FilesystemManagementService::slotFilesystemUnmounted( Filesystem fs )
{
    QString mountpoint = fs.mountpoint();
    m_filesystemList.remove( mountpoint );
    fs.writeToNepomuk();
}


NEPOMUK_EXPORT_SERVICE( Nepomuk::FilesystemManagementService, "nepomukfilesystemmanagementservice")

#include "filesystemmanagementservice.moc"
