/* This file is part of the KDE Project
   Copyright (c) 2009 Daniel Winter <dw@danielwinter.de>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QTimer>

#include <kmountpoint.h>

#include "kdebug.h"
#include "kmountpointhandler.h"
#include "filesystem.h"


namespace Nepomuk
{

KMountPointHandler::KMountPointHandler(QObject *parent, int pollingInterval, int defaultPriority, int maxPriority )
        : AbstractFilesystemHandler( parent ), m_pollingInterval( pollingInterval), m_defaultPriority( defaultPriority )
        ,m_maxPriority( maxPriority ), m_valid( false )

{
    // others missing? can we really ignore none (could be bind mounts for example)
    // FIXME: what about Windows and other non Linux systems?

    m_fsTypeIgnoreList << "ramfs" << "proc" << "sysfs" << "rootfs" << "devpts" << "nfsd" << "tmpfs"
            << "none" << "usbfs" << "securityfs" << "rpc_pipefs";

    m_timer = new QTimer( this );

    connect( m_timer, SIGNAL( timeout() ),
            this, SLOT( slotUpdate() ) );
    m_timer->start( m_pollingInterval );
    update( false );
}


QList<Filesystem> KMountPointHandler::mountedFilesystems() const
{
    return m_mountedFilesystems.values();
}


void KMountPointHandler::update( bool emitSignals )
{
    QList< KMountPoint::Ptr > newList;
    QStringList oldList = m_mountedFilesystems.keys();
    KMountPoint::List mplist = KMountPoint::currentMountPoints( KMountPoint::NeedMountOptions );

    foreach( KMountPoint::Ptr point, mplist ) {
        if (!shouldIgnoreMountPoint( point ) ) {
            newList << point;
        }
    }
    while ( !newList.isEmpty() )
    {
        KMountPoint::Ptr mp = newList.takeFirst();
        bool removed = oldList.removeOne( mp->mountPoint() );
        // if mointpoint was not in old list, it is new
        if ( !removed )
        {
            Filesystem fs = filesystemForKMountPoint( mp );
            m_mountedFilesystems.insert( fs.mountpoint(), fs );
            if ( emitSignals ) {
                emit filesystemMounted( fs );
            }
        }
    }
    // allmountpoints still in oldList got unmounted
    foreach( const QString& mp, oldList ){
        Filesystem fs  = m_mountedFilesystems.take( mp );
        fs.setMounted( false, QString() );
        if ( emitSignals )
            emit filesystemUnmounted( fs );
    }
}


Filesystem KMountPointHandler::filesystemForKMountPoint( KMountPoint::Ptr mp ) const
{
    QString mountpoint = mp->mountPoint();
    QStringList options = mp->mountOptions();
    QString mountSource = mp->mountedFrom();
    QString type = mp->mountType();

    // identifer: should be something like nepomuk_kmountpoint://ext3@/dev/sda3
    // TODO:    be smart for loop mounts if source is a file use nepomuk_uri for that file
    //          so even the images could be moved arround and still matched

    QString idString = QString( "%1@%2").arg( type, mountSource );

    QUrl identifer;
    identifer.setScheme( "nepomuk_kmountpoint" );
    identifer.setPath( idString );

    // label: FIXME: translation? better way to create a label?
    QString label = QString("mount source: %1 mount type: %2").arg( mountSource, type );
    bool mounted = true;

    // TODO: find a way to get the size of the filesystem, if that makes sense at all
    qulonglong byteSize = 0;

    int priority = m_defaultPriority;

    // for filesystem types where kmountpoint is the only reporting one
    // report them with maxPriority to avoid any delay in signal filter

    if ( type=="nfs"  || type=="nfs4" || type=="smbfs" ) {
        priority = m_maxPriority;
    }
    Filesystem filesystem( identifer, mountpoint, label, mounted, byteSize, priority );

    return filesystem;
}


bool KMountPointHandler::shouldIgnoreMountPoint( KMountPoint::Ptr mp ) const
{
    QString type = mp->mountType();
    QString mountpoint= mp->mountPoint();

    // we can not ignore root (only if it rootfs)
    if ( mountpoint=="/" && type!="rootfs" )
        return false;

    if ( m_fsTypeIgnoreList.contains( type ) ) {
        return true;
    } else {
        return false;
    }
}


void KMountPointHandler::slotUpdate()
{
    update( true );
}

}


#include "kmountpointhandler.moc"
