/*
Copyright (c) 2010 Oszkar Ambrus <aoszkar@gmail.com>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) version 3, or any
later version accepted by the membership of KDE e.V. (or its
successor approved by the membership of KDE e.V.), which shall
act as a proxy defined in Section 6 of version 3 of the license.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "abstractresourceeditplugin.h"
#include "propertyeditwidget.h"
#include "searchwidget.h"

#include <QWidget>
#include <QVariant>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QAction>

#include <KDebug>
#include <KAction>
#include <KMessageBox>

#include <Nepomuk/Query/QueryServiceClient>
#include <Nepomuk/Query/ComparisonTerm>
#include <Nepomuk/Query/ResourceTerm>
#include <Nepomuk/Query/ResourceTypeTerm>
#include <Nepomuk/Query/OrTerm>
#include <Nepomuk/Query/Result>
#include <Nepomuk/Types/Property>
#include <Nepomuk/Resource>

#include <Soprano/Vocabulary/NAO>

class Nepomuk::AbstractResourceEditPlugin::Private
{
public:
    Resource resource;
    QList<PropertyEditWidget*> propertyEditWidgets;

    bool editingEnabled;
};

Nepomuk::AbstractResourceEditPlugin::AbstractResourceEditPlugin(QWidget *parent) :
        QWidget(parent),
        d(new Private())
{
    setFocusPolicy(Qt::StrongFocus);
}

Nepomuk::AbstractResourceEditPlugin::~AbstractResourceEditPlugin()
{
    delete d;
}

void Nepomuk::AbstractResourceEditPlugin::setResource(const Resource &resource)
{
    d->resource = resource;
    setupActions();
    initGui();
}

Nepomuk::Resource Nepomuk::AbstractResourceEditPlugin::resource() const
{
    return d->resource;
}

bool Nepomuk::AbstractResourceEditPlugin::editingEnabled()
{
    return d->editingEnabled;
}


void Nepomuk::AbstractResourceEditPlugin::setEditingEnabled(bool enable)
{
    d->editingEnabled = enable;

    foreach(PropertyEditWidget* pew, d->propertyEditWidgets) {
        pew->setEditingEnabled(enable);
    }
}


void Nepomuk::AbstractResourceEditPlugin::revertChanges()
{
    setResource(resource());

    foreach(PropertyEditWidget* pew, d->propertyEditWidgets) {
        pew->setValue(resource().property(pew->nepomukProperty().uri()));
    }
}


void Nepomuk::AbstractResourceEditPlugin::saveChanges()
{
    foreach(PropertyEditWidget* pew, d->propertyEditWidgets) {
        if(pew->value().isValid()) {
            kDebug() << "*** Trying to save:" << pew->nepomukProperty().uri().toString() << ", " << pew->value();
            resource().setProperty(pew->nepomukProperty().uri(), pew->value());
        }
        else {
            kDebug() << "*** Not saving invalid variant for property:" << pew->nepomukProperty().uri().toString();
        }

    }
}

void Nepomuk::AbstractResourceEditPlugin::addPropertyEditWidget(PropertyEditWidget *widget)
{
    d->propertyEditWidgets.append(widget);
    connect(widget, SIGNAL(geometryChanged()), this, SIGNAL(geometryChanged()));
}

void Nepomuk::AbstractResourceEditPlugin::setupActions()
{
    actions().clear();

    QAction *showRelated = new QAction(KIcon("edit-find-project"), i18n("Show related resources"), this);
    connect(showRelated, SIGNAL(triggered(bool)), this, SLOT(showRelatedResources()));
    addAction(showRelated);
}

void Nepomuk::AbstractResourceEditPlugin::showRelatedResources()
{
    if(resource().isValid()) {
        Nepomuk::Query::ResourceTerm resourceTerm(resource());
        Nepomuk::Types::Property isRelatedProperty(Soprano::Vocabulary::NAO::isRelated());
        Nepomuk::Query::ComparisonTerm isRelated(isRelatedProperty, resourceTerm);

        Nepomuk::Utils::SearchWidget::searchResource(this, Nepomuk::Query::Query(Nepomuk::Query::OrTerm(isRelated, isRelated.inverted())));
    }
    else {
        KMessageBox::sorry(this, "Resource is not valid", "Invalid resource");
    }
}

#warning implement initGui to cleanup list of PropertyEditWidgets
