/*  This file is part of the KDE semantic clipboard
    Copyright (C) 2008 Tobias Wolf <twolf@access.unizh.ch>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

#include <iostream>

#include <QByteArray>
#include <QMap>
#include <QString>
#include <QStringList>

#include <KAboutData>
#include <KApplication>
#include <KCmdLineArgs>
#include <KUrl>
#include <kzip.h>

const QString prefixes        = QString( "prefixes" );
const QString queries         = QString( "queries" );
const QString lineSeperator   = QString( "\n" );
const QString prefixSeperator = QString( ":=" );
const QString querySeperator  = QString( "###QUERY-SEPERATOR###\n" );

QMap< QString, QString > readPrefixes( const KZip* zipFile )
{
    QMap< QString, QString > result;
    const KArchiveEntry* entry = zipFile->directory()->entry( prefixes );

    if ( entry && entry->isFile() ) {
        const KArchiveFile* file = static_cast< const KArchiveFile* >( entry );
        const QStringList lines = QString::fromUtf8( file->data() ).split( lineSeperator, QString::SkipEmptyParts );

        foreach ( QString line, lines ) {
            const QStringList parts = line.split( prefixSeperator );
            const QString lhs = parts.value( 0 );
            const QString rhs = parts.value( 1 );

            if ( !lhs.isEmpty() && !rhs.isEmpty() ) {
                result.insert( lhs, rhs );
            }
        }
    }

    return result;
}

QStringList readQueries( const KZip* zipFile )
{
    QStringList result;
    const KArchiveEntry* entry = zipFile->directory()->entry( queries );

    if ( entry && entry->isFile() ) {
        const KArchiveFile* file = static_cast< const KArchiveFile* >( entry );
        result = QString::fromUtf8( file->data() ).split( querySeperator, QString::SkipEmptyParts );
    }

    return result;
}

bool writePrefixes( KZip* zipFile, const QMap< QString, QString >& prefixMap )
{
    QByteArray data;
    foreach ( QString key, prefixMap.keys() ) {
        data.append( key.toUtf8() );
        data.append( prefixSeperator.toUtf8() );
        data.append( prefixMap.value( key ).toUtf8() );
        data.append( lineSeperator.toUtf8() );
    }

    return zipFile->writeFile( prefixes, QString(), QString(), data.constData(), data.size() );
}

bool writeQueries( KZip* zipFile, const QStringList& queryList )
{
    QByteArray data;
    foreach ( QString query, queryList ) {
        data.append( query.toUtf8() );
        data.append( querySeperator.toUtf8() );
    }

    return zipFile->writeFile( queries, QString(), QString(), data.constData(), data.size() );
}

int main( int argc, char* argv[] )
{
    KAboutData aboutData( QByteArray( "ZipFile" ), QByteArray(), KLocalizedString(), QByteArray() );
    KCmdLineArgs::init( argc, argv, &aboutData );
    KCmdLineOptions options;
    options.add( "+[file]", KLocalizedString() );
    KCmdLineArgs::addCmdLineOptions( options );
    KApplication app;

    KCmdLineArgs *args = KCmdLineArgs::parsedArgs();
    if ( !args->count() ) {
        return 1;
    }

    KZip input( args->url( 0 ).toLocalFile() );
    input.open( QIODevice::ReadOnly );
    QMap< QString, QString > prefixes = readPrefixes( &input );
    QStringList queries = readQueries( &input );
    input.close();

    foreach ( QString prefix, prefixes.keys() ) {
        std::cout << prefix.toUtf8().constData() << "\t" << prefixes.value( prefix ).toUtf8().constData() << std::endl;
    }
    std::cout << std::endl;

    foreach ( QString query, queries ) {
        std::cout << query.toUtf8().constData() << std::endl;
    }
    std::cout << std::endl;

    KZip output( "zipfile.zip" );
    output.open( QIODevice::WriteOnly );
    writePrefixes( &output, prefixes );
    writeQueries( &output, queries );
    output.close();

    return 0;
}
