/*
 * This file is part of Search Project.
 *
 * Copyright (C) 2007 Daniele Galdi <daniele.galdi@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include <QDebug>

#include <strigi/qtdbus/strigiasyncclient.h>

#include "StrigiHitsModel.h"

bool sortByMime( const StrigiHit &h1, const StrigiHit &h2 )
{
  return h1.mimetype < h2.mimetype;
}

bool sortByUri( const StrigiHit &h1, const StrigiHit &h2 )
{
  return h1.uri < h2.uri;
}

bool sortByDate( const StrigiHit &h1, const StrigiHit &h2 )
{
  return h1.mtime > h2.mtime;
}

struct StrigiHitsModel::Private 
{
  Private()
  {}

  QList<StrigiHit> hits;

  QList<QString> groups;  
  QMap<int, QList<StrigiHit> *> leaf;
};

StrigiHitsModel::StrigiHitsModel( QObject * parent )
  : QAbstractItemModel( parent )
{
  d = new Private;
}
  
StrigiHitsModel::~StrigiHitsModel()
{
  delete d;
}	

QModelIndex StrigiHitsModel::parent( const QModelIndex &child ) const
{
  if ( !child.isValid() )
    return QModelIndex();

  if ( child.internalId() == 2 )
    return createIndex( 0, 0, 1);

  return QModelIndex();
}

QModelIndex StrigiHitsModel::index( int row, int column, const QModelIndex& parent ) const
{
  if ( parent.internalId() == 0 )
    // Group node
    return createIndex(row, column, 1);

  // Leaf node
  return createIndex(row, column, 2);
}

int StrigiHitsModel::columnCount(const QModelIndex& parent) const
{
  return 1;
}

int StrigiHitsModel::rowCount( const QModelIndex& parent ) const
{
  if ( !parent.isValid() )
    return d->groups.size();

  if ( parent.internalId() == 1 )
    return d->leaf.value( parent.row() )->size();

  return 0;
}

QVariant StrigiHitsModel::data( const QModelIndex & index, int role ) const
{
  if ( !index.isValid() )
  {
    return QVariant();
  }
 
  if ( role == Qt::DisplayRole )
  {
    if ( index.internalId() == 1 && index.column() == 0 )
    {
      return d->groups.value( index.row() );
    }
    
    if ( index.internalId() == 2 )
    {
      int parentRow = index.parent().row();

      qDebug() << "########################################" << parentRow;

      if ( index.column() == 0 )
	return d->leaf.value( parentRow )->value( index.row() ).uri;
      
      if ( index.column() == 1 )
	return d->hits.value( index.row() ).mimetype;

      if ( index.column() == 2 )
	return d->hits.value( index.row() ).size;

      if ( index.column() == 3 )
	return d->hits.value( index.row() ).mtime;
    }
  }

  return QVariant();
}

void StrigiHitsModel::setHits( const QList<StrigiHit> &hits )
{
  d->hits = hits;
  d->groups.append( "Unknow" );   // 0
  d->groups.append( "Text/Plain" ); // 1

  d->leaf.insert( 0, new QList<StrigiHit>() );
  d->leaf.insert( 1, new QList<StrigiHit>() );

  QListIterator<StrigiHit> iter( hits );
  while ( iter.hasNext() )
  {
    StrigiHit hit = iter.next();
    if ( hit.mimetype == "text/plain" )
    {
      qDebug() << "append in text/plain";
      d->leaf.value( 1 )->append( hit );
    }
    else 
    {
      qDebug() << "append in unknow";
      d->leaf.value( 0 )->append( hit );
    }
  }

  reset();
}

void StrigiHitsModel::invalidateLayout()
{
  emit layoutChanged();
}

void StrigiHitsModel::groupBy( const QString &group )
{
  qDebug() << "StrigiHitsModel::groupBy" << group;

  if ( group == "kind" ) 
    qSort( d->hits.begin(), d->hits.end(), sortByMime );

  if ( group == "date" ) 
    qSort( d->hits.begin(), d->hits.end(), sortByDate );

  emit layoutChanged();
}

#include "StrigiHitsModel.moc"
