#!/usr/bin/env python

import os
import unittest
import dbus
import sys
import subprocess
import time
from PyQt4 import QtCore
from PyKDE4 import kdecore, nepomuk

# It is never wrong to have a QCoreApplication around
kcomponent = kdecore.KComponentData("nepomuk-test")
qapp = QtCore.QCoreApplication(sys.argv)

# The list of known Nepomuk services. This list may contain deprecated or experimental services.
allNepomukServices = [ 'nepomukremovablestorageservice', 'nepomukfilesystemmanagementservice', 'nepomukmigrationservice', 'nepomukmigration1', 'nepomukusercontextservice', 'nepomukpimoservice', 'nepomukstorage', 'nepomukontologyloader', 'nepomukfileannotationservice', 'nepomukqueryservice', 'digikamnepomukservice', 'nepomukfilewatch', 'nepomukstrigiservice']

class NepomukConfig(object):
    """
    Allows to modify the Nepomuk config
    """

    def __init__(self):
        self.serverConfig = kdecore.KConfig("nepomukserverrc")
        self.strigiConfig = kdecore.KConfig("nepomukstrigirc")
        
    def enableNepomuk(self, enable):
        self.serverConfig.group("Basic Settings").writeEntry("Start Nepomuk", enable)
        self.serverConfig.sync()

    def enableService(self,serviceName,enable):
        self.serverConfig.group("Service-"+serviceName).writeEntry("autostart", enable)
        self.serverConfig.sync()
            
    def enableStrigi(self, enable):
        self.enableService("Service-nepomukstrigiservice", enable)

    def setStrigiIndexDirs(self, dirs):
        self.strigiConfig.group("General").writePathEntry("folders", dirs)
        self.strigiConfig.sync()


class NepomukServer(object):
    def __init__(self):
        self.serverPid = 0
        self.running = False
        
    def start(self):
        "Start the nepomuk server"
        if not self.running:
            print "Starting the nepomukserver..."
            self.process = subprocess.Popen(["nepomukserver", "--nofork"])
            self.running = True
            
    def stop(self):
        "Stop the nepomuk server and wait for it to start"
        if self.running:
            print "Stopping the nepomukserver..."
            self.process.terminate()
            self.process.wait()
            self.running = False
                     

# a global nepomuk server object so the server can properly be shut down at any time
# and is not started more than once
nepomukServer = NepomukServer()


class NepomukServiceTestBase(unittest.TestCase):
    """
    Base class for all Nepomuk service unit tests.
    """
    
    def __init__(self, name='runTest'):
        # super constructor
        unittest.TestCase.__init__(self, name)

        # a few convinience things for subclasses to use
        self.nepomukConfig = NepomukConfig()
        self.testDataDir = os.getenv('NEPOMUK_TEST_DATA_DIR')
        if self.testDataDir is None:
            self.testDataDir = os.getcwd() + '/testdata'

        # empty list -> all services
        self.services = []
        

    def run(self, result=None):
        # setup test case
        self.setUpTestCase()

        # make sure our services are actually started (in case there are other global KDE defaults)
        self.nepomukConfig.enableNepomuk(True)
        for service in allNepomukServices:
            if not service in self.services:
                self.nepomukConfig.enableService(service,False)
        for service in self.services:
            self.nepomukConfig.enableService(service,True)
            
        # start the nepomuk server
        self.startServer()

        # time the tests
        startTime = time.time()

        # make sure the necessary services are up and initialized
        for service in self.services:
            self.assert_(self.waitForService(service, 60))

        # actually run the test case
        unittest.TestCase.run(self, result)

        duration = time.time() - startTime

        # stop the nepomuk server (in case of an exception nepomukUnittestMain jumps in)
        self.stopServer()

        # tear down the test case
        self.tearDownTestCase()

        print "==================================================================="
        print "Ran Nepomuk services test in %f" % duration
        print "==================================================================="


    def setUpTestCase(self):
        "Set up the whole test case. The standard implementation does nothing."
        pass


    def tearDownTestCase(self):
        "Tear down the whole test case. Called after all tests have finished."
        pass
            

    def startServer(self):
        "Start the global Nepomuk server"
        nepomukServer.start()


    def stopServer(self):
        "Stop the global Nepomuk server"
        nepomukServer.stop()
            

    def getTestDataBaseDir(self):
        "Return the base dir of the test data."
        return self.testDataDir


    def setNecessaryServices(self, services):
        "Set the necessary services for the test. By default all services are started."
        self.services = services


    def serviceDBusInterface(self, serviceName, interfaceName):
        "Return the DBus proxy interface to a service"
        serviceObject = dbus.SessionBus().get_object('org.kde.nepomuk.services.' + serviceName, '/' + serviceName)
        return dbus.Interface(serviceObject, interfaceName)


    def waitForService(self, serviceName, timeout = -1):
        "Wait for a Nepomuk service to be initialized. timeout is in seconds."
        startTime = time.time()
        sbus = dbus.SessionBus()
        while True:
            try:
                serviceObject = sbus.get_object('org.kde.nepomuk.services.' + serviceName, '/servicecontrol')
                serviceControlInterface = dbus.Interface(serviceObject,'org.kde.nepomuk.ServiceControl')
                if serviceControlInterface.isInitialized():
                    return True;
                break;
            except dbus.exceptions.DBusException:
                pass
            if timeout >= 0 and time.time()-startTime >= timeout:
                return False
            time.sleep(2)


def nepomukUnittestMain():
    """
    The main test method to be used instead of unittest.main()
    which makes sure the nepomukserver is properly stopped
    even if any error occurs.
    """
    try:
        unittest.main()
    except:
        nepomukServer.stop()
