/*
  Copyright (C) 2010 Michael Leupold <lemma@confuego.org>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "opener.h"

#include <QtCore/QCoreApplication>
#include <QtCore/QTimer>
#include <QtDBus/QDBusConnectionInterface>
#include <QtDBus/QDBusInterface>
#include <QtDBus/QDBusReply>
#include <QtCore/QTextStream>
#include <KToolInvocation>

Opener::Opener()
 : m_stderr(stderr, QIODevice::WriteOnly), m_openLocalWallet(false), m_openNetworkWallet(true),
   m_maxwait(30), m_sessionTimeout(120)
{
}

void Opener::start()
{
   QTimer::singleShot(0, this, SLOT(realStart()));
}

void Opener::realStart()
{
   QDBusConnection sessionBus(QDBusConnection::sessionBus());
   // check if dbus was launched
   if (!sessionBus.isConnected()) {
      m_stderr << i18n("D-Bus session bus not available") << endl;
      QCoreApplication::instance()->exit(4);
      return;
   }

   // check if KWalletD is already available
   if (sessionBus.interface()->isServiceRegistered("org.kde.kwalletd")) {
      if (tryOpen()) {
         QCoreApplication::instance()->exit(0);
         return;
      }
   }

   // check if klauncher is already available
   if (m_startDaemon && sessionBus.interface()->isServiceRegistered("org.kde.klauncher")) {
      startDaemon();
   } else {
      // now, wait for kwalletd and klauncher to appear so we can perform out duty.
      connect(&m_klauncherWatcher, SIGNAL(serviceRegistered(const QString&)),
                                 SLOT(slotKLauncherRegistered(const QString&)));
      m_klauncherWatcher.setWatchMode(QDBusServiceWatcher::WatchForRegistration);
      m_klauncherWatcher.addWatchedService("org.kde.klauncher");
      m_klauncherWatcher.setConnection(sessionBus);
   }

   connect(&m_kwalletdWatcher, SIGNAL(serviceRegistered(const QString&)),
                               SLOT(slotKWalletDRegistered(const QString&)));
   m_kwalletdWatcher.setWatchMode(QDBusServiceWatcher::WatchForRegistration);
   m_kwalletdWatcher.addWatchedService("org.kde.kwalletd");
   m_kwalletdWatcher.setConnection(sessionBus);

   // maximum waiting time for the daemon to appear
   QTimer::singleShot(m_maxwait, this, SLOT(timeout()));
}

Opener::~Opener()
{
}

void Opener::timeout()
{
   QCoreApplication::instance()->exit(4);
}

void Opener::slotKLauncherRegistered(const QString &serviceName)
{
   Q_ASSERT(serviceName == QLatin1String("org.kde.klauncher"));
   startDaemon();
}

void Opener::slotKWalletDRegistered(const QString &serviceName)
{
   Q_ASSERT(serviceName == QLatin1String("org.kde.kwalletd"));
   int ret = 0;
   if (!tryOpen()) {
      ret = 5;
   }
   QCoreApplication::instance()->exit(ret);
}

bool Opener::startDaemon()
{
   QString error;
   int ret = KToolInvocation::startServiceByDesktopPath("kwalletd.desktop", QStringList(),
                                                        &error);
   if (ret > 0) {
      m_stderr << "Couldn't start kwalletd: " << error << endl;
      return false;
   }
   return true;
}

bool Opener::tryOpen()
{
   QDBusInterface iface("org.kde.kwalletd", "/modules/kwalletd", "org.kde.KWallet");
   if (!iface.isValid()) {
      // couldn't open as interface is not available
      return false;
   } else {
      if (m_openLocalWallet) {
         QDBusReply<QString> wallet = iface.call("localWallet");
         QDBusMessage reply = iface.call("pamOpen", wallet.value(), m_hashedPw.toByteArray(), m_sessionTimeout);
         if (reply.type() == QDBusMessage::ErrorMessage) {
            m_stderr << "Error: " << reply.errorMessage() << endl;
         }
      }
      if (m_openNetworkWallet) {
         QDBusReply<QString> wallet = iface.call("networkWallet");
         m_stderr << "Opening " << wallet.value() << endl;
         QDBusMessage reply = iface.call("pamOpen", wallet.value(), m_hashedPw.toByteArray(), m_sessionTimeout);
         if (reply.type() == QDBusMessage::ErrorMessage) {
            m_stderr << "Error: " << reply.errorMessage() << endl;
         }
      }
      return true;
   }
}

#include "opener.moc"
