/***************************************************************************
 *   Copyright © 2009 Reza Fatahilah Shah <rshah0385@kireihana.com>        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/
#include "videogst.h"
#include <KDebug>
#include <gst/gst.h>
#include <gst/base/gstbasesink.h>
#include "photosavebin.h"
#include "avrecordbin.h"
#include "videoeffectbin.h"
#include "videocapturebin.h"
#include "libqtgstreamer/qgstobject.h"
#include "libqtgstreamer/qgstpipeline.h"
#include "libqtgstreamer/qgstpad.h"
#include "libqtgstreamer/qgstbin.h"
#include "libqtgstreamer/qgstelement.h"
#include "libqtgstreamer/qgstelementfactory.h"
#include "libqtgstreamer/qgstcaps.h"

#include "videoinputdevice.h"

static bool bus_message_cb (GstBus *bus, GstMessage *message, VideoGst *videoGst)
{
    if (GST_MESSAGE_TYPE (message) == GST_MESSAGE_EOS)
    {
        if (videoGst->isRecording()) {
            kDebug() << "Received EOS message";
            videoGst->stopVideoRecording(true);
            return true;
        }
    }
}

VideoGst::VideoGst(QObject *parent) 
    : QObject(parent)
{
    mIsRunning = false;
    mIsRecording = false;
    mFrameSize = QSize(320,240);
    setupPipeline();

    /*GstBus *bus = gst_pipeline_get_bus (GST_PIPELINE(mPipeline));
    gst_bus_add_watch (bus, busCall, this);
    gst_object_unref(bus);*/
}

VideoGst::~VideoGst()
{
    //g_source_remove_by_user_data(this);
}

gboolean VideoGst::busCall(GstBus *bus, GstMessage *msg, gpointer data)
{
    Q_UNUSED(bus);
    Q_ASSERT(msg);
    //TODO do something if needed here
    return true;
}


void VideoGst::start()
{
    if (mPipeline) {
        kDebug() << "Starting pipeline";
        mPipeline->setState(QtGstreamer::QGstElement::Playing);
        mIsRunning = true;
    }
}

void VideoGst::stop()
{
    if (mPipeline) {
        kDebug() << "Stopping pipeline";
        mPipeline->setState(QtGstreamer::QGstElement::Null);
        mIsRunning = false;
    }
}

bool VideoGst::isRunning()
{
    return mIsRunning;
}

bool VideoGst::isRecording()
{
    return mIsRecording;
}

void VideoGst::slotScreenshotTaken(QImage *image)
{
    using namespace QtGstreamer;
    kDebug() << "Emit photo taken signal";
    QGstPadPtr photoPad = mPhotoSaveBin->getStaticPad("sink");
    QGstPadPtr sourcePad = mVideoEffectBin->getStaticPad("src1");
    
    if (sourcePad->setBlocked(true)) {
        sourcePad->unlink(photoPad);
        mPipeline->remove(mPhotoSaveBin);
        mPhotoSaveBin->setState(QGstElement::Null);
        sourcePad->setBlocked(false);
    }
    emit photoTaken(image);
}

void VideoGst::setupPipeline()
{
    mPipeline = QtGstreamer::QGstPipeline::newPipeline("main_pipeline");
    mVideoCaptureBin = VideoCaptureBin::createVideoCaptureBin();
    mPipeline->add(mVideoCaptureBin);

    mPhotoSaveBin = PhotoSaveBin::createPhotoSaveBin(mFrameSize);
    mVideoRecordingBin = AVRecordBin::createAVRecordBin();
    
    mVideoEffectBin = VideoEffectBin::createVideoEffectBin();
    mPipeline->add(mVideoEffectBin);

    connect(&(*mPhotoSaveBin),SIGNAL(photoTaken(QImage *)), this, SLOT(slotScreenshotTaken(QImage *)));
}

void VideoGst::linkPipeline()
{
    using namespace QtGstreamer;

    kDebug() << "Link capture to effect bin";
    QGstElement::link(mVideoCaptureBin,mVideoEffectBin);
    kDebug() << "Link effect bin to video sink";
    
    QGstElementPtr queue = QGstElementFactory::make("queue");
    QGstElementPtr csp = QGstElementFactory::make("ffmpegcolorspace");
    mPipeline->add(queue);
    mPipeline->add(csp);
    QGstElement::link(queue,csp,mVideoSinkPtr);

    QGstPadPtr displayPad = queue->getStaticPad("sink");
    QGstPadPtr sourcePad = mVideoEffectBin->getStaticPad("src0");

    sourcePad->link(displayPad);
    kDebug() << "Done linking effect bin";
    /*mPipelineBus = gst_element_get_bus (mPipeline);
    gst_bus_add_signal_watch (mPipelineBus);
    g_signal_connect (G_OBJECT (mPipelineBus), "message",G_CALLBACK (bus_message_cb), this);*/
}

void VideoGst::setVideoSink(const QtGstreamer::QGstElementPtr &videoSink)
{
    this->stop();
    kDebug() << "Set video sink";

    //remove old sink
    if (mVideoSinkPtr) {
        kDebug() << "Removing old video sink";
        mPipeline->remove(mVideoSinkPtr);
    }
    //add new sink
    mVideoSinkPtr = videoSink;
    kDebug() << "Add new video sink";
    mPipeline->add(mVideoSinkPtr);
}

void VideoGst::setVideoSourceProperties(QString &devicePath,QSize frameSize,QSize frameRate)
{
    mDevicePath = devicePath;
    mFrameSize = frameSize;
    mFrameRate = frameRate;
}

void VideoGst::updatePipeline()
{
    this->stop();
    //update capture
    updateVideoSource();
    //update photo capture
    updatePhotoSaveBin();
}

void VideoGst::updateVideoSource()
{
    mVideoCaptureBin->updateCapabilities(mFrameSize,mFrameRate);
}

void VideoGst::updatePhotoSaveBin()
{
    kDebug() << "Trying to update photosink for size=" << mFrameSize;
    mPhotoSaveBin->updatePhotoSize(mFrameSize);
}

void VideoGst::startVideoRecording(QString &fileName)
{
    kDebug() << "Start Video Recording";
    using namespace QtGstreamer;
    QGstPadPtr recordPad = mVideoRecordingBin->getStaticPad("sink");
    QGstPadPtr sourcePad = mVideoEffectBin->getStaticPad("src1");

    if (sourcePad->setBlocked(true)) {
        mPipeline->add(mVideoRecordingBin);
        mVideoRecordingBin->syncStateWithParent();
        sourcePad->link(recordPad);
        mVideoRecordingBin->start(fileName);
        sourcePad->setBlocked(false);
        mVideoCaptureBin->newTimeStamp();
        mIsRecording = true;
    }
}

void VideoGst::stopVideoRecording(bool fromCallback)
{
    if (!mIsRecording) {
        return;
    }
    using namespace QtGstreamer;

    QGstPadPtr recordPad = mVideoRecordingBin->getStaticPad("sink");
    QGstPadPtr sourcePad = mVideoEffectBin->getStaticPad("src1");
    
    if (sourcePad->setBlocked(true)) {
        sourcePad->unlink(recordPad);
        mPipeline->remove(mVideoRecordingBin);
        mVideoRecordingBin->setState(QGstElement::Null);
        sourcePad->setBlocked(false);
        mVideoRecordingBin->stop();
        kDebug() << "Stop video recording";    
        mIsRecording = false;
    }
}

void VideoGst::takeScreenshot()
{
    using namespace QtGstreamer;
    kDebug() << "Link effect bin to photo capture bin";
    QGstPadPtr photoPad = mPhotoSaveBin->getStaticPad("sink");
    QGstPadPtr sourcePad = mVideoEffectBin->getStaticPad("src1");

    if (sourcePad->setBlocked(true)) {
        mPipeline->add(mPhotoSaveBin);
        mPhotoSaveBin->syncStateWithParent();
        sourcePad->link(photoPad);
        mPhotoSaveBin->takePhoto();
        sourcePad->setBlocked(false);
    }
}

#include "videogst.moc"