/*
    Copyright (C) 2010 Ariya Hidayat <ariya@kde.org>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "morphingclock.h"

#include <QGraphicsSceneMouseEvent>
#include <QPainter>
#include <QTimeLine>
 
#include <math.h>

#ifdef M_PI
#define M_PI 3.14159265358979323846
#endif


// see http://ariya.blogspot.com/2010/03/morphing-clock.html
// for details on how morphing clock works

static QPolygonF subdivide(const QPolygonF &p, qreal delta = 8)
{
    QPolygonF poly;
    poly << p.at(0);
    for (int i = 0; i < p.count() - 1; ++i) {
        QPointF d = p.at(i + 1) - p.at(i);
        qreal len = sqrt(d.x() * d.x() + d.y() * d.y());
        poly << p.at(i);
        int split = len / delta;
        for (int c = 1; c < split; ++c)
            poly << p.at(i) + static_cast<qreal>(c) * d / split;
    }
    return poly;
}

static QPolygonF createCircle(qreal radius, int points)
{
    QPolygonF poly;
    for (int i = 0; i < points; ++i) {
        qreal a = i * M_PI * 2 / points;
        poly << QPointF(cos(a), sin(a)) * radius;
    }
    return poly;
}

static QPolygonF createSlab(qreal thickness, qreal length, qreal angle, int points)
{
    QPolygonF poly;

    int cx = points / 2 * thickness / (thickness + length);
    int cy = points / 2 * length / (thickness + length);

    qreal xo = thickness / 2;
    qreal yo = length;

    for (int x = 0; x < cx; ++x)
        poly << QPointF(-xo + x * thickness / cx, length);
    for (int y = 0; y < cy; ++y)
        poly << QPointF(xo, yo - y * length / cy);
    for (int x = 0; x < cx; ++x)
        poly << QPointF(xo - x * thickness / cx, 0);
    for (int y = 0; y < cy; ++y)
        poly << QPointF(-xo, y * length / cy);

    while (poly.count() < points)
        poly << QPointF(-xo, yo);

    poly = QTransform().rotateRadians(M_PI + angle).map(poly);

    return poly;
}

static QPolygonF sort(const QPolygonF &p)
{
    QPolygonF poly;
    int index = 0;
    for (int c = 0; c < p.count(); ++c)
        if (p.at(c).y() > p.at(index).y())
            index = c;
    for (int c = 0; c < p.count(); ++c)
        poly << p.at((index + c) % p.count());
    return poly;
}

template <class T>
static T tween(const T &a, const T &b, qreal t)
{
    return a + t * (b - a);
}

static QPolygonF tween(const QPolygonF &poly1, const QPolygonF &poly2, qreal t)
{
    if (qFuzzyCompare(t, 1 + t))
        return poly1;
    Q_ASSERT(poly1.count() == poly2.count());
    QPolygonF poly;
    poly.reserve(poly1.count());
    for (int i = 0; i < poly1.count(); ++i)
        poly << tween(poly1.at(i), poly2.at(i), t);
    return poly;
}
 
MorphingClock::MorphingClock(QObject *parent, const QVariantList &args)
    : Plasma::Applet(parent, args)
{
    resize(256, 256);
    setAspectRatioMode(Plasma::Square);
    setBackgroundHints(NoBackground);

    targetLevel[0] = 224;
    targetLevel[1] = 64;
    targetLevel[2] = 64;
    targetLevel[3] = 192;

    timeLine = new QTimeLine(503, this);
    timeLine->setCurveShape(QTimeLine::EaseInOutCurve);
    connect(timeLine, SIGNAL(valueChanged(qreal)), this, SLOT(updateTransition()));
    timeLine->setCurrentTime(0);
    timeLine->setUpdateInterval(25);
    timeLine->setDirection(QTimeLine::Backward);

    adjustSize();
    connect(this, SIGNAL(geometryChanged()), SLOT(adjustSize()));
}
 
void MorphingClock::init()
{
    Plasma::DataEngine* timeEngine = dataEngine("time");
    timeEngine->connectSource("Local", this, 6000, Plasma::AlignToMinute);
}

void MorphingClock::adjustSize()
{
    radius = qMin(boundingRect().height(), boundingRect().width()) * 0.45;
    font = QFont("Helvetica", 72);
    qreal fw = QFontMetrics(font).width("4");
    font.setPointSize(72 * (2 * radius / 4.5) / fw);
    digitWidth = QFontMetrics(font).width("4");
}

void MorphingClock::updateTransition()
{
    update();
}
 
void MorphingClock::dataUpdated(const QString& source, const Plasma::DataEngine::Data &data)
{
    Q_UNUSED(source);
    int hour = data["Time"].toTime().hour();
    int minute = data["Time"].toTime().minute();
    QChar text[4];
    text[0] = '0' + (hour / 10);
    text[1] = '0' + (hour % 10);
    text[2] = '0' + (minute / 10);
    text[3] = '0' + (minute % 10);

    for (int c = 0; c < 4; ++c) {
        QPainterPath path;
        path.addText(0, 0, font, text[c]);
        QRectF box = path.boundingRect();
        qreal ofs = (c-2)*(digitWidth);
        ofs += (c < 2) ? -digitWidth/2 : digitWidth/2;
        path.translate(ofs, box.height() / 2);
        sourcePoly[c] = sort(subdivide(path.toFillPolygon()));
    }

    qreal angle1 = hour * M_PI / 6 + minute * M_PI / 360;
    qreal angle2 = minute * M_PI / 30;

    targetPoly[0] = createCircle(radius, sourcePoly[0].count());
    targetPoly[1] = createSlab(radius / 6, 0.5 * radius, angle1, sourcePoly[1].count());
    targetPoly[2] = createSlab(radius / 6, 0.9 * radius, angle2, sourcePoly[2].count());
    targetPoly[3] = createCircle(0.8 * radius, sourcePoly[3].count());

    for (int c = 0; c < 4; ++c)
        targetPoly[c] = sort(targetPoly[c]);

    update();
}

void MorphingClock::constraintsEvent(Plasma::Constraints constraints)
{
    Plasma::Applet::constraintsEvent(constraints);

    if (constraints & Plasma::SizeConstraint) {
        // FIXME: update our cache (when finally it is implemented)
    }

    if (constraints & Plasma::FormFactorConstraint) {
        if (formFactor() == Plasma::Planar || formFactor() == Plasma::MediaCenter) {
            setPreferredSize(256, 256);
        } else {
            setPreferredSize(-1, -1);
        }
    }
}

QPainterPath MorphingClock::shape() const
{
    QPainterPath path;
    path.addEllipse(boundingRect().adjusted(-2, -2, 2, 2));
    return path;
}

void MorphingClock::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    if (timeLine->state() == QTimeLine::Running)
        return;
    timeLine->toggleDirection();
    timeLine->start();
    event->accept();
}

void MorphingClock::paintInterface(QPainter *p,
        const QStyleOptionGraphicsItem *option, const QRect &contentsRect)
{
    qreal frame = timeLine->currentValue();
    p->save();
    p->setRenderHint(QPainter::Antialiasing);
    p->translate(contentsRect.center());
    p->setBrush(Qt::lightGray);
    p->setPen(Qt::NoPen);
    for (int c = 0; c < 4; ++c) {
        static int order[] = { 0, 3, 1, 2 };
        int i = order[c];
        int level = tween(128, targetLevel[i], frame);
        p->setBrush(QColor(level, level, level));
        p->drawPolygon(tween(sourcePoly[i], targetPoly[i], frame));
    }
    p->setOpacity(1 - frame);
    p->drawEllipse(QPointF(0, radius / 10), radius / 20, radius / 20);
    p->drawEllipse(QPointF(0, -radius / 10), radius / 20, radius / 20);
    p->setOpacity(frame);
    p->drawEllipse(QPointF(0, 0), radius / 5, radius / 5);
    p->restore();
}
 
#include "morphingclock.moc"

