/*
 *   Copyright (C) 2007 Percy Leonhardt <percy@eris23.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "network.h"

#include <QPainter>
#include <QStyleOptionGraphicsItem>
#include <QGraphicsSceneWheelEvent>

#include <KDebug>
#include <KLocale>
#include <KDialog>
#include <KApplication>
#include <KSharedConfig>
#include <kio/global.h>

#include <plasma/svg.h>
#include <plasma/dataengine.h>
#include <plasma/widgets/signalplotter.h>

// ----------------------------------------------------------------------

#define PANEL_SIZE 44.0
#define USED_ON_PANEL (formFactor() == Plasma::Vertical ||      \
                       formFactor() == Plasma::Horizontal)

// ----------------------------------------------------------------------

Network::Label::Label()
    :  m_textWidth(0)
{
}

void Network::Label::setText(const QString &text)
{
    if (m_text != text && m_parent) {
        m_textWidth = m_parent->m_fontMetrics.width(text);
    }
    m_text = text;
}

void Network::Label::setParent(const Network *parent)
{
    m_parent = parent;
}

void Network::Label::invalidate()
{
    m_textWidth = m_parent->m_fontMetrics.width(m_text);
}

int Network::Label::width() const
{
    return m_textWidth;
}

QString Network::Label::text() const
{
    return m_text;
}

// ----------------------------------------------------------------------

Network::Container::Container(const Network *parent)
    : m_parent(parent)
{
}

Network::Container::~Container()
{
    for (int row = 0; row < m_grid.size(); row++) {
        m_grid[row].clear();
    }
}

void Network::Container::invalidateAll()
{
    for (int row = 0; row < m_grid.size(); row++) {
        for (int col = 0; col < m_grid[row].size(); col++) {
            m_grid[row][col].invalidate();
        }
    }
}

Network::Label& Network::Container::getLabel(int row, int col)
{
    bool labelCreated = false;
    if (m_grid.size() < row) {
        m_grid.resize(row);
        m_grid[row - 1].resize(col);
        labelCreated = true;
    } else if (m_grid[row - 1].size() < col) {
        m_grid[row - 1].resize(col);
        labelCreated = true;
    }

    if (labelCreated) {
        m_grid[row - 1][col - 1].setParent(m_parent);
    }
    return m_grid[row - 1][col - 1];
}

int Network::Container::getNumRows()
{
    return m_grid.size();
}

int Network::Container::getRowWidth(int row)
{
    int width = 0;
    foreach(Label label, m_grid[row]) {
        width += label.width();
    }
    return width;
}

// ----------------------------------------------------------------------

Network::Network(QObject *parent, const QVariantList &args)
    : Plasma::Applet(parent, args),
      m_margin(5),
      m_spacing(0),
      m_rxPackets(0),
      m_txPackets(0),
      m_isWireless(false),
      m_connectionActive(false),
      m_rxState("offline"),
      m_txState("offline"),
      m_quality("0percent"),
      m_dialog(0),
      m_fontMetrics(kapp->font()),
      m_container(this)
{
    setAcceptsHoverEvents(true);
    setHasConfigurationInterface(true);
    updateSize();
    resize(sizeHint());
}

Network::~Network()
{
    delete m_dialog;
}

void Network::init()
{
    KConfigGroup cg = config();
    m_container.getLabel(1, 1).setText(cg.readEntry("alias", i18n("Network connection 0")));
    m_container.getLabel(1, 2).setText(cg.readEntry("interface", "eth0"));
    m_textFont = cg.readEntry("textFont", kapp->font());
    m_textColor = cg.readEntry("textColor", QColor(Qt::white));
    m_incomingColor = cg.readEntry("incomingColor", QColor(0x1889FF));
    m_outgoingColor = cg.readEntry("outgoingColor", QColor(0xFF7F08));
    m_detailLevel = (DetailLevel) cg.readEntry("detailLevel", int(LowDetails));
    m_fontMetrics = QFontMetrics(m_textFont);

    m_traffic = new Plasma::Svg("widgets/network_traffic", this);
    m_traffic->setContentType(Plasma::Svg::ImageSet);
    m_traffic->resize(sizeHint());

    m_plotter = new Plasma::SignalPlotter(this);
    m_plotter->addPlot(m_incomingColor);
    m_plotter->addPlot(m_outgoingColor);
    m_plotter->setFont(m_textFont);
    m_plotter->setFontColor(m_textColor);
    m_plotter->scale(1024.0);
    m_plotter->setUseAutoRange(true);
    m_plotter->setUnit("kB/s"); // TODO: use localized string
    m_plotter->setSvgBackground("widgets/plot-background");
    configurePlotter();
    if (m_detailLevel == MinimumDetails || USED_ON_PANEL) {
        m_plotter->hide();
//        m_plotter->setGeometry(QRectF(0.0, 0.0, 5.0, 5.0));
    }

    m_container.getLabel(3, 1).setText(i18n("IN:"));
    m_container.getLabel(4, 1).setText(i18n("OUT:"));
    m_container.getLabel(5, 1).setText(i18n("IP:"));
    m_container.getLabel(6, 1).setText(i18n("MAC:"));
    m_container.getLabel(7, 1).setText(i18n("ESSID:"));
    m_container.getLabel(8, 1).setText(i18n("Encryption:"));

    dataEngine("network")->connectSource(m_container.getLabel(1, 2).text(), this, 1000);
}

QSizeF Network::contentSizeHint() const
{
    return m_size;
}

void Network::constraintsEvent(Plasma::Constraints constraints)
{
    if (constraints & Plasma::FormFactorConstraint) {
        updateSize();
        if (USED_ON_PANEL) {
            m_traffic->resize(sizeHint());
        }
        resize(sizeHint());
    }
}

void Network::dataUpdated(const QString &source, const Plasma::DataEngine::Data &data)
{
    Q_UNUSED(source);
    if (data.size() > 0) {
        m_isWireless = data["Wireless"].toBool();
        m_connectionActive = (data["ConnectionState"].toString() == "Activated");

        int rxPackets = data["RX packets"].toInt();
        int txPackets = data["TX packets"].toInt();
        int linkQuality = data["Link quality"].toInt();

        QString newRxState;
        QString newTxState;
        QString newQuality;
        if (!m_connectionActive) {
            newRxState = "offline";
            newTxState = "offline";
            newQuality = "0percent";
        }
        else {
            if (m_rxPackets < rxPackets &&
                m_txPackets < txPackets) {
                newRxState = "traffic";
                newTxState = "traffic";
            }
            else if (m_rxPackets == rxPackets &&
                     m_txPackets == txPackets) {
                newRxState = "notraffic";
                newTxState = "notraffic";
            }
            else if (m_rxPackets < rxPackets) {
                newRxState = "traffic";
                newTxState = "notraffic";
            }
            else {
                newRxState = "notraffic";
                newTxState = "traffic";
            }

            if (linkQuality < 25) {
                newQuality = "25percent";
            }
            else if (linkQuality < 50) {
                newQuality = "50percent";
            }
            else if (linkQuality < 75) {
                newQuality = "75percent";
            }
            else {
                newQuality = "100percent";
            }
        }

        m_rxPackets = data["RX packets"].toInt();
        m_txPackets = data["TX packets"].toInt();

        m_rxState = newRxState;
        m_txState = newTxState;
        m_quality = newQuality;

        if (m_connectionActive) {
            int upsecs = data["Uptime"].toInt();
            int hrs = upsecs / 3600;
            int mins = (upsecs - hrs * 3600) / 60;
            int secs = upsecs - hrs * 3600 - mins * 60;
            QString uptime;
            if (hrs > 0) {
                uptime.sprintf("%d:%02d:%02d", hrs, mins, secs);
            }
            else {
                uptime.sprintf("%02d:%02d", mins, secs);
            }
            m_container.getLabel(2, 1).setText(i18n("Connection established."));
            m_container.getLabel(2, 2).setText(uptime);

            m_container.getLabel(3, 2).setText(KIO::convertSize(data["RX bytes/s"].toInt()) + i18n("/s"));
            m_container.getLabel(3, 3).setText(KIO::convertSize(data["RX bytes"].toInt()));
            m_container.getLabel(4, 2).setText(KIO::convertSize(data["TX bytes/s"].toInt()) + i18n("/s"));
            m_container.getLabel(4, 3).setText(KIO::convertSize(data["TX bytes"].toInt()));
            m_container.getLabel(5, 2).setText(data["IP"].toString());
            m_container.getLabel(6, 2).setText(data["HW addr"].toString());
            if (data["Wireless"].toBool()) {
                if (data["Encryption"].toBool()) {
                    m_container.getLabel(8, 2).setText(i18n("On"));
                } else {
                    m_container.getLabel(8, 2).setText(i18n("Off"));
                }
                m_container.getLabel(7, 2).setText(data["ESSID"].toString());
                m_container.getLabel(7, 3).setText(data["Bitrate"].toString());
                m_container.getLabel(8, 3).setText(QString("%1%").arg(data["Link quality"].toInt()));
            }
        } else {
            m_container.getLabel(2, 1).setText(i18n("Not connected."));
            m_container.getLabel(2, 2).setText("00:00");
        }

        QList<double> list;
        list.append(data["RX bytes/s"].toInt());
        list.append(data["TX bytes/s"].toInt());
        m_plotter->addSample(list);
        update();
    }
}

void Network::showConfigurationInterface()
{
    if (m_dialog == 0) {
        m_dialog = new KDialog;
        m_dialog->setCaption(i18n("Configure Network Monitor"));
        ui.setupUi(m_dialog->mainWidget());
        m_dialog->mainWidget()->layout()->setMargin(0);
        m_dialog->setButtons(KDialog::Ok | KDialog::Cancel | KDialog::Apply);
        connect(m_dialog, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
        connect(m_dialog, SIGNAL(okClicked()), this, SLOT(configAccepted()));
    }
    ui.fontRequester->setFont(m_textFont);
    ui.textColorButton->setColor(m_textColor);
    ui.incomingColorButton->setColor(m_incomingColor);
    ui.outgoingColorButton->setColor(m_outgoingColor);
    ui.interfaceEdit->setText(m_container.getLabel(1, 2).text());
    ui.aliasEdit->setText(m_container.getLabel(1, 1).text());

    m_dialog->show();
}

bool Network::sceneEvent(QEvent *event)
{
    switch (event->type()) {
    case QEvent::GraphicsSceneWheel:
    {
        if (USED_ON_PANEL) {
            break;
        }
        QGraphicsSceneWheelEvent *wheelEvent = static_cast<QGraphicsSceneWheelEvent*>(event);
        if (wheelEvent->delta() > 0) {
            switch (m_detailLevel) {
            case LowDetails:
                m_detailLevel = MinimumDetails;
                m_plotter->hide();
                break;
            case MediumDetails:
                m_detailLevel = LowDetails;
                break;
            case HighDetails:
                m_detailLevel = MediumDetails;
                break;
            case AllDetails:
                m_detailLevel = HighDetails;
                configurePlotter();
                break;
            default:
                break;
            }
        } else {
            switch (m_detailLevel) {
            case MinimumDetails:
                m_detailLevel = LowDetails;
                m_plotter->show();
                break;
            case LowDetails:
                m_detailLevel = MediumDetails;
                break;
            case MediumDetails:
                m_detailLevel = HighDetails;
                break;
            case HighDetails:
                m_detailLevel = AllDetails;
                configurePlotter();
                break;
            default:
                break;
            }
        }
        updateSize();
        resize(sizeHint());

        KConfigGroup cg = config();
        cg.writeEntry("detailLevel", int(m_detailLevel));
        emit configNeedsSaving();
        break;
    }
    case QEvent::GraphicsSceneHoverEnter:
    {
        if (!USED_ON_PANEL) {
            break;
        }
        QString subText;
        Plasma::DataEngine::Data data = dataEngine("network")->query(m_container.getLabel(1, 2).text());
        if (data["ConnectionState"].toString() == "Activated") {
            subText +="<table cellspacing=1 cellpadding=0 border=0>";
            subText += "<tr><td>" + i18n("IP:") + "</td><td>" + data["IP"].toString() + "</td></tr>";
            subText += "<tr><td>" + i18n("MAC:") + "</td><td>" + data["HW addr"].toString() + "</td></tr>";
            if (data["Wireless"].toBool()) {
                subText += "<tr><td>" + i18n("ESSID:") + "</td><td>" + data["ESSID"].toString() + "</td></tr>";
                subText += "<tr><td>" + i18n("Quality:") + "</td><td>" + QString("%1%").arg(data["Link quality"].toInt()) + "</td></tr>";
            }
            subText += "<tr><td>" + i18n("Incoming:") + "</td><td>" + KIO::convertSize(data["RX bytes/s"].toInt()) + i18n("/s") + "</td></tr>";
            subText += "<tr><td>" + i18n("Outgoing:") + "</td><td>" + KIO::convertSize(data["TX bytes/s"].toInt()) + i18n("/s") + "</td></tr>";
            subText += "</table>";
        } else {
            subText += i18n("Not connected.");
        }

        Plasma::ToolTipData toolTipData;
        toolTipData.mainText = QString("%1 (%2)").arg(m_container.getLabel(1, 1).text()).arg(m_container.getLabel(1, 2).text());
        toolTipData.subText = subText;
        setToolTip(toolTipData);
        break;
    }
    default:
        break;
    }

    return Plasma::Widget::sceneEvent(event);
}

void Network::configAccepted()
{
    if (m_container.getLabel(1, 2).text() != ui.interfaceEdit->text()) {
        dataEngine("network")->disconnectSource(m_container.getLabel(1, 2).text(), this);
        m_container.getLabel(1, 2).setText(ui.interfaceEdit->text());
        dataEngine("network")->connectSource(m_container.getLabel(1, 2).text(), this, 1000);
        m_rxState = "offline";
        m_txState = "offline";
    }

    m_container.getLabel(1, 1).setText(ui.aliasEdit->text());
    m_textFont = ui.fontRequester->font();
    m_textColor = ui.textColorButton->color();
    m_incomingColor = ui.incomingColorButton->color();
    m_outgoingColor = ui.outgoingColorButton->color();
    m_fontMetrics = QFontMetrics(m_textFont);

    KConfigGroup cg = config();
    cg.writeEntry("interface", m_container.getLabel(1, 2).text());
    cg.writeEntry("alias", m_container.getLabel(1, 1).text());
    cg.writeEntry("textFont", m_textFont);
    cg.writeEntry("textColor", m_textColor);
    cg.writeEntry("incomingColor", m_incomingColor);
    cg.writeEntry("outgoingColor", m_outgoingColor);
    emit configNeedsSaving();

    m_plotter->setFont(m_textFont);
    m_plotter->setFontColor(m_textColor);

    m_container.invalidateAll();
    updateSize(); // fontsize can have changed
    resize(sizeHint());
}

void Network::paintInterface(QPainter *p, const QStyleOptionGraphicsItem *option, const QRect &contentsRect)
{
    Q_UNUSED(option);

    p->setRenderHint(QPainter::SmoothPixmapTransform);

    if (USED_ON_PANEL) {
        paintTrafficIcon(p, contentsRect);
    }
    else // desktop or media center mode
    {
        int w = (int) contentsRect.width();
        int h = (int) contentsRect.height();
        if (m_detailLevel == AllDetails) {
            w /= 2;
        }

        p->setFont(m_textFont);
        if (m_connectionActive) {
            p->setPen(QPen(m_textColor));
        } else {
            p->setPen(QPen(m_textColor.darker(200)));
        }

        int fontHeight = m_fontMetrics.height();
        int y = fontHeight;

        // draw first two lines
        p->drawText(m_margin, y, m_container.getLabel(1, 1).text());
        if (m_detailLevel > MinimumDetails) {
            p->drawText(w - m_margin - m_container.getLabel(1, 2).width(), y, m_container.getLabel(1, 2).text());
        }

        y += fontHeight + m_spacing;
        p->drawText(m_margin, y, m_container.getLabel(2, 1).text());
        if (m_detailLevel > MinimumDetails) {
            p->drawText(w - m_margin - m_container.getLabel(2, 2).width(), y, m_container.getLabel(2, 2).text());
        }

        if (m_detailLevel == MinimumDetails) {
            // draw traffic icon
            int iconSize = 2 * fontHeight;
            m_traffic->resize(iconSize, iconSize);
            paintTrafficIcon(p, QRect(w - m_margin - iconSize, m_margin, iconSize, iconSize));
        } else if (m_detailLevel < AllDetails) {
            // draw mini plotter
            y += fontHeight / 2;
            m_plotter->setGeometry(QRectF(m_margin, y, w - 2 * m_margin, 2 * fontHeight));
        } else {
            // draw full plotter
            m_plotter->setGeometry(QRectF(w + m_margin, m_margin, w - 2* m_margin, h - 2 * m_margin));
        }

        if (m_detailLevel > LowDetails)
        {
            // draw the traffic lines
            if (m_detailLevel < AllDetails) {
                // skip the mini plotter
                y += fontHeight * 3;
            } else {
                // only skip a text line
                y += fontHeight + m_spacing;
            }
            p->drawText(m_margin, y, m_container.getLabel(3, 1).text());
            p->drawText((int)(w * 0.6 - m_container.getLabel(3, 2).width()), y, m_container.getLabel(3, 2).text());
            p->drawText(w - m_margin - m_container.getLabel(3, 3).width(), y, m_container.getLabel(3, 3).text());

            y += fontHeight + m_spacing;
            p->drawText(m_margin, y, m_container.getLabel(4, 1).text());
            p->drawText((int)(w * 0.6 - m_container.getLabel(4, 2).width()), y, m_container.getLabel(4, 2).text());
            p->drawText(w - m_margin - m_container.getLabel(4, 3).width(), y, m_container.getLabel(4, 3).text());
        }

        if (m_detailLevel >= HighDetails) {
            // draw IP lines
            y += fontHeight + m_spacing;
            p->drawText(m_margin, y, m_container.getLabel(5, 1).text());
            p->drawText(w - m_margin - m_container.getLabel(5, 2).width(), y, m_container.getLabel(5, 2).text());

            y += fontHeight + m_spacing;
            p->drawText(m_margin, y, m_container.getLabel(6, 1).text());
            p->drawText(w - m_margin - m_container.getLabel(6, 2).width(), y, m_container.getLabel(6, 2).text());

            if (m_isWireless) {
                // draw wireless lines
                y += fontHeight + m_spacing;
                p->drawText(m_margin, y, m_container.getLabel(7, 1).text());
                p->drawText(2 * m_margin + m_container.getLabel(7, 1).width(), y, m_container.getLabel(7, 2).text());
                p->drawText(w - m_margin - m_container.getLabel(7, 3).width(), y, m_container.getLabel(7, 3).text());

                y += fontHeight + m_spacing;
                p->drawText(m_margin, y, m_container.getLabel(8, 1).text());
                p->drawText(2 * m_margin + m_container.getLabel(8, 1).width(), y, m_container.getLabel(8, 2).text());
                p->drawText(w - m_margin - m_container.getLabel(8, 3).width(), y, m_container.getLabel(8, 3).text());
            }
        }
    }
}

Qt::Orientations Network::expandingDirections() const
{
    return 0;
}

void Network::updateSize()
{
    if (USED_ON_PANEL) {
        m_size = QSizeF(PANEL_SIZE, PANEL_SIZE);
        return;
    }

    int h = 0;
    int fontHeight = m_fontMetrics.height();

    switch (m_detailLevel) {
    case MinimumDetails:
        h = 2 * (fontHeight + m_spacing);
        break;
    case LowDetails:
        h = (int) (4.5 * (fontHeight + m_spacing));
        break;
    case MediumDetails:
        h = (int) (6.5 * (fontHeight + m_spacing));
        break;
    case HighDetails:
        if (m_isWireless) {
            h = (int) (10.5 * (fontHeight + m_spacing));
        } else {
            h = (int) (8.5 * (fontHeight + m_spacing));
        }
        break;
    case AllDetails:
        if (m_isWireless) {
            h = 8 * (fontHeight + m_spacing);
        } else {
            h = 6 * (fontHeight + m_spacing);
        }
        break;
    }

    // Calculate the width independent from the current detail level to
    // avoid changes of the applets width when switching the detail level.
    int w = 0;
    for (int row = 0; row < m_container.getNumRows(); row++) {
        w = qMax(w, m_container.getRowWidth(row));
    }

    w += 4 * m_margin; // give some extra space between labels

    if (m_detailLevel == AllDetails) {
        w *= 2;
        w += m_margin;
    }

    h += 2 * m_margin;
    w += 2 * m_margin;
    m_size = QSizeF(w, h);
}

void Network::configurePlotter()
{
    if (m_detailLevel == AllDetails) { // large version
        m_plotter->setThinFrame(false);
        m_plotter->setShowLabels(true);
        m_plotter->setShowTopBar(false);
        m_plotter->setShowVerticalLines(true);
        m_plotter->setShowHorizontalLines(true);
        m_plotter->setHorizontalLinesCount(5);
    } else { // small version
        m_plotter->setThinFrame(false);
        m_plotter->setShowLabels(false);
        m_plotter->setShowTopBar(false);
        m_plotter->setShowVerticalLines(false);
        m_plotter->setShowHorizontalLines(false);
        m_plotter->setHorizontalLinesCount(3);
    }
}

void Network::paintTrafficIcon(QPainter *p, const QRect &geometry)
{
    QRect rect = geometry;
    qreal x = geometry.x();
    qreal y = geometry.y();
    qreal size = geometry.width();
    qreal stateSize = size * 0.75;
    qreal qualitySize = size * 0.25;

    if (m_isWireless == true) {
        // paint quality level
        int w = m_traffic->elementSize(m_quality).width();
        int h = m_traffic->elementSize(m_quality).height();
        int wMax = m_traffic->elementSize("100percent").width();
        int hMax = m_traffic->elementSize("100percent").height();
        float hScaled = ((float) h / hMax) * (size * 0.5);
        rect.setTopLeft(QPoint((int)(x + stateSize), (int)(y + (size * 0.5) - (hScaled / 2))));
        rect.setSize(QSize((int)(((float) w / wMax) * qualitySize), (int)hScaled));
        if (m_traffic->hasElement(m_quality)) {
            m_traffic->paint(p, rect, m_quality);
        }
    } else {
        // paint back monitor
        rect.setTopLeft(QPoint((int)(x + size - stateSize), (int)y));
        rect.setSize(QSize((int)stateSize, (int)stateSize));
        if (m_traffic->hasElement(m_rxState)) {
            m_traffic->paint(p, rect, m_rxState);
        }
    }
    // paint the front monitor
    rect.setTopLeft(QPoint((int)x, (int)(y + size - stateSize)));
    rect.setSize(QSize((int)stateSize, (int)stateSize));
    if (m_traffic->hasElement(m_txState)) {
        m_traffic->paint(p, rect, m_txState);
    }
}

#include "network.moc"
