#include <limits.h>

#include <QtCore>
#include <QtGui>

#include <KConfigDialog>
#include <KGlobalSettings>
#include <KStandardDirs>
#include <KIcon>
#include <KIconLoader>
#include <KDebug>

#include <Plasma/IconWidget>
#include <Plasma/Label>
#include <Plasma/LineEdit>

#include "openbrain.h"

OpenBrain::OpenBrain(QObject *parent, const QVariantList &args) : Plasma::PopupApplet(parent, args), main_widget(0)
{
    setPopupIcon("preferences-desktop-user");
    setHasConfigurationInterface(true);
    setBackgroundHints(DefaultBackground);

    m_layout = 0;
    resize(290,150);
}

QGraphicsWidget *OpenBrain::graphicsWidget()
{
    if (!main_widget) {
        main_widget = new QGraphicsWidget(this);
        QSizePolicy expanding_policy(QSizePolicy::Expanding, QSizePolicy::Expanding);
        QSizePolicy minimal_policy(QSizePolicy::Expanding, QSizePolicy::Maximum);
        m_layout = new QGraphicsLinearLayout(Qt::Vertical);
        m_inputLayout = new QGraphicsLinearLayout(Qt::Horizontal);
        m_inputLineEdit = new Plasma::LineEdit(this);
        m_inputLineEdit->setEnabled(false);
        m_response = new Plasma::Label(this);
        m_inputLayout->addItem(m_inputLineEdit);
        m_go = new Plasma::IconWidget(this);
        m_goAction = new QAction(KIcon("go-jump-locationbar"), QString(), this);
        m_go->setAction(m_goAction);
        m_go->setEnabled(false);
        m_go->setPreferredSize(m_go->sizeFromIconSize(IconSize(KIconLoader::Toolbar)));
        m_inputLayout->addItem(m_go);
        m_home = new Plasma::IconWidget(this);
        m_homeAction = new QAction(KIcon("go-home"), QString(), this);
        m_home->setAction(m_homeAction);
        m_home->setEnabled(false);
        m_home->setPreferredSize(m_home->sizeFromIconSize(IconSize(KIconLoader::Toolbar)));
        m_inputLayout->addItem(m_home);
        m_layout->addItem(m_inputLayout);
        m_layout->addItem(m_response);
        m_response->nativeWidget()->setSizePolicy(expanding_policy);
        m_response->nativeWidget()->setWordWrap(true);
        m_response->nativeWidget()->setTextInteractionFlags(Qt::NoTextInteraction);
        m_response->nativeWidget()->setAlignment(Qt::AlignCenter);
        main_widget->setLayout(m_layout);
        main_widget->setPreferredSize(260, 200);
        setupConnections();
    }

    return main_widget;
}

void OpenBrain::init()
{
    m_aimlset = config().readEntry("aimlset", "default");

    if (!QDir(QDir::homePath() + "/.openbrain").exists()) {
        QProcess p(this);
        p.start("cp " + executeCommand("kde4-config --install data") + "/openbrain/openbrain_configuration " + QDir::homePath() + "/.openbrain -a");
        if (!p.waitForFinished()) {
            QMessageBox::information(0, i18n("Error"), i18n("~/.openbrain did not exist, and copying the default 'brain' from %1 failed.", executeCommand("kde4-config --install data")));
            return;
        }
    }

    loadBrain();
}

void OpenBrain::setupConnections()
{
connect(m_goAction, SIGNAL(triggered()), this, SLOT(returnPressed()));
connect(m_homeAction, SIGNAL(triggered()), this, SLOT(displayHome()));
connect(m_inputLineEdit, SIGNAL(returnPressed()), this, SLOT(returnPressed()));
}

QString OpenBrain::executeCommand(const QString &commandStr)
{
    QString r;
    FILE *file = popen(commandStr.toAscii().data(), "r");
    if (!file) {
        return "";
    }
    char c = 0;
    while (c != EOF) {
        c = (char)getc(file);
        r += c;
    }
    fclose(file);
    return r.left(r.length() - 1).trimmed();
}

void OpenBrain::returnPressed()
{
    m_inputLineEdit->setText(m_inputLineEdit->text().replace("?", ""));
    m_inputLineEdit->setText(m_inputLineEdit->text().replace(".", ""));
    m_inputLineEdit->setText(m_inputLineEdit->text().replace(",", ""));
    m_inputLineEdit->setText(m_inputLineEdit->text().replace("/", ""));
    setText(parser->getResponse(m_inputLineEdit->text()));

    QFile f(QDir::homePath() + "/.openbrain/send_to_author");
    QTextStream out(&f);
    if (f.open(QIODevice::Append))
    out << m_inputLineEdit->text() << "\n";
    out << m_response->text() << "\n\n";

    f.close();

    m_inputLineEdit->setText("");
    m_inputLineEdit->setFocus();
}

void OpenBrain::createConfigurationInterface(KConfigDialog *parent)
{
    m_settings = new OpenBrainSettings;
    parent->addPage(m_settings, i18n("OpenBrain Settings"), "preferences-desktop-user");

    m_settings->comboAimlSet->setItemText(0, m_aimlset);

    readConfig();
    QDir aiml_paths(QDir::homePath() + "/.openbrain/aiml");
    QStringList aiml_sets = aiml_paths.entryList(QDir::NoDotAndDotDot);
    m_settings->comboAimlSet->addItems(aiml_sets);

    connect(m_settings->buttonInstallAIML, SIGNAL(clicked()), m_settings, SLOT(installAimlSet()));

    connect(parent, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
    connect(parent, SIGNAL(okClicked()), this, SLOT(configAccepted()));
}

void OpenBrain::configAccepted()
{
    m_aimlset = m_settings->comboAimlSet->currentText();

    config().writeEntry("aimlset", m_aimlset);
    config().sync();

    emit configNeedsSaving();

    readConfig();
}

void OpenBrain::readConfig()
{
    m_aimlset = config().readEntry("aimlset", m_aimlset);
    setText(m_response->text());
}

void OpenBrain::writeConfig()
{
    m_aimlset = m_settings->comboAimlSet->currentText();

    config().writeEntry("aimlset", m_aimlset);
    config().sync();
}

void OpenBrain::loadBrain()
{
    parser = new AIMLParser();

    connect(parser, SIGNAL(aimlLoad_doc(QString)), this, SLOT(updateLoadingPage(QString)));
    connect(parser, SIGNAL(finished()), this, SLOT(doneLoading()));

    QDir dir(QDir::homePath() + "/.openbrain/aiml/default");
    QStringList files = dir.entryList(QStringList("*.aiml"));
    parser->start();
}

void OpenBrain::displayHome()
{
    m_inputLineEdit->setText(i18n("openbrain home display"));
    emit returnPressed();
}

void OpenBrain::updateLoadingPage(const QString &str)
{
    Q_UNUSED(str);
    setText("Loading...");
}

void OpenBrain::doneLoading()
{
    m_go->setEnabled(true);
    m_home->setEnabled(true);
    m_inputLineEdit->setEnabled(true);
    m_inputLineEdit->setFocus();
    displayHome();
}

void OpenBrain::setText(const QString &str)
{
    m_response->nativeWidget()->setText(str);
}

OpenBrain::~OpenBrain()
{
}

#include "openbrain.moc"
