/***************************************************************************
 *   Copyright (C) 2009 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// own
#include "hoverwidget.h"

// Qt
#include <QPainter>
#include <QGraphicsSceneHoverEvent>

// KDE
#include <KDebug>
#include <Plasma/FrameSvg>
#include <Plasma/Animator>
#include <Plasma/Animation>

#define SPACING 5


HoverWidget::HoverWidget(QGraphicsWidget *parent)
        : QGraphicsWidget(parent)
{
        
    m_svg = new Plasma::FrameSvg(this);
    m_svg->setImagePath("widgets/viewitem");
    m_svg->setElementPrefix("hover");
    connect(m_svg, SIGNAL(repaintNeeded()), this, SLOT(syncBorders()));
  
    m_moveAnimation = Plasma::Animator::create(Plasma::Animator::SlideAnimation);
    m_moveAnimation->setTargetWidget(this);
   
    m_fadeAnimation = Plasma::Animator::create(Plasma::Animator::FadeAnimation);
    m_fadeAnimation->setTargetWidget(this);
    
    syncBorders();
    
}



HoverWidget::~HoverWidget()
{

    delete m_svg;
    delete m_moveAnimation;
    delete m_fadeAnimation;
    
}


qreal HoverWidget::spacing() const
{

    qreal left, top, right, bottom;   
    m_svg->getMargins(left, top, right, bottom);
    return top+bottom;
    
}


QRectF HoverWidget::getHoverRect(QRectF targetRect, const QPointF &pos)
{

    qreal left, top, right, bottom;
    m_svg->getMargins(left, top, right, bottom);
    
    targetRect.setHeight(targetRect.height()+top+bottom);
    targetRect.setWidth(targetRect.width()+left+right);
      
    targetRect.setTopLeft(pos-QPointF(left, top));

    return targetRect;
    
}


void HoverWidget::getMargins(qreal &left, qreal &top, qreal &right, qreal &bottom)
{

    m_svg->getMargins(left, top, right, bottom);

}


void HoverWidget::show()
{
 
    if (!isVisible()) {
        setOpacity(0.0);
        QGraphicsWidget::show();
    }
    if ((opacity() == 1.0 && m_fadeAnimation->state() != Plasma::Animation::Running) ||
        (m_fadeAnimation->state() == Plasma::Animation::Running && 
        m_fadeAnimation->property("targetOpacity") == 1.0)) {
        return;
    }
    
    if (m_fadeAnimation->state() == Plasma::Animation::Running) {
        m_fadeAnimation->stop();
    }
    m_fadeAnimation->setProperty("startOpacity", opacity());
    m_fadeAnimation->setProperty("targetOpacity", 1.0);
    m_fadeAnimation->start();
  
}


void HoverWidget::hide()
{
 
     if ((opacity() == 0.0 && m_fadeAnimation->state() != Plasma::Animation::Running) ||
        (m_fadeAnimation->state() == Plasma::Animation::Running && 
        m_fadeAnimation->property("targetOpacity") == 0.0)) {
        return;
    }
    
    if (m_fadeAnimation->state() == Plasma::Animation::Running) {
        m_fadeAnimation->stop();
    }
    m_fadeAnimation->setProperty("startOpacity", opacity());
    m_fadeAnimation->setProperty("targetOpacity", 0.0);
    m_fadeAnimation->start();

}


void HoverWidget::setTarget(QGraphicsWidget *widget)
{

    if (widget && widget == m_target) {
        return;
    }
    
    if (m_target) {
        m_target->removeEventFilter(this);
    }
    
    m_target = widget;
    if (m_target) {
        m_target->installEventFilter(this);
        move(m_target->pos());
    } else {
        hide();
    }
    
}


void HoverWidget::move(const QPointF &newPos)
{
    
    if (!m_target) {
        return;
    }
    
    const QRectF rect = getHoverRect(m_target->geometry(), newPos);    
    resize(rect.size());
    
    show();
    
    if (m_moveAnimation->state() == Plasma::Animation::Running) {
        m_moveAnimation->stop();
    }
    
    m_moveAnimation->setProperty("distance", qAbs(rect.top()-pos().y()));
    if (rect.top() < pos().y()) {
        m_moveAnimation->setDirection(Plasma::Animation::Forward);
    } else {
        m_moveAnimation->setDirection(Plasma::Animation::Backward);
    }
    m_moveAnimation->start();
    
}


void HoverWidget::syncBorders()
{
    
    update();
    
}


void HoverWidget::hideFinished()
{

    QGraphicsWidget::hide();

}


void HoverWidget::paint(QPainter *painter, const QStyleOptionGraphicsItem *options, QWidget *widget)
{

    Q_UNUSED(options);
    Q_UNUSED(widget);

    m_svg->paintFrame(painter, rect().topLeft());
    
}


void HoverWidget::mousePressEvent(QGraphicsSceneMouseEvent *event)
{

    event->ignore();

}


void HoverWidget::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{

    event->ignore();

}


void HoverWidget::mouseDoubleClickEvent(QGraphicsSceneMouseEvent *event)
{

    event->ignore();

}


void HoverWidget::keyPressEvent(QKeyEvent *event)
{

    Q_UNUSED(event)

}


void HoverWidget::wheelEvent(QGraphicsSceneWheelEvent *event)
{

    event->ignore();
    
}


void HoverWidget::hoverMoveEvent(QGraphicsSceneHoverEvent *event)
{
    
    event->ignore();

}


void HoverWidget::hoverLeaveEvent(QGraphicsSceneHoverEvent *event)
{

    event->ignore();
  
}




void HoverWidget::resizeEvent(QGraphicsSceneResizeEvent *event)
{

    m_svg->resizeFrame(event->newSize());
    QGraphicsWidget::resizeEvent(event);

}


bool HoverWidget::eventFilter(QObject *watched, QEvent *event)
{

    switch (event->type()) {
        case QEvent::GraphicsSceneResize: {
            if (m_target) {
                move(m_target->pos());
            }
            break;
        }
        case QEvent::GraphicsSceneMove: {
            QGraphicsSceneMoveEvent *nEvent = static_cast<QGraphicsSceneMoveEvent*>(event);
            move(QPointF(nEvent->newPos()));
            break;
        }
        default: break;
    }
    return QGraphicsWidget::eventFilter(watched, event);
    
}


#include "hoverwidget.moc"

