/*
 *   Copyright (C) 2009 Andrew Coles  <andrew.coles@yahoo.co.uk>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "coordinate_backend.h"

#include <QFile>
#include <QStringList>
#include <QTextStream>
#include <kio/netaccess.h>
#include <KTemporaryFile>

class FileBackend::Private {

public:

    QHash<QString, Coordinate> records;
    KUrl pathToData;
};

FileBackend::FileBackend(const KUrl & toOpen) : d(new Private) {
    d->pathToData = toOpen;
    
    QString tmpFile;
    if( KIO::NetAccess::download(toOpen, tmpFile, 0)) {	
	QFile fromDisk(tmpFile);
	
	if (fromDisk.open(QIODevice::ReadOnly | QIODevice::Text)) {
	    QTextStream in(&fromDisk);
            while (!in.atEnd()) {
	        const QString line = in.readLine();
		QStringList entries = line.split(',');
		QStringList::iterator itr = entries.begin();
		const QString mac = (itr++)->remove('"');
			
		const double lati = (itr++)->toDouble();
		const double longi = (itr++)->toDouble();
		const double alti = (itr++)->toDouble();
		const double error = (itr++)->toDouble();

		QString ccode;
		QString cname;
		QString city;

		if (itr != entries.end()) {
		    ccode = (itr++)->remove('"');
		    cname = (itr++)->remove('"');
		    city = itr->remove('"');
		}
		
		d->records.insert(mac, Coordinate(lati, longi, alti, error, ccode, cname, city));		
	    }

	}
	
        KIO::NetAccess::removeTempFile(tmpFile);
    }
}

Coordinate FileBackend::resolveMAC(const QString & mac) {
    const QHash<QString, Coordinate>::const_iterator i = d->records.constFind(mac);
    if (i == d->records.constEnd()) return Coordinate();
    return i.value();
}

void FileBackend::addAccessPointData(const QString & mac, const Coordinate & data) {
    const QHash<QString, Coordinate>::iterator i = d->records.find(mac);
    if (i == d->records.end()) {
        d->records.insert(mac, data);
    } else if (data.error < i.value().error) {
        i.value() = data;
    }
}

bool FileBackend::saveAccessPointData() {

    KTemporaryFile tmp;
    if (!tmp.open()) {
        return false;
    }
    
    const QString tmpFile = tmp.fileName();
    
    
    QHash<QString, Coordinate>::iterator itr = d->records.begin();
    const QHash<QString, Coordinate>::iterator itrEnd = d->records.end();
    
    
    QTextStream out(&tmp);

    for (; itr != itrEnd; ++itr) {
        out << '"' << itr.key() << '"'
	    << ',' << itr.value().latitude
	    << ',' << itr.value().longitude
	    << ',' << itr.value().altitude
	    << ',' << itr.value().error
	    << ',' << '"' << itr.value().countryCode << '"'
	    << ',' << '"' << itr.value().countryName << '"'
	    << ',' << '"' << itr.value().city << '"'
	    << "\n";
    }
    
    tmp.close();
    
    return KIO::NetAccess::upload(tmpFile, d->pathToData, 0);
}
