/***************************************************************************
 *   Copyright (C) 2009 by Shawn Starr <shawn.starr@rogers.com>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Ion for BuienRadar.NL XML data */

#include "ion_buienradar.h"
// #include "../../time/solarposition.h"

class BuienradarIon::Private : public QObject
{

public:
    struct XMLMapInfo {
        QString stationName;
        QString stationCode;
    };

    // Key dicts
    QHash<QString, BuienradarIon::Private::XMLMapInfo> m_places;

    // Weather information
    QHash<QString, WeatherData> m_weatherData;

    // Store KIO jobs
    QMap<KJob *, QXmlStreamReader*> m_jobXml;
    QMap<KJob *, QString> m_jobList;
    QXmlStreamReader m_xmlData;

    QDateTime m_dateFormat;
    QString m_source;
    bool m_foundStation;

    bool emitWhenSetup;
    Private() : emitWhenSetup(false) {
    }
};

QMap<QString, IonInterface::WindDirections> BuienradarIon::setupWindIconMappings(void) const
{
    QMap<QString, WindDirections> windDir;

    // Convert Wind directions from Dutch to English for wind icon SVG elements are English keys.
   
    windDir["n"] = N;
    windDir["no"] = NE;
    windDir["nw"] = NW;
    windDir["nno"] = NNE;
    windDir["o"] = E;
    windDir["zo"] = SE;
    windDir["zw"] = SW;
    windDir["ono"] = ENE;
    windDir["zzo"] = SSE;
    windDir["ozo"] = ESE;
    windDir["z"] = S;
    windDir["nnw"] = NNW;
    windDir["wnw"] = WNW;
    windDir["w"] = W;
    windDir["zzw"] = SSW;
    windDir["wzw"] = WSW;

    return windDir;
}

QMap<QString, IonInterface::ConditionIcons> BuienradarIon::setupDayIconMappings(void) const
{

    QMap<QString, ConditionIcons> conditionList;
    conditionList["bewolkt"] = Overcast;
    conditionList["heldere lucht"] = ClearDay;
    return conditionList;
}

QMap<QString, IonInterface::ConditionIcons> BuienradarIon::setupNightIconMappings(void) const
{

    QMap<QString, ConditionIcons> conditionList;
    conditionList["bewolkt"] = Overcast;
    conditionList["heldere lucht"] = ClearNight;
    return conditionList;
}

QMap<QString, IonInterface::ConditionIcons> BuienradarIon::setupForecastIconMappings(void) const
{
    QMap<QString, ConditionIcons> forecastList;
    forecastList["a"] = Overcast;
    return forecastList;
}

QMap<QString, IonInterface::ConditionIcons> const& BuienradarIon::dayIcons(void) const
{
    static QMap<QString, ConditionIcons> const condval = setupDayIconMappings();
    return condval;
}

QMap<QString, IonInterface::ConditionIcons> const& BuienradarIon::nightIcons(void) const
{
    static QMap<QString, ConditionIcons> const condval = setupNightIconMappings();
    return condval;
}

QMap<QString, IonInterface::WindDirections> const& BuienradarIon::windIcons(void) const
{
    static QMap<QString, WindDirections> const wval = setupWindIconMappings();
    return wval;
}

QMap<QString, IonInterface::ConditionIcons> const& BuienradarIon::forecastIcons(void) const
{
    static QMap<QString, ConditionIcons> const foreval = setupForecastIconMappings();
    return foreval;
}

// ctor, dtor
BuienradarIon::BuienradarIon(QObject *parent, const QVariantList &args)
        : IonInterface(parent, args), d(new Private())
{
    Q_UNUSED(args)
}

BuienradarIon::~BuienradarIon()
{
    // Destroy dptr
    delete d;
}

// Get the master list of locations to be parsed
void BuienradarIon::init()
{
    // Get the real city XML URL so we can parse this
    d->m_foundStation = false;
    getXML();
}

void BuienradarIon::reset()
{
    delete d;
    d = new Private();
    d->emitWhenSetup = true;
    setInitialized(false);
    getXML();
}

QStringList BuienradarIon::validate(const QString& source) const
{
    QStringList placeList;
    QString sourceNormalized = source.toUpper();

    QHash<QString, BuienradarIon::Private::XMLMapInfo>::const_iterator it = d->m_places.constBegin();

    while (it != d->m_places.constEnd()) {
        if (it.key().toUpper().contains(sourceNormalized)) {
            placeList.append(QString("place|").append(it.key()));
        }
        ++it;
    }

    // Check if placeList is empty if so, return nothing.
    if (placeList.isEmpty()) {
        return QStringList();
    }

    placeList.sort();
    return placeList;
}

bool BuienradarIon::updateIonSource(const QString& source)
{
    // We expect the applet to send the source in the following tokenization:
    // ionname:validate:place_name - Triggers validation of place
    // ionname:weather:place_name - Triggers receiving weather of place

    QStringList sourceAction = source.split('|');

    // Guard: if the size of array is not 2 then we have bad data, return an error
    if (sourceAction.size() < 2) {
        setData(source, "validate", "buienradar|malformed");
        return true;
    }
    
    if (sourceAction[1] == "validate" && sourceAction.size() > 2) {
        kDebug() << "Initiate Validating of place: " << sourceAction[2];
        QStringList result = validate(sourceAction[2]);

        if (result.size() == 1) {
            setData(source, "validate", QString("buienradar|valid|single|%1").arg(result.join("|")));
            return true;
        } else if (result.size() > 1) {
            setData(source, "validate", QString("buienradar|valid|multiple|%1").arg(result.join("|")));
            return true;
        } else if (result.size() == 0) {
            setData(source, "validate", QString("buienradar|invalid|single|%1").arg(sourceAction[2]));
            return true;
        }

    } else if (sourceAction[1] == "weather" && sourceAction.size() > 2) {
        getXMLData(source);
        return true;
    } else {
        setData(source, "validate", "buienradar|malformed");
        return true;
    }

    return false;
}

// Parses city list and gets the correct city based on ID number
void BuienradarIon::getXML() const
{
    KIO::TransferJob *job = KIO::get(KUrl("http://xml.buienradar.nl"), KIO::NoReload, KIO::HideProgressInfo);

    if (job) {
        connect(job, SIGNAL(data(KIO::Job *, const QByteArray &)), this,
                SLOT(setup_slotDataArrived(KIO::Job *, const QByteArray &)));
        connect(job, SIGNAL(result(KJob *)), this, SLOT(setup_slotJobFinished(KJob *)));
    }
}

// Gets specific city XML data
void BuienradarIon::getXMLData(const QString& source)
{
    KUrl url;

    QString dataKey = source;
    dataKey.remove("buienradar|weather|");
    url = "http://xml.buienradar.nl";

    kDebug() << "URL Location: " << url.url();

    // If this is empty we have no valid data, send out an error and abort.
    //
    if (url.url().isEmpty()) { 
        setData(source, "validate", QString("buienradar|malformed"));
        return;
    }

    KIO::TransferJob *myJob = KIO::get(url.url(), KIO::Reload, KIO::HideProgressInfo);
    d->m_jobXml.insert(myJob, new QXmlStreamReader);
    d->m_jobList.insert(myJob, source);

    if (myJob) {
        connect(myJob, SIGNAL(data(KIO::Job *, const QByteArray &)), this,
                SLOT(slotDataArrived(KIO::Job *, const QByteArray &)));
        connect(myJob, SIGNAL(result(KJob *)), this, SLOT(slotJobFinished(KJob *)));
    }
}

void BuienradarIon::setup_slotDataArrived(KIO::Job *job, const QByteArray &data)
{
    Q_UNUSED(job)

    if (data.isEmpty()) {
        return;
    }

    // Send to xml.
    d->m_xmlData.addData(data);
}

void BuienradarIon::slotDataArrived(KIO::Job *job, const QByteArray &data)
{
    if (data.isEmpty() || !d->m_jobXml.contains(job)) {
        return;
    }

    // Send to xml.
    d->m_jobXml[job]->addData(data);
}

void BuienradarIon::slotJobFinished(KJob *job)
{
    // Dual use method, if we're fetching location data to parse we need to do this first
    setData(d->m_jobList[job], Data());
    readXMLData(d->m_jobList[job], *d->m_jobXml[job]);
    d->m_jobList.remove(job);
    delete d->m_jobXml[job];
    d->m_jobXml.remove(job);
}

void BuienradarIon::setup_slotJobFinished(KJob *job)
{
    Q_UNUSED(job)
    const bool success = readXML();
    setInitialized(success);
    if (d->emitWhenSetup) {
        d->emitWhenSetup = false;
        emit(resetCompleted(this, success));
    }

}

void BuienradarIon::parseStation(WeatherData& data, QXmlStreamReader& xml)
{
    data.temperature_C = i18n("N/A");
    data.weather = i18n("N/A");
    data.stationCode = i18n("N/A");
    data.pressure = i18n("N/A");
    data.visibility = i18n("N/A");
    data.humidity = i18n("N/A");
    data.windSpeed_MS = i18n("N/A");
    data.windGust = i18n("N/A");

    while (!xml.atEnd()) {
        xml.readNext();

        if (xml.isEndElement() && xml.name() == "weerstation") {
            break;
        }

        if (xml.isStartElement()) {
            if (xml.name() == "stationcode") {
                data.stationCode = xml.readElementText();
            } else if (xml.name() == "stationnaam") {
                data.locationName = xml.readElementText().replace("Meetstation ", "");

                if (d->m_source.isEmpty()) {
                    BuienradarIon::Private::XMLMapInfo info;
                    info.stationName = data.locationName;
                    info.stationCode = data.stationCode;
                    d->m_places[data.locationName] = info;
                }

            } else if (xml.name() == "lat") {
                data.stationLat = xml.readElementText();
            } else if (xml.name() == "lon") {
                data.stationLon = xml.readElementText();
            } else if (xml.name() == "datum") {
                data.observationTime = xml.readElementText();
                //QStringList tmpDateStr = data.observationTime.split(' ');
               //data.observationTime = QString("%1 %2").arg(tmpDateStr[5]).arg(tmpDateStr[6]);
                //d->m_dateFormat = QDateTime::fromString(data.observationTime, "h:mm ap");
            } else if (xml.name() == "luchtvochtigheid") {
                data.humidity = xml.readElementText();
                if (data.humidity == "-") {
                    data.humidity = i18n("N/A");
                }
            } else if (xml.name() == "temperatuurGC") {
                data.temperature_C = xml.readElementText();
                if (data.temperature_C == "-") {
                    data.temperature_C = i18n("N/A");
                }
            } else if (xml.name() == "windsnelheidMS") {
                data.windSpeed_MS = xml.readElementText();
                if (data.windSpeed_MS == "-") {
                    data.windSpeed_MS = i18n("N/A");
                }
            } else if (xml.name() == "windsnelheidBF") {
                xml.readElementText(); // Skip it
            } else if (xml.name() == "windrichtingGR") {
                xml.readElementText(); // Skip it
            } else if (xml.name() == "windrichting") {
                data.windDirection = xml.readElementText();
                if (data.windDirection == "-") {
                    data.windDirection = i18n("N/A");
                }
            } else if (xml.name() == "luchtdruk") {
                data.pressure = xml.readElementText();
                if (data.pressure == "-") {
                    data.pressure = i18n("N/A");
                }
            } else if (xml.name() == "zichtmeters") {
                data.visibility = xml.readElementText();
                if (data.visibility == "-") {
                    data.visibility = i18n("N/A");
                } else {
                    float tmpVisibility = data.visibility.toFloat() / 1000; // Convert from meters to km
                    data.visibility = QString::number(tmpVisibility, 'f', 1);
                }
            } else if (xml.name() == "windstotenMS") {
                data.windGust = xml.readElementText();
                if (data.windGust == "0") {
                    data.windGust = i18n("N/A");
                }
            } else if (xml.name() == "regenMMPU") {
                xml.readElementText(); // Skip it
            } else if (xml.name() == "icoonactueel") {
                data.weather = xml.attributes().value("zin").toString();
                // zonnig - Not available
                if (data.weather == "zonnig" || data.weather == "-") {
                    data.weather = i18n("N/A");
                }
            } else {
                parseUnknownElement(xml);
            }
        }
    }
}

void BuienradarIon::parseForecast(WeatherData& data, QXmlStreamReader& xml)
{
    while (!xml.atEnd()) {
        xml.readNext();
  
        if (xml.isEndElement()) {
            break;
        }
        
        if (xml.name() == "url") {
            xml.readElementText();
        } else if (xml.name() == "urlbackup") {
            xml.readElementText();
        } else if (xml.name() == "tekst_middellang") {
            xml.readElementText();
        } else if (xml.name() == "tekst_lang") {
            xml.readElementText();
        } else if (xml.name().toString().contains("dag-plus")) {
            parseDayForecast(data, xml);
        } else {
            parseUnknownElement(xml);
        }
   }
}

void BuienradarIon::parseDayForecast(WeatherData& data, QXmlStreamReader& xml)
{
    while (!xml.atEnd()) {
        xml.readNext();
 
        if (xml.isEndElement()) {
            break;
        }

        if (xml.name() == "datum") {
            xml.readElementText();
        } else if (xml.name() == "dagweek") {
            kDebug() << "Day of Week: " << xml.readElementText();
        } else if (xml.name() == "kanszon") {
            xml.readElementText();
        } else if (xml.name() == "kansregen") {
            kDebug() << "Probability of Precipitation: " << xml.readElementText();
        } else if (xml.name() == "minmmregen") {
            xml.readElementText();
        } else if (xml.name() == "maxmmregen") {
            xml.readElementText();
        } else if (xml.name() == "mintemp") {
            kDebug() << "Min Temperature: " << xml.readElementText();
        } else if (xml.name() == "maxtemp") {
            kDebug() << "Max Temperature: " << xml.readElementText();
        } else if (xml.name() == "windrichting") {
            kDebug() << "Wind Direction: " << xml.readElementText();
        } else if (xml.name() == "windkracht") {
            kDebug() << "Wind Speed: " << xml.readElementText();
        } else if (xml.name() == "icoon") {
            kDebug() << "Wind Icon to use: " << xml.attributes().value("ID").toString();
        } else {
            parseUnknownElement(xml);
        }
    }
}

void BuienradarIon::parseWeatherStations(WeatherData& data, QXmlStreamReader& xml)
{
    while (!xml.atEnd()) {
        xml.readNext();

        if (xml.isEndElement()) {
            break;
        }

        if (xml.isStartElement() && xml.name() == "weerstation") {
            if (xml.name() == "weerstation") {
                if (!d->m_foundStation) {
                    parseStation(data, xml);
                }
                if (!d->m_source.isEmpty()) {
                    if (d->m_source.split('|')[2] == data.locationName) {
                        d->m_foundStation = true;
                        return;
                    }
                }
            } 
        } else {
             parseUnknownElement(xml);
        }
    }
}

void BuienradarIon::parseStationList(WeatherData& data, QXmlStreamReader& xml)
{
    while (!xml.atEnd()) {
        xml.readNext();

        if (xml.isStartElement() && xml.name() == "weerstations") {
            if (xml.name() == "weerstations") {
                parseWeatherStations(data, xml);
            } else {
                parseUnknownElement(xml);
            }
        } else if (xml.isStartElement() && xml.name() == "verwachting_meerdaags") {
            if (xml.name() == "verwachting_meerdaags") {
                parseForecast(data, xml);
            } else {
                parseUnknownElement(xml);
            }
       }
    }
 
}

// Parse the city list and store into a QMap
bool BuienradarIon::readXML()
{
    WeatherData data;
    bool success = false;

    while (!d->m_xmlData.atEnd()) {
        d->m_xmlData.readNext();

        if (d->m_xmlData.isStartElement()) {
            if (d->m_xmlData.name() == "buienradarnl") {
                d->m_source = QString();
                parseStationList(data, d->m_xmlData);
                success = true;
            }
        }
    }
    return (!d->m_xmlData.error() && success);
}

// Parse Weather data main loop, from here we have to decend into each tag pair
bool BuienradarIon::readXMLData(const QString& source, QXmlStreamReader& xml)
{
    WeatherData data;
    while (!xml.atEnd()) {
        xml.readNext();

        if (xml.isEndElement()) {
            break;
        }

        if (xml.isStartElement()) {
            if (xml.name() == "buienradarnl") {
                d->m_source = source;
                parseStationList(data, xml);
            } else {
                parseUnknownElement(xml);
            }
        }
    }

    d->m_weatherData[source] = data;
    updateWeather(source);
    return !xml.error();
}

// handle when no XML tag is found
void BuienradarIon::parseUnknownElement(QXmlStreamReader& xml) const
{

    while (!xml.atEnd()) {
        xml.readNext();

        if (xml.isEndElement()) {
            break;
        }

        if (xml.isStartElement()) {
            parseUnknownElement(xml);
        }
    }
}

void BuienradarIon::updateWeather(const QString& source)
{
    QMap<QString, QString> dataFields;
    Plasma::DataEngine::Data data;
    QStringList fieldList;

    data.insert("Country", country(source));
    data.insert("Place", QString("%1, %2").arg(place(source)).arg(country(source)));
    data.insert("Station", station(source));

    data.insert("Latitude", latitude(source));
    data.insert("Longitude", longitude(source));

    // Real weather - Current conditions
    data.insert("Observation Period", observationTime(source));
    data.insert("Current Conditions", condition(source));
    data.insert("Condition Icon", getWeatherIcon(dayIcons(), condition(source)));

    dataFields = temperature(source);
    data.insert("Temperature", dataFields["temperature"]);
    data.insert("Temperature Unit", dataFields["temperatureUnit"]);

    dataFields = pressure(source);
    data.insert("Pressure", dataFields["pressure"]);
    data.insert("Pressure Unit", dataFields["pressureUnit"]);

    dataFields = visibility(source);
    data.insert("Visibility", dataFields["visibility"]);
    data.insert("Visibility Unit", dataFields["visibilityUnit"]);

    dataFields = humidity(source);
    data.insert("Humidity", dataFields["humidity"]);
    data.insert("Humidity Unit", dataFields["humidityUnit"]);

    // Set number of forecasts per day/night supported, none for this ion right now
    data.insert(QString("Total Weather Days"), 0);

    dataFields = wind(source);
    data.insert("Wind Speed", dataFields["windSpeed"]);
    data.insert("Wind Speed Unit", dataFields["windUnit"]);

    data.insert("Wind Gust", dataFields["windGust"]);
    data.insert("Wind Gust Unit", dataFields["windGustUnit"]);
    data.insert("Wind Direction", getWindDirectionIcon(windIcons(), dataFields["windDirection"].toLower()));
    data.insert("Credit", i18n("Provided by: BuienRadar.NL - http://www.buienradar.nl"));
    data.insert("Credit Url", "http://www.buienradar.nl");
    setData(source, data);
}

QString const BuienradarIon::country(const QString& source) const
{
    Q_UNUSED(source);
    return QString("Netherlands");
}
QString BuienradarIon::place(const QString& source) const
{
    return d->m_weatherData[source].locationName;
}
QString BuienradarIon::station(const QString& source) const
{
    return d->m_weatherData[source].stationCode;
}

QString BuienradarIon::latitude(const QString& source) const
{
    return d->m_weatherData[source].stationLat;
}

QString BuienradarIon::longitude(const QString& source) const
{
    return d->m_weatherData[source].stationLon;
}

QString BuienradarIon::observationTime(const QString& source) const
{
    return d->m_weatherData[source].observationTime;
}

QString BuienradarIon::condition(const QString& source) const
{
    if (d->m_weatherData[source].weather == "N/A") {
        d->m_weatherData[source].weather = i18n("N/A");
    }
    return i18nc("weather condition", d->m_weatherData[source].weather.toUtf8());
}

QMap<QString, QString> BuienradarIon::humidity(const QString& source) const
{
    QMap<QString, QString> humidityInfo; 
    if (d->m_weatherData[source].humidity == "N/A") {
        humidityInfo.insert("humidity", QString(i18n("N/A")));
        humidityInfo.insert("humidityUnit", QString::number(WeatherUtils::NoUnit));
        return humidityInfo;
    } else {
        humidityInfo.insert("humidity", d->m_weatherData[source].humidity);
        humidityInfo.insert("humidityUnit", QString::number(WeatherUtils::Percent));
    }
    
    return humidityInfo;
}

QMap<QString, QString> BuienradarIon::visibility(const QString& source) const
{
    QMap<QString, QString> visibilityInfo;
    if (d->m_weatherData[source].visibility == "N/A") {
        visibilityInfo.insert("visibility", QString("N/A"));
        visibilityInfo.insert("visibilityUnit", QString::number(WeatherUtils::NoUnit));
        return visibilityInfo;
    }
    visibilityInfo.insert("visibility", d->m_weatherData[source].visibility);
    visibilityInfo.insert("visibilityUnit", QString::number(WeatherUtils::Kilometers));
    return visibilityInfo;
}

QMap<QString, QString> BuienradarIon::temperature(const QString& source) const
{
    QMap<QString, QString> temperatureInfo;
    temperatureInfo.insert("temperature", d->m_weatherData[source].temperature_C);
    temperatureInfo.insert("temperatureUnit", QString::number(WeatherUtils::Celsius));

    return temperatureInfo;
}

QMap<QString, QString> BuienradarIon::pressure(const QString& source) const
{
    QMap<QString, QString> pressureInfo;
    if (d->m_weatherData[source].pressure == "N/A") {
        pressureInfo.insert("pressure", "N/A");
        pressureInfo.insert("pressureUnit", QString::number(WeatherUtils::NoUnit));
        return pressureInfo;
    }

    pressureInfo.insert("pressure", d->m_weatherData[source].pressure);
    pressureInfo.insert("pressureUnit", QString::number(WeatherUtils::Hectopascals));
    return pressureInfo;
}

QMap<QString, QString> BuienradarIon::wind(const QString& source) const
{
    QMap<QString, QString> windInfo;

    // May not have any winds
    if (d->m_weatherData[source].windSpeed_MS == "N/A") {
        windInfo.insert("windSpeed", "Calm");
        windInfo.insert("windUnit", QString::number(WeatherUtils::NoUnit));
    } else {
        windInfo.insert("windSpeed", QString::number(d->m_weatherData[source].windSpeed_MS.toFloat(), 'f', 1));
        windInfo.insert("windUnit", QString::number(WeatherUtils::MetersPerSecond));
    }

    // May not always have gusty winds
    if (d->m_weatherData[source].windGust == "N/A") {
        windInfo.insert("windGust", "N/A");
        windInfo.insert("windGustUnit", QString::number(WeatherUtils::NoUnit));
    } else {
        windInfo.insert("windGust", QString::number(d->m_weatherData[source].windGust.toFloat(), 'f', 1));
        windInfo.insert("windGustUnit", QString::number(WeatherUtils::MetersPerSecond));
    }

    if (d->m_weatherData[source].windDirection == "N/A") {
        windInfo.insert("windDirection", "N/A");
    } else {
        windInfo.insert("windDirection", d->m_weatherData[source].windDirection);
    }
    return windInfo;
}

#include "ion_buienradar.moc"
