/*
 *   Copyright 2010 Brian Pritchett <batenkaitos@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include "storageresource.h"
#include "akonadi/cachepolicy.h"
#include "akonadi/collectionfetchjob.h"
#include <QtDBus/QDBusConnection>
#include <Akonadi/CollectionCreateJob>
#include <Akonadi/CollectionFetchScope>
#include <Akonadi/ChangeRecorder>
#include <Akonadi/ItemFetchScope>
#include <KConfig>
#include <KConfigGroup>

using namespace Akonadi;

StorageResource::StorageResource( const QString &id )
        : ResourceBase( id )
{
//    new SettingsAdaptor( Settings::self() );
//    QDBusConnection::sessionBus().registerObject( QLatin1String( "/Settings" ),
//            Settings::self(), QDBusConnection::ExportAdaptors );
    changeRecorder()->itemFetchScope().fetchFullPayload( true );
    changeRecorder()->fetchCollection( true );
    initStorage();
}

StorageResource::~StorageResource()
{
}

void StorageResource::initStorage()
{
    QStringList mimeTypes;
    mimeTypes << "application/x-vnd.kde.storage" << "application/octet-stream";

    Akonadi::Collection root;
    root.setParent(Akonadi::Collection::root());
    root.setName( "Storage" );
    root.setRemoteId("plasmastorage");
    root.setContentMimeTypes( QStringList() << Akonadi::Collection::mimeType() << mimeTypes );
    Akonadi::Collection::Rights rootRights = Akonadi::Collection::CanChangeCollection;
    rootRights |= Akonadi::Collection::CanCreateCollection;
    rootRights |= Akonadi::Collection::CanDeleteCollection;
    root.setRights(rootRights);

    Akonadi::CachePolicy policy;
    policy.setInheritFromParent( false );
    policy.setSyncOnDemand( true );
    policy.setIntervalCheckTime( -1 );
    root.setCachePolicy( policy );

    storageBase = root;

    setHierarchicalRemoteIdentifiersEnabled(true);
    synchronizeCollectionTree();
}

void StorageResource::retrieveCollections()
{
    QHash<QString, Akonadi::Collection> collections;
    QStringList mimeTypes;
    mimeTypes << "application/x-vnd.kde.storage" << "application/octet-stream";
    collections["rootfolderunique"] = storageBase;

    Akonadi::Collection::Rights sourceRights = Akonadi::Collection::CanChangeItem;
    sourceRights |= Akonadi::Collection::CanDeleteItem;
    sourceRights |= Akonadi::Collection::CanCreateItem;

    KConfig config("plasma-storagerc");
    foreach (QString engine, config.groupList()) {
        Akonadi::Collection c;
        c.setContentMimeTypes(QStringList() << Akonadi::Collection::mimeType() << mimeTypes);
        c.setName(engine);
        c.setParentCollection(storageBase);
        c.setRights(storageBase.rights());
        c.setCachePolicy(storageBase.cachePolicy());
        c.setRemoteId(engine);
        collections[engine] = c;
        KConfigGroup engineGroup(&config, engine);
        foreach (QString source, engineGroup.groupList()) {
            Akonadi::Collection s;
            s.setContentMimeTypes(QStringList() << mimeTypes);
            s.setName(source);
            s.setParentCollection(c);
            s.setRights(sourceRights);
            s.setCachePolicy(storageBase.cachePolicy());
            s.setRemoteId(source);
            collections[engine + "-" + source] = s;
        }
    }
    collectionsRetrieved(collections.values());
}

void StorageResource::collectionAdded( const Akonadi::Collection &collection,
        const Akonadi::Collection &parent )
{
    Collection newCollection( collection );
    newCollection.setRemoteId( collection.name() );
    newCollection.setParentCollection(parent);
    changeCommitted( newCollection ); 
}

void StorageResource::retrieveItems( const Akonadi::Collection &collection )
{
    Akonadi::CollectionFetchJob *job = new Akonadi::CollectionFetchJob(collection, 
            Akonadi::CollectionFetchJob::Base);
    Akonadi::CollectionFetchScope scope;
    Item::List items;
    scope.setAncestorRetrieval(Akonadi::CollectionFetchScope::Parent);
    job->setFetchScope(scope);
    if (!job->exec()) {
        itemsRetrieved(items);
        return;
    }
    Akonadi::Collection fetched(job->collections().first());

    QString source = fetched.remoteId();
    QString dataEngine = fetched.parentCollection().remoteId();
    KConfig config("plasma-storagerc");
    KConfigGroup dataEngineGroup(&config, dataEngine); 
    KConfigGroup sourceGroup(&dataEngineGroup, source);

    if (collection.parentCollection().name() == "Storage") {
        itemsRetrieved(items);
        return;
    }

    foreach (QString key, sourceGroup.keyList()){
        Item item( "application/x-vnd.kde.storage" ) ;
        QVariantHash vh;
        vh[QString("key")] = key;
        vh[QString("data")] = sourceGroup.readEntry(key).toAscii();
        item.setPayload<QVariantHash>(vh);
        item.setRemoteId(key);
        items << item;
    }
    itemsRetrieved(items);
}

bool StorageResource::retrieveItem( const Akonadi::Item &item, const QSet<QByteArray> &parts )
{
    Q_UNUSED( item );
    Q_UNUSED( parts );

    // TODO: this method is called when Akonadi wants more data for a given item.
    // You can only provide the parts that have been requested but you are allowed
    // to provide all in one go

    return true;
}

void StorageResource::aboutToQuit()
{
    // TODO: any cleanup you need to do while there is still an active
    // event loop. The resource will terminate after this method returns
}

void StorageResource::configure( WId windowId )
{
    Q_UNUSED( windowId );

    // TODO: this method is usually called when a new resource is being
    // added to the Akonadi setup. You can do any kind of user interaction here,
    // e.g. showing dialogs.
    // The given window ID is usually useful to get the correct
    // "on top of parent" behavior if the running window manager applies any kind
    // of focus stealing prevention technique
    //
    // If the configuration dialog has been accepted by the user by clicking Ok,
    // the signal configurationDialogAccepted() has to be emitted, otherwise, if
    // the user canceled the dialog, configurationDialogRejected() has to be emitted.
}

void StorageResource::itemAdded( const Akonadi::Item &item, const Akonadi::Collection &collection )
{
    Akonadi::CollectionFetchJob *job = new Akonadi::CollectionFetchJob(collection, 
            Akonadi::CollectionFetchJob::Base);
    Akonadi::CollectionFetchScope scope;
    scope.setAncestorRetrieval(Akonadi::CollectionFetchScope::Parent);
    job->setFetchScope(scope);
    if (!job->exec()) {
        return;
    }
    Akonadi::Collection fetched(job->collections().first());

    QString source = fetched.remoteId();
    QString dataEngine = fetched.parentCollection().remoteId();
    KConfig config("plasma-storagerc");
    KConfigGroup dataEngineGroup(&config, dataEngine);
    KConfigGroup sourceGroup(&dataEngineGroup, source);

    QVariantHash v;
    if (item.hasPayload<QVariantHash>()) {
        v = item.payload<QVariantHash>();
    }
    sourceGroup.writeEntry(v[QString("key")].toString(), v[QString("data")]);

    Item result(item);
    result.setRemoteId(v[QString("key")].toString());
    result.setPayload<QVariantHash>(v);
    changeCommitted(result);
}

void StorageResource::itemChanged( const Akonadi::Item &item, const QSet<QByteArray> &parts )
{
    Q_UNUSED( item );
    Q_UNUSED( parts );

    // TODO: this method is called when somebody else, e.g. a client application,
    // has changed an item managed by your resource.

    // NOTE: There is an equivalent method for collections, but it isn't part
    // of this template code to keep it simple
}

void StorageResource::itemRemoved( const Akonadi::Item &item )
{
    Q_UNUSED( item );

    const QString source = item.parentCollection().name();
    const QString dataEngine = item.parentCollection().parentCollection().name();
    KConfig config("plasma-storagerc");
    KConfigGroup dataEngineGroup(&config, dataEngine);
    KConfigGroup sourceGroup(&dataEngineGroup, source);
    sourceGroup.writeEntry(item.remoteId(), "");
}

AKONADI_RESOURCE_MAIN( StorageResource )

#include "storageresource.moc"
