/*
 *   Copyright 2008 Aike J Sommer <dev@aikesommer.name>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2,
 *   or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


#include "screensengine.h"
#include "outputservice.h"
#include "configurationsservice.h"

#include <kephal/outputs.h>
#include <kephal/screens.h>
#include <kephal/configurations.h>


ScreensEngine::ScreensEngine(QObject* parent, const QVariantList& args)
    : Plasma::DataEngine(parent, args)
{
    Q_UNUSED(args)
    
    setMinimumPollingInterval(1000);
}

bool ScreensEngine::sourceRequestEvent(const QString &name) {
    return updateSourceEvent(name);
}

bool ScreensEngine::updateSourceEvent(const QString &name) {
    Q_UNUSED(name)
    
    return false;
}

Plasma::Service * ScreensEngine::serviceForSource(const QString& source)
{
    if (m_services.contains(source)) {
        return m_services[source];
    } else if (source.startsWith("output-")) {
        Kephal::Output * output = outputForSource(source);
        if (output) {
            return m_services.insert(source, new OutputService(this, output)).value();
        }
    } else if (source == "configurations") {
        return m_services.insert(source, new ConfigurationsService(this)).value();
    }
    return DataEngine::serviceForSource(source);
}

Kephal::Output * ScreensEngine::outputForSource(const QString & source) {
    QList<Kephal::Output *> outputs = Kephal::Outputs::self()->outputs();
    foreach (Kephal::Output * output, outputs) {
        if (source == ("output-" + output->id())) {
            return output;
        }
    }
    return 0;
}

void ScreensEngine::init() {
    Kephal::Screens * screens = Kephal::Screens::self();
    foreach (Kephal::Screen * screen, screens->screens()) {
        screenAdded(screen);
    }
    
    Kephal::Outputs * outputs = Kephal::Outputs::self();
    foreach (Kephal::Output * output, outputs->outputs()) {
        dataForOutput(output);
    }
    
    connect(Kephal::Screens::self(), SIGNAL(screenResized(Kephal::Screen *, QSize, QSize)), this, SLOT(screenResized(Kephal::Screen *, QSize, QSize)));
    connect(Kephal::Screens::self(), SIGNAL(screenMoved(Kephal::Screen *, QPoint, QPoint)), this, SLOT(screenMoved(Kephal::Screen *, QPoint, QPoint)));
    connect(Kephal::Screens::self(), SIGNAL(screenAdded(Kephal::Screen *)), this, SLOT(screenAdded(Kephal::Screen *)));
    connect(Kephal::Screens::self(), SIGNAL(screenRemoved(int)), this, SLOT(screenRemoved(int)));
    
    connect(Kephal::Outputs::self(), SIGNAL(outputConnected(Kephal::Output *)), this, SLOT(outputConnected(Kephal::Output *)));
    connect(Kephal::Outputs::self(), SIGNAL(outputDisconnected(Kephal::Output *)), this, SLOT(outputDisconnected(Kephal::Output *)));
    connect(Kephal::Outputs::self(), SIGNAL(outputActivated(Kephal::Output *)), this, SLOT(outputActivated(Kephal::Output *)));
    connect(Kephal::Outputs::self(), SIGNAL(outputDeactivated(Kephal::Output *)), this, SLOT(outputDeactivated(Kephal::Output *)));
    connect(Kephal::Outputs::self(), SIGNAL(outputResized(Kephal::Output *, QSize, QSize)), this, SLOT(outputResized(Kephal::Output *, QSize, QSize)));
    connect(Kephal::Outputs::self(), SIGNAL(outputMoved(Kephal::Output *, QPoint, QPoint)), this, SLOT(outputMoved(Kephal::Output *, QPoint, QPoint)));
    connect(Kephal::Outputs::self(), SIGNAL(outputRateChanged(Kephal::Output *, float, float)), this, SLOT(outputRateChanged(Kephal::Output *, float, float)));
    connect(Kephal::Outputs::self(), SIGNAL(outputRotated(Kephal::Output *, Kephal::Rotation, Kephal::Rotation)), this, SLOT(outputRotated(Kephal::Output *, Kephal::Rotation, Kephal::Rotation)));
    connect(Kephal::Outputs::self(), SIGNAL(outputReflected(Kephal::Output *, bool, bool, bool, bool)), this, SLOT(outputReflected(Kephal::Output *, bool, bool, bool, bool)));
    
    connect(Kephal::Configurations::self(), SIGNAL(confirmTimeout(int)), this, SLOT(confirmTimeout(int)));
    connect(Kephal::Configurations::self(), SIGNAL(confirmed()), this, SLOT(confirmed()));
    connect(Kephal::Configurations::self(), SIGNAL(reverted()), this, SLOT(reverted()));
    
    setData("configurations", I18N_NOOP("Confirmation-Timeout"), 0);
    setData("configurations", I18N_NOOP("Confirmation-Timeout-Max"), CONFIRMATION_TIME);
}

void ScreensEngine::screenResized(Kephal::Screen * screen, QSize oldSize, QSize newSize) {
    Q_UNUSED(oldSize)
    setData("screen-" + QString::number(screen->id()), I18N_NOOP("Size"), newSize);
}

void ScreensEngine::screenMoved(Kephal::Screen * screen, QPoint oldPosition, QPoint newPosition) {
    Q_UNUSED(oldPosition)
    setData("screen-" + QString::number(screen->id()), I18N_NOOP("Position"), newPosition);
}

void ScreensEngine::screenAdded(Kephal::Screen * screen) {
    setData("screen-" + QString::number(screen->id()), I18N_NOOP("Id"), screen->id());
    screenResized(screen, screen->size(), screen->size());
    screenMoved(screen, screen->position(), screen->position());
}

void ScreensEngine::screenRemoved(int id) {
    removeAllData("screen-" + QString::number(id));
}

void ScreensEngine::dataForOutput(Kephal::Output * output) {
    outputDisconnected(output);
    if (output->isConnected()) {
        outputConnected(output);
        if (output->isActivated()) {
            outputActivated(output);
        } else {
            outputDeactivated(output);
        }
    }
}

void ScreensEngine::outputConnected(Kephal::Output * output) {
    setData("output-" + output->id(), I18N_NOOP("Connected"), true);
    
    QList<QSize> sizes = output->availableSizes();
    for (int i = 0; i < sizes.size(); ++i) {
        QSize size = sizes[i];
        setData("output-" + output->id(), I18N_NOOP("Size " + QString::number(i + 1)), size);
        setData("output-" + output->id(), I18N_NOOP("HasSize " + QString::number(size.width())) + "x" + QString::number(size.height()), true);
    }
    setData("output-" + output->id(), I18N_NOOP("NumSizes"), sizes.size());
    
    QList<QPoint> positions = output->availablePositions();
    for (int i = 0; i < positions.size(); ++i) {
        QPoint pos = positions[i];
        setData("output-" + output->id(), I18N_NOOP("Position " + QString::number(i + 1)), pos);
        setData("output-" + output->id(), I18N_NOOP("HasPosition " + QString::number(pos.x())) + "x" + QString::number(pos.y()), true);
    }
    setData("output-" + output->id(), I18N_NOOP("NumPositions"), positions.size());
}

void ScreensEngine::outputDisconnected(Kephal::Output * output) {
    removeAllData("output-" + output->id());
    setData("output-" + output->id(), I18N_NOOP("Id"), output->id());
    setData("output-" + output->id(), I18N_NOOP("Connected"), false);
}

void ScreensEngine::outputActivated(Kephal::Output * output) {
    setData("output-" + output->id(), I18N_NOOP("Activated"), true);
    outputResized(output, output->size(), output->size());
    outputMoved(output, output->position(), output->position());
    outputRateChanged(output, output->rate(), output->rate());
    outputRotated(output, output->rotation(), output->rotation());
    outputReflected(output, output->reflectX(), output->reflectY(), output->reflectX(), output->reflectY());
}

void ScreensEngine::outputDeactivated(Kephal::Output * output) {
    setData("output-" + output->id(), I18N_NOOP("Activated"), false);
}

void ScreensEngine::outputResized(Kephal::Output * output, QSize oldSize, QSize newSize) {
    Q_UNUSED(oldSize)
    setData("output-" + output->id(), I18N_NOOP("Size"), newSize);
    setData("output-" + output->id(), I18N_NOOP("Geometry"), output->geom());
}

void ScreensEngine::outputMoved(Kephal::Output * output, QPoint oldPosition, QPoint newPosition) {
    Q_UNUSED(oldPosition)
    setData("output-" + output->id(), I18N_NOOP("Position"), newPosition);
    setData("output-" + output->id(), I18N_NOOP("Geometry"), output->geom());
}

void ScreensEngine::outputRateChanged(Kephal::Output * output, float oldRate, float newRate) {
    Q_UNUSED(oldRate)
    setData("output-" + output->id(), I18N_NOOP("Refresh-Rate"), newRate);
}

void ScreensEngine::outputRotated(Kephal::Output * output, Kephal::Rotation oldRotation, Kephal::Rotation newRotation) {
    Q_UNUSED(oldRotation)
    setData("output-" + output->id(), I18N_NOOP("Rotation"), newRotation);
}

void ScreensEngine::outputReflected(Kephal::Output * output, bool oldX, bool oldY, bool newX, bool newY) {
    Q_UNUSED(oldX)
    Q_UNUSED(oldY)
    setData("output-" + output->id(), I18N_NOOP("Reflect-X"), newX);
    setData("output-" + output->id(), I18N_NOOP("Reflect-Y"), newY);
}

void ScreensEngine::confirmTimeout(int seconds) {
    setData("configurations", I18N_NOOP("Confirmation-Timeout"), seconds);
}

void ScreensEngine::confirmed() {
    setData("configurations", I18N_NOOP("Confirmation-Timeout"), 0);
}

void ScreensEngine::reverted() {
    setData("configurations", I18N_NOOP("Confirmation-Timeout"), 0);
}


K_EXPORT_PLASMA_DATAENGINE(screens, ScreensEngine)

