/*
  Copyright (c) 2010 by Drake Justice <djustice.kde@gmail.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
*/

#include "movingscene.h"

#include <QPainter>
#include <QFile>


K_EXPORT_PLASMA_WALLPAPER(movingscene, MovingScene)

QTime *m_time;

MovingScene::MovingScene(QObject *parent, const QVariantList &args)
    : Plasma::Wallpaper(parent, args)
{
    m_time = new QTime();
    m_time->start();
    m_cursor = new QCursor();
    m_timer = new QTimer();
    x_d = 0;
    y_d = 0;
    x = 0;
    y = 0;
    
    m_timer->start(100);
}

MovingScene::~MovingScene()
{
    delete m_cursor;
    delete m_timer;
}

void MovingScene::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
    if (event->buttons() == Qt::LeftButton) {
        event->accept();
        x = event->screenPos().x();
        y = event->screenPos().y();
        x_d = x / 4;
        y_d = y / 4;
        emit update(boundingRect());
    }
}

void MovingScene::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    if (event->buttons() == Qt::LeftButton) {
        event->accept();
        if (x_d > 0 && y_d > 0) {
            m_cursor->setPos(x_d*4, y_d*4);
            emit update(boundingRect());
        } else {
            m_cursor->setPos(boundingRect().width() / 2, boundingRect().height() / 2);
            emit update(boundingRect());
        }
    }
}

void MovingScene::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    event->accept();
    x = event->screenPos().x();
    y = event->screenPos().y();
    x_d = x / 4;
    y_d = y / 4;
    emit update(boundingRect());
}

QWidget* MovingScene::createConfigurationInterface(QWidget* parent)
{
    m_configWidget = new QWidget(parent);

    m_kurl = new KUrlRequester();
    m_moveWithMouseCheckBox = new QCheckBox();
    m_layout = new QVBoxLayout();
    m_layout->addWidget(m_kurl);
    m_layout->addWidget(m_moveWithMouseCheckBox);
    m_kurl->setText(m_wallpaper);
    m_moveWithMouseCheckBox->setText(i18n("Sync wallpaper with mouse movement"));
    m_moveWithMouseCheckBox->setChecked(m_moveWithMouse);
    m_configWidget->setLayout(m_layout);

    connect(m_kurl, SIGNAL(textChanged(QString)), this, SLOT(setWallpaper(QString)));
    connect(m_kurl, SIGNAL(returnPressed(QString)), this, SLOT(setWallpaper(QString)));
    connect(m_moveWithMouseCheckBox, SIGNAL(stateChanged(int)), this, SLOT(changeMouseBehavior(int)));
    connect(m_kurl, SIGNAL(urlSelected(KUrl)), this, SLOT(setWallpaper(KUrl)));

    connect(this, SIGNAL(settingsChanged(bool)), parent, SLOT(settingsChanged(bool)));
    
    return m_configWidget;
}

void MovingScene::save(KConfigGroup &config)
{
    config.writeEntry("wallpaper", m_wallpaper);
    config.writeEntry("movewithmouse", m_moveWithMouse);
    config.writeEntry("current_x", x_d);
    config.writeEntry("current_y", y_d);
}

void MovingScene::init(const KConfigGroup &config)
{
     m_wallpaper = config.readEntry("wallpaper", "/usr/share/wallpapers/Air/contents/images/1280x800.jpg");
     m_moveWithMouse = config.readEntry("movewithmouse", false);
     x_d = config.readEntry("current_x", 0);
     y_d = config.readEntry("current_y", 0);
     m_image = QImage(m_wallpaper);
     
     if(m_moveWithMouse) {
         connect(m_timer, SIGNAL(timeout()), this, SLOT(updateWallpaper()));
     } else {
         disconnect(m_timer, SIGNAL(timeout()), this, SLOT(updateWallpaper()));
     }
}

void MovingScene::setWallpaper(const QString& url)
{
    if(QFile::exists(url)) {
        m_wallpaper = url;
        emit configNeedsSaving();
        emit settingsChanged(true);
    }
}

void MovingScene::setWallpaper(const KUrl& url)
{
    setWallpaper(url.toLocalFile());
}

void MovingScene::paint(QPainter *painter, const QRectF& exposedRect)
{
    Q_UNUSED(exposedRect);

//    QTime cTime = QTime::currentTime();

    int expandW = boundingRect().width() / 3;
    int expandH = boundingRect().height() / 4;

    QRectF area(-x_d, -y_d, boundingRect().width() + expandW, boundingRect().height() + expandH);
    painter->drawImage(area, m_image.transformed(QTransform::fromTranslate(-x_d, -y_d)), boundingRect());
//    qDebug() << "paint needed " << cTime.msecsTo(QTime::currentTime())<<"ms";
}

void MovingScene::updateWallpaper()
{
    x = m_cursor->pos().x();
    y = m_cursor->pos().y();
    x_d = x / 4;
    y_d = y / 4;
    emit update(boundingRect());  
}

void MovingScene::changeMouseBehavior(int movewithmouse)
{
    if (movewithmouse) {
        m_moveWithMouse = true;
	emit configNeedsSaving();
	emit settingsChanged(true);
    } else {
        m_moveWithMouse = false;
	emit configNeedsSaving();
	emit settingsChanged(true);      
    }
}

#include "movingscene.moc"
