#include "koratbuttongroup.h"

#include "koratdefines.h"
#include "koratitem.h"
#include "koratbutton.h"

#include <KDebug>

#include <QList>
#include <QMap>
#include <QGraphicsLinearLayout>

namespace Korat
{

class ButtonGroupPrivate
{
public:
    ButtonGroupPrivate(ButtonGroup *buttonGroup)
        : q(buttonGroup)
        , largestSize(BoxButtonLarge)
        , minimumSize(BoxButtonSmall)
        , layout(new QGraphicsLinearLayout())
        , currentWeight(-1)
    {

    }
    
    void calculateWidths()
    {
        widths.clear();

        widths << 0 << 0 << 0;
        foreach (Button *button, buttonList) {
            qreal smallWidth;
            qreal smallWithTextWidth;
            qreal largeWidth;

            WeightList weightList;

            weightList << 0;

            weightList[0] = (int)BoxButtonSmall;
            smallWidth = button->getWidth(weightList);

            weightList[0] = (int)BoxButtonSmallWithText;
            smallWithTextWidth = button->getWidth(weightList);

            weightList[0] = (int)BoxButtonLarge;
            largeWidth = button->getWidth(weightList);

            widths[0] = qMax(widths[0],smallWidth);
            widths[1] = qMax(widths[1],smallWithTextWidth);
            widths[2] += largeWidth;
        }

//        kDebug() << widths;
    }

    void sizeUpdate()
    {
        calculateWidths();
        emit q->sizeChanged();
    }

    QList<Button *> buttonList;
    ButtonGroup *q;

    LayoutWeight largestSize;
    LayoutWeight minimumSize;

    QGraphicsLinearLayout *layout;

    QList<qreal> widths;
    int currentWeight;
};

ButtonGroup::ButtonGroup(QGraphicsItem *parent)
    : Item(parent)
    , d(new ButtonGroupPrivate(this))
{
    //setSizePolicy(QSizePolicy::Fixed,QSizePolicy::Fixed);
    setMinimumHeight(LARGEST_HEIGHT);
    setMaximumHeight(LARGEST_HEIGHT);
    setContentsMargins(0,0,0,0);
    d->layout->setSpacing(0);
    d->layout->setContentsMargins(0,0,0,0);
    setLayout(d->layout);
}

ButtonGroup::~ButtonGroup()
{
    delete d;
    d = 0;
}

bool ButtonGroup::isContainer()
{
    return true;
}

QList<WeightList> ButtonGroup::availableWidths()
{
    QList<WeightList> result;
    WeightList weightList;

    weightList.clear();
    result.clear();

    if ((BoxButtonLarge >= d->minimumSize) && (BoxButtonLarge <= d->largestSize)) {
        weightList << (int)BoxButtonLarge;
    }
    if ((BoxButtonSmallWithText >= d->minimumSize) && (BoxButtonSmallWithText <= d->largestSize)) {
        weightList << (int)BoxButtonSmallWithText;
    }
    if ((BoxButtonSmall >= d->minimumSize) && (BoxButtonSmall <= d->largestSize)) {
        weightList << (int)BoxButtonSmall;
    }

    result << weightList;

    //kDebug() << result;
    return result;
}

qreal ButtonGroup::getWidth(const WeightList &weightList)
{
    qreal width = 0;
    d->calculateWidths();
    switch (weightList.at(0)) {
    case BoxButtonSmall:
        width = d->widths[0];
        break;
    case BoxButtonSmallWithText:
        width = d->widths[1];
        break;
    case BoxButtonLarge:
        width = d->widths[2];
        break;
    default:
        kWarning() << "Trouble:" << weightList;
    }
    //kDebug() << weightList;
    return width;
}

void ButtonGroup::adjustWidth(const WeightList &weightList)
{
    setMinimumWidth(getWidth(weightList));
    setMaximumWidth(getWidth(weightList));
    foreach (Button *button, d->buttonList) {
        button->adjustWidth(weightList);
    }
    switch (weightList.at(0)) {
    case BoxButtonSmall:
        if (d->layout->orientation() != Qt::Vertical) {
            d->layout->setOrientation(Qt::Vertical);
        }
        break;
    case BoxButtonSmallWithText:
        if (d->layout->orientation() != Qt::Vertical) {
            d->layout->setOrientation(Qt::Vertical);
        }
        break;
    case BoxButtonLarge:
        if (d->layout->orientation() != Qt::Horizontal) {
            d->layout->setOrientation(Qt::Horizontal);
        }
        break;
    default:
        kWarning() << "Trouble:" << weightList;
    }
}

void ButtonGroup::setLargestSizeMode(LayoutWeight mode)
{
    d->largestSize = mode;
}

void ButtonGroup::setMinimumSizeMode(LayoutWeight mode)
{
    d->minimumSize = mode;
}

void ButtonGroup::setFixedSizeMode(LayoutWeight mode)
{
    d->largestSize = d->minimumSize = mode;
}

void ButtonGroup::addButton(Button *button)
{
    // we allow maximum 3 icons in one column
    if (d->buttonList.size() < 3) {
        d->buttonList << button;
        connect(button,SIGNAL(sizeChanged()),this,SLOT(sizeUpdate()));
        d->layout->addItem(button);
//        d->layout->setAlignment(button,Qt::AlignCenter);
        d->sizeUpdate();
    }
}

void ButtonGroup::removeButton(int index)
{
    if ((index >= 0) && (index < d->buttonList.size())) {
        Button *button = d->buttonList.takeAt(index);
        disconnect(button,SIGNAL(sizeChanged()),this,SLOT(sizeUpdate()));
        d->layout->removeAt(index);
        d->sizeUpdate();
    }
}

void ButtonGroup::removeButton(Button *button)
{
    removeButton(d->buttonList.indexOf(button));
}

Button *ButtonGroup::getButton(int index)
{
    return d->buttonList.at(index);
}

} // namespace Korat

#include "koratbuttongroup.moc"
// vim: sw=4 sts=4 et tw=100
