/*
 *   Copyright (C) 2007  Jon de Andres Frias <jondeandres@gmail.com>
 *   Copyright (C) 2007  Christopher Blauvelt <cblauvelt@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 or later as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <QPoint>
#include <QPainter>
#include <QItemDelegate>

#include <kdebug.h>

#include "listview.h"
#include "listview_p.h"


namespace Plasma
{
    ListView::Private::Private() { }

    ListView::Private::~Private() {}

    ListView::ListView(Widget *parent)
        : Plasma::Widget(parent),
          d(new Private())
    {

        setItemDelegate(new QItemDelegate(0));
        setFlag(QGraphicsItem::ItemIsFocusable);

        d->model = NULL;
        d->firstItem = -1;
        d->lastItem = -1;
        d->offset = 0;
        d->scrollPosition = 0;

        d->verticalScrollBar = new Plasma::ScrollBar(0);
        d->verticalScrollBar->setRange(0, 0);
        d->verticalScrollBar->setRange(0, 100);
        d->verticalScrollBar->setValue(40);
        d->verticalScrollBar->setGeometry(QRectF(0, 0, 40, 300));

//         connect(d->verticalScrollBar, SIGNAL(valueChanged(int)),
//                 this, SLOT(scrollChanged(int)));
    }

    ListView::~ListView() { }

    void ListView::setModel(QAbstractItemModel *model)
    {
        //no model change.
        if (model == d->model)
            return;

        //if model is different then disconnect the signals.
        if (d->model) {
        disconnect(d->model, SIGNAL(dataChanged(QModelIndex, QModelIndex)),
                   this, SLOT(dataChanged(QModelIndex, QModelIndex)));
        }

        d->model = model;

        //sannity checks
        Q_ASSERT_X(d->model->index(0,0) == d->model->index(0,0),
                   "QAbstractItemView::setModel",
                   "A model should return the exact same index "
                   "(including its internal id/pointer) when asked for it twice in a row.");
        Q_ASSERT_X(d->model->index(0,0).parent() == QModelIndex(),
                   "QAbstractItemView::setModel",
                   "The parent of a top level index should be invalid");

        if (d->model) {
        connect(d->model, SIGNAL(dataChanged(QModelIndex, QModelIndex)),
                   this, SLOT(dataChanged(QModelIndex, QModelIndex)));
        }

        setSelectionModel(new QItemSelectionModel(d->model));
        if(d->selectionModel){
            connect(d->selectionModel, SIGNAL(selectionChanged(const QItemSelection&, const QItemSelection&)),
                    this, SLOT(selectionChanged(const QItemSelection&, const QItemSelection&)));
        }
        update();
    }

    void ListView::initPaintRange()
    {
        if (d->model == NULL) {
            return;
        }

        int modelItems = d->model->rowCount();
        QSizeF size = geometry().size();
        QSizeF actualSize(0, 0);

        for (int i = 0; i < modelItems; i++) {
            QStyleOptionViewItem options;
            options.decorationSize = QSize(64, 64);

            QModelIndex index = d->model->index(i, 0);
            QSize itemSize = d->delegate->sizeHint(options, index);

            if (i == 0) {
                d->firstItem = 0;
                d->firstItemOffset = itemSize.height();
            }

            if (actualSize.height() > size.height()) {
                break;
            }
            else{
                actualSize.setHeight(actualSize.height() + itemSize.height());
                d->lastItem = index.row();
                if (actualSize.height() > size.height()) {
                    d->lastItemOffset = itemSize.height() - (actualSize.height() - size.height());
                }
                else{
                    d->lastItemOffset = itemSize.height();
                }
            }
        }

    }

    QAbstractItemModel * ListView::model()
    {
        return d->model;
    }

    void ListView::setItemDelegate(QAbstractItemDelegate *delegate)
    {
        d->delegate = delegate;
    }

    QAbstractItemDelegate * ListView::delegate()
    {
        return d->delegate;
    }

    void ListView::toUpdate()
    {
        initPaintRange();
        update();
    }

    void ListView::dataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight)
    {
        Q_UNUSED(topLeft)
        Q_UNUSED(bottomRight)
            initPaintRange();
        update();
    }

    void ListView::selectionChanged(const QItemSelection &topLeft, const QItemSelection &bottomRight)
    {
        Q_UNUSED(topLeft)
        Q_UNUSED(bottomRight)
        update();
    }

    void ListView::scrollChanged(int value)
    {
        kDebug() << "scrolling!! " << value << endl;

        if (value > d->scrollPosition) {
            scrollDown(value - d->scrollPosition);
        }
        else if (value < d->scrollPosition){
            scrollUp(d->scrollPosition - value);
        }

        d->scrollPosition = value;


    }

    void ListView::scrollUp(int value)
    {
        do{
            if ((d->firstItem == 0) &&
                d->offset == 0) {
                return;
            }

//            d->lastItemOffset--;
            d->lastItemOffset -= 20;

            QStyleOptionViewItem options;
            options.decorationSize = QSize(64, 64);

            QModelIndex index = d->model->index(d->lastItem, 0);
            QSize itemSize = d->delegate->sizeHint(options, index);

            if (d->lastItemOffset < 0) {
                d->lastItem--;
                d->lastItemOffset = d->delegate->sizeHint(options, d->model->index(d->lastItem, 0)).height();;
            }

//            d->firstItemOffset++;
            d->firstItemOffset += 20;

            index = d->model->index(d->firstItem, 0);
            itemSize = d->delegate->sizeHint(options, index);

            if (d->firstItemOffset > itemSize.height()) {
                d->firstItem--;
                d->firstItemOffset = 0;
                d->offset = -(d->delegate->sizeHint(options, d->model->index(d->firstItem, 0)).height());
            }
            else{
//                d->offset++;
                d->offset += 20;
            }

        } while (value-=20 > 0);
            update();
//         if (value-- >= 0) {
//             scrollUp(value);
//         }


        kDebug() << "" << endl;
//        kDebug() << "firstItemOffset: " << d->firstItemOffset << " lastItemOffset: " << d->lastItemOffset << " offset: " << d->offset <<  endl;
    }

    void ListView::scrollDown(int value)
    {
        do {
//            d->lastItemOffset++;
            d->lastItemOffset += 20;

            QStyleOptionViewItem options;
            options.decorationSize = QSize(64, 64);

            QModelIndex index = d->model->index(d->lastItem, 0);
            QSize itemSize = d->delegate->sizeHint(options, index);

            if (d->lastItem == (d->model->rowCount() -1 ) &&
                d->lastItemOffset >= itemSize.height()) {
                return;
            }

            if (d->lastItemOffset > itemSize.height()) {
                d->lastItem++;
                d->lastItemOffset = 0;
            }

//            d->firstItemOffset--;
            d->firstItemOffset -= 20;

            index = d->model->index(d->firstItem, 0);
            itemSize = d->delegate->sizeHint(options, index);

            if (d->firstItemOffset < 0) {
                d->firstItem++;
                d->firstItemOffset = d->delegate->sizeHint(options, d->model->index(d->firstItem, 0)).height();
                d->offset = 0;
            }
            else{
//                d->offset--;
                d->offset -= 20;
            }
//         if (value-- >= 0) {
//             scrollDown(value);

            kDebug() << "" << endl;

//        kDebug() << "firstItemOffset: " << d->firstItemOffset << " lastItemOffset: " << d->lastItemOffset << " offset: " << d->offset <<  endl;

    }while (value-=20 > 0);
            update();

//     QSizeF ListView::sizeHint() const
//     {
//         return QSizeF(400, 400);

//     }
    }

    QRectF ListView::boundingRect() const
    {
        return QRectF(QPoint(0, 0), geometry().size());
    }

    void ListView::mousePressEvent(QGraphicsSceneMouseEvent *event)
    {
        event->accept();
//        kDebug() << "mouse pressed event: "<< event->pos() << endl;
    }

    void ListView::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
    {
        event->accept();
//        kDebug() << "mouse pressed event: " << event->pos() << endl;
        QModelIndex clickedIndex = indexAt(event->pos());
        if(event->button() & Qt::LeftButton){      //click is a selection
            //clear selection but retain the current selection
            if(!clickedIndex.isValid()){
                d->selectionModel->clearSelection();
                update();
                return;
            }

            QItemSelectionModel::SelectionFlags command = QItemSelectionModel::ClearAndSelect;

            if(event->modifiers() & Qt::ControlModifier){
                command = QItemSelectionModel::Select;
                if(d->selectionModel->isSelected(clickedIndex)){
                    command |= QItemSelectionModel::Toggle;
                }
            }

            if(event->modifiers() & Qt::ShiftModifier){     //click is setting a range

                d->selectionModel->select(QItemSelection(clickedIndex,d->selectionModel->currentIndex()),command);
            }
            else {      //if there is no keyboard modifier, this is a new selection
                d->selectionModel->select(clickedIndex,command);
            }
        }
        //update current index
        d->selectionModel->setCurrentIndex(clickedIndex, QItemSelectionModel::NoUpdate);
    }

    void ListView::keyPressEvent(QKeyEvent *event)
    {
        event->accept();
        kDebug() << "keypress event: " << event->key() << endl;
        kDebug() << "keypress event: " << event->text() << endl;

        //get current selection, if no current selection select item (0,0)
        QModelIndex currentIndex = d->selectionModel->currentIndex();
        if(!currentIndex.isValid()){
            currentIndex = d->model->index(0,0);
            if(!currentIndex.isValid()){
                kDebug() << "Error!  A default selection could not be created." << endl;
                return;
            }
        }

        //handle arrow keys
        if(event->key() & (Qt::Key_Left | Qt::Key_Up | Qt::Key_Right | Qt::Key_Down) != 0){
            //find QModelIndex in direction of key
            QModelIndex newIndex;
            switch (event->key()) {
                case Qt::Key_Left:
                    newIndex = d->model->index(currentIndex.row(),currentIndex.column()-1);
                    break;
                case Qt::Key_Up:
                    newIndex = d->model->index(currentIndex.row()-1,currentIndex.column());
                    break;
                case Qt::Key_Right:
                    newIndex = d->model->index(currentIndex.row(),currentIndex.column()+1);
                    break;
                case Qt::Key_Down:
                    newIndex = d->model->index(currentIndex.row()+1,currentIndex.column());
                    break;
            }
            if(!newIndex.isValid()){
                return;
            }
            d->selectionModel->setCurrentIndex(newIndex,QItemSelectionModel::ClearAndSelect);
        }
        else if(event->text() != QString()) {
            //find the first QModelIndex whose text starts with the pressed letter.
            return;
        }
    }

    void ListView::paintWidget(QPainter *painter,  const QStyleOptionGraphicsItem *option,  QWidget *widget)
    {
         Q_UNUSED(option)
         Q_UNUSED(widget)

         QPalette::ColorGroup cg;
         cg = QPalette::Normal;

//         d->verticalScrollBar->update();
//         d->verticalScrollBar->paintWidget(painter, option, widget);

         if (d->firstItem == -1 || d->lastItem == -1) {
             return;
         }

//         painter->fillRect(QRectF(QPoint(0, 0), QSizeF(geometry().size())), Qt::red);
         kDebug() << "Painted items: " << d->firstItem << " to " << d->lastItem << endl;
         for ( int i = d->firstItem; i <= d->lastItem; i++) {

             QStyleOptionViewItem options;

             if (!d->model->hasIndex(i, 0)) {
                 continue;
            }

            QModelIndex index = d->model->index(i, 0);
            options.decorationSize = QSize(64, 64);
            options.rect = QRect(QPoint(40, d->offset + (i - d->firstItem)*64), QSize(256, 256));
            options.font = font();
            options.state |= QStyle::State_Enabled;

            if(d->selectionModel->isSelected(index)){
                options.state |= QStyle::State_Selected;
            }

            options.palette.setCurrentColorGroup(cg);
            options.textElideMode = Qt::ElideNone;

            d->delegate->paint(painter,  options, index);
        }
    }

    void ListView::setSelectionModel(QItemSelectionModel *selectionModel)
    {
        d->selectionModel = selectionModel;
    }

    QItemSelectionModel* ListView::selectionModel() const
    {
        return d->selectionModel;
    }

    QModelIndex ListView::indexAt(const QPointF &p) const
    {
        QStyleOptionViewItem options;
        options.decorationSize = QSize(64, 64);
        QSize firstItemSize = d->delegate->sizeHint(options, d->model->index(d->firstItem, 0));

        int row = d->firstItem + ((int)p.y() + firstItemSize.height() - d->firstItemOffset)/64; //FIXME: the divisor should not be hardcoded.  It should be the height of the items.
        QModelIndex index = d->model->index(row,0);
        if(index.isValid()){
            return index;
        }
        return QModelIndex();
    }
}



#include "listview.moc"
