#include "nodeitem.h"

#include "edge.h"

#include "composition.h"
#include "node.h"
#include "layer.h"

#include <QGraphicsScene>
#include <QGraphicsSceneContextMenuEvent>
#include <QMenu>
#include <QAction>
#include <QDebug>

NodeItem::NodeItem()
{
    m_rect = QRectF(0, 0, 160, 100);
    setFlag(QGraphicsItem::ItemIsMovable, true);
    setFlag(QGraphicsItem::ItemIsSelectable, true);
    //setBackgroundColor(QColor(193, 193, 193, 127));
    setBackgroundColor(Qt::white);

    setZValue(10);
}

static inline QPainterPath headerPath(const QRectF &r, int roundness,
                                      int headerHeight=10)
{
    QPainterPath path;
    int xRnd = roundness;
    int yRnd = roundness;
    if (r.width() > r.height())
        xRnd = int(roundness * r.height()/r.width());
    else
        yRnd = int(roundness * r.width()/r.height());

    if(xRnd >= 100)                          // fix ranges
        xRnd = 99;
    if(yRnd >= 100)
        yRnd = 99;
    if(xRnd <= 0 || yRnd <= 0) {             // add normal rectangle
        path.addRect(r);
        return path;
    }

    QRectF rect = r.normalized();

    if (rect.isNull())
        return path;

    qreal x = rect.x();
    qreal y = rect.y();
    qreal w = rect.width();
    qreal h = rect.height();
    qreal rxx = w*xRnd/200;
    qreal ryy = h*yRnd/200;
    // were there overflows?
    if (rxx < 0)
        rxx = w/200*xRnd;
    if (ryy < 0)
        ryy = h/200*yRnd;
    qreal rxx2 = 2*rxx;
    qreal ryy2 = 2*ryy;

    path.arcMoveTo(x, y, rxx2, ryy2, 90);
    path.arcTo(x,        y, rxx2, ryy2, 90, 90);
    QPointF pt = path.currentPosition();
    path.lineTo(x, pt.y()+headerHeight);
    path.lineTo(x+w, pt.y()+headerHeight);
    path.lineTo(x+w, pt.y());
    path.arcTo(x+w-rxx2, y, rxx2, ryy2, 0, 90);
    path.closeSubpath();

    return path;
}


void NodeItem::createBoxConnections(QPainterPath &path,
                                    const QList<Quasar::Property> &props)
{
    QList<Quasar::Property> inputs;
    QList<Quasar::Property> outputs;

    const int conWidth = 10;
    const int xOffset  = 7;

    foreach(Quasar::Property prop, props) {
        if (prop.kind == Quasar::Property::InputProperty)
            inputs.append(prop);
        else
            outputs.append(prop);
    }

    int i = 0;
    foreach(Quasar::Property prop, inputs) {
        QRectF rect(xOffset,
                    conWidth + 20 * (i+1),
                    conWidth, conWidth);
        path.addEllipse(rect);
        m_propLocations.insert(prop.name, rect);
        ++i;
    }
    i = 0;
    foreach(Quasar::Property prop, outputs) {
        QRectF rect(m_rect.right() - conWidth - xOffset,
                    conWidth + 20 * (i+1),
                    conWidth, conWidth);
        path.addEllipse(rect);
        m_propLocations.insert(prop.name, rect);
        ++i;
    }
}

void NodeItem::paint(QPainter *p, const QStyleOptionGraphicsItem *,
                     QWidget *)
{
    QFont font("ComicSans", 10);
    const QSize  bsize(160, 100);

    QList<Quasar::Property> pargs = m_node->properties();
    QRectF boxRect = m_rect;

    QPainterPath header = headerPath(boxRect, 25);

    p->save();
    p->setRenderHint(QPainter::Antialiasing);
    p->setRenderHint(QPainter::SmoothPixmapTransform);

    p->setPen(Qt::black);
    p->setBrush(QBrush(Qt::white));
    p->drawPath(m_shape);

    p->save();
    p->setPen(QPen(Qt::NoPen));
    p->setBrush(QColor(120, 255, 185));
    p->drawPath(header);
    p->restore();

    p->setPen(QPen(Qt::black));
    font.setBold(true);
    p->setFont(font);
    p->drawText(QPointF(boxRect.x()+10, boxRect.y()+15),
                m_node->name());

    p->setPen(QPen(Qt::black));
    font.setPointSize(8);
    font.setBold(false);
    p->setFont(font);

    QList<Quasar::Property>::const_iterator itr;
    QFontMetrics fm(font);
    for (itr = pargs.constBegin(); itr != pargs.constEnd(); ++itr) {
        QRectF rect = m_propLocations[itr->name];
        if (itr->kind == Quasar::Property::InputProperty) {
            if (m_node->inputNodes().contains(itr->name)) {
                p->save();
                p->setBrush(Qt::yellow);
                p->drawEllipse(propertyRect(itr->name));
                p->restore();
            }
            p->drawText(QPointF(rect.right() + 5,
                                rect.bottom() - fm.descent()/4),
                        itr->name);
        } else {
            if (m_node->outputNodes().contains(itr->name)) {
                p->save();
                p->setBrush(Qt::yellow);
                p->drawEllipse(propertyRect(itr->name));
                p->restore();
            }
            p->drawText(QPointF(
                            rect.left() - 5 - fm.width(itr->name),
                            rect.bottom() - fm.descent()/4),
                        itr->name);
        }
    }
    p->restore();
}


QRectF NodeItem::boundingRect() const
{
    return m_rect;
}

void NodeItem::setNode(Quasar::Node *node)
{
    m_node = node;
    updateSize();
    update();
}

void NodeItem::contextMenuEvent(QGraphicsSceneContextMenuEvent *)
{
    QMenu menu;
    menu.setTitle(QLatin1String("Filter Options"));
    QAction *action = new QAction(QLatin1String("Remove Filter"), &menu);
    connect(action, SIGNAL(triggered()),
            this, SLOT(deleteLater()));
    menu.addAction(action);
    menu.exec(QCursor::pos());
}

NodeItem::~NodeItem()
{
    /*
    EdgeSource *prev = 0;
    EdgeSource *next = 0;
    if (m_destEdge)
        prev = m_destEdge->sourceNode();
    if (m_sourceEdge)
        next = m_sourceEdge->destNode();
    delete m_sourceEdge;
    delete m_destEdge;

    if (prev && next) {
        Edge *e = new Edge(prev, next);
        scene()->addItem(e);
        e->show();
    }
    */
}

void NodeItem::updateSize()
{
    QFontMetrics fm(QFont("ComicSans", 10));
    m_rect.setWidth(qMax(150, fm.width(m_node->name()) + 15*2));

    QList<Quasar::Property> pargs = m_node->properties();
    m_rect.setHeight(qMax(100, (pargs.count() * 12) + 35)); // too many hardcoded values. =/

    m_baseShape = QPainterPath();
    m_baseShape.addRoundRect(m_rect, 25);
    m_shape = m_baseShape;
    createBoxConnections(m_shape, pargs);
    m_shape.setFillRule(Qt::OddEvenFill);
}

void NodeItem::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    m_clickedProperty = propertyAt(event->pos());

    if (m_clickedProperty.isEmpty())
        return QGraphicsItem::mousePressEvent(event);
}

void NodeItem::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    if (m_clickedProperty.isEmpty())
        QGraphicsItem::mouseReleaseEvent(event);
    m_clickedProperty = QString();
}

QString NodeItem::propertyAt(const QPointF &pt) const
{
    QMap<QString, QRectF>::const_iterator itr;

    for (itr = m_propLocations.constBegin();
         itr != m_propLocations.constEnd();
         ++itr) {
        if (itr.value().contains(pt))
            return itr.key();
    }
    return QString();
}

QPainterPath NodeItem::shape() const
{
    return m_baseShape;
}

QRectF NodeItem::propertyRect(const QString &prop) const
{
    return m_propLocations[prop];
}

void NodeItem::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
    if (m_clickedProperty.isEmpty())
        QGraphicsItem::mouseMoveEvent(event);
}

#include "moc_nodeitem.cpp"
