#define GL_GLEXT_PROTOTYPES
#include "mesh.h"

#include "texture.h"

#include <GL/glext.h>
#include <GL/gl.h>

#include <QVector>
#include <QList>
#include <QDebug>

namespace Quasar
{

static inline float mapVal(float cur, float min, float max)
{
    Q_UNUSED(min);
    return cur * max;
}

struct Vertex {
    Vertex()
        :x(0), y(0), z(0)
    {}
    Vertex(float ax, float ay, float az)
        : x(ax), y(ay), z(az)
    {}
    Vertex(const Vertex &other)
        : x(other.x), y(other.y), z(other.z)
    {
    }
    float x;
    float y;
    float z;
};
struct TexCoord {
    TexCoord()
        :s(0), t(0)
    {}
    TexCoord(float as, float at)
        : s(as), t(at)
    {}
    TexCoord(const TexCoord &other)
        : s(other.s), t(other.t)
    {
    }
    float s;
    float t;
};
struct VertexColor {
    VertexColor()
        :r(0), g(0), b(0)
    {}
    VertexColor(GLubyte ar, GLubyte ag, GLubyte ab)
        : r(ar), g(ag), b(ab)
    {}
    VertexColor(const VertexColor &other)
        : r(other.r), g(other.g), b(other.b)
    {
    }
    GLubyte r;
    GLubyte g;
    GLubyte b;
};

class Mesh::Private
{
public:
    Private()
        : maxMesh(2)
    {}

    void createMesh(int minx, int maxx, int miny, int maxy)
    {
        int k, i, swap = 0;
        const int end_i = maxMesh - 1;
        const int end_k = maxMesh - 1;
        const float oneOverEnd_i = 1.0 / end_i;
        const float oneOverEnd_k = 1.0 / end_k;

        for (k = 0; k < end_k; k++) {
            if (swap) {
                for (i = end_i; i >= 0; i--) {
                    const float xpercent   = i * oneOverEnd_i;
                    const float ypercent   = k * oneOverEnd_k;
                    const float ypercentp1 = (k + 1) * oneOverEnd_k;
                    const float ri  = mapVal(xpercent, minx, maxx);
                    const float rk  = mapVal(ypercent, miny, maxy);
                    const float rk1 = mapVal(ypercentp1, miny, maxy);
                    colors.append(VertexColor(255, 0, 0));
                    texCoords.append(TexCoord(xpercent, ypercent));
                    vertices.append(Vertex(ri, rk, 0));
                    colors.append(VertexColor(0, 0, 255));
                    texCoords.append(TexCoord(xpercent, ypercentp1));
                    vertices.append(Vertex(ri, rk1, 0));
                    if (i == 0) {  /* degenerate */
                        colors.append(VertexColor(0, 0, 255));
                        texCoords.append(TexCoord(xpercent, ypercentp1));
                        vertices.append(Vertex(ri, rk1, 0));
                    }
                }
            } else {
                for (i = 0; i < maxMesh; i++) {
                    const float xpercent   = i * oneOverEnd_i;
                    const float ypercent   = k * oneOverEnd_k;
                    const float ypercentp1 = (k + 1) * oneOverEnd_k;
                    const float ri  = mapVal(xpercent, minx, maxx);
                    const float rk  = mapVal(ypercent, miny, maxy);
                    const float rk1 = mapVal(ypercentp1, miny, maxy);
                    colors.append(VertexColor(255, 0, 0));
                    texCoords.append(TexCoord(xpercent, ypercent));
                    vertices.append(Vertex(ri, rk, 0));
                    colors.append(VertexColor(0, 0, 255));
                    texCoords.append(TexCoord(xpercent, ypercentp1));
                    vertices.append(Vertex(ri, rk1, 0));
                    if (i == maxMesh-1) { /* degenerate */
                        colors.append(VertexColor(0, 0, 255));
                        texCoords.append(TexCoord(xpercent, ypercentp1));
                        vertices.append(Vertex(ri, rk1, 0));
                    }
                }
            }
            swap ^= 1;
        }
    }

    void createVbos()
    {
        QVector<Vertex> vertData   = vertices.toVector();
        QVector<VertexColor> clrData    = colors.toVector();
        QVector<TexCoord> texMapData = texCoords.toVector();

        glTexParameteri(GL_TEXTURE_2D,
                        GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D,
                        GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_COLOR_ARRAY);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glGenBuffersARB(3, vbos);

        glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[0]);
        glBufferDataARB(GL_ARRAY_BUFFER_ARB,
                        sizeof(Vertex) * vertData.size(),
                        vertData.data(), GL_STATIC_DRAW_ARB);
        glVertexPointer(3, GL_FLOAT, 0, 0);

        glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[1]);
        glBufferDataARB(GL_ARRAY_BUFFER_ARB,
                        sizeof(VertexColor) * clrData.size(),
                        clrData.data(), GL_STATIC_DRAW_ARB);
        glColorPointer(3, GL_UNSIGNED_BYTE, 0, 0);

        glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbos[2]);
        glBufferDataARB(GL_ARRAY_BUFFER_ARB,
                        sizeof(TexCoord) * texMapData.size(),
                        texMapData.data(), GL_STATIC_DRAW_ARB);
        glTexCoordPointer(2, GL_FLOAT, 0, 0);
    }

    int maxMesh;
    Texture texture;
    QList<Vertex>      vertices;
    QList<TexCoord>    texCoords;
    QList<VertexColor> colors;

    GLuint  vbos[3];
};

Mesh::Mesh(const Texture &tex, int size)
    : d(new Private)
{
    d->maxMesh = size;
    d->texture = tex;

    d->createMesh(0, tex.width(), 0, tex.height());
    d->createVbos();
}

int Mesh::size() const
{
    return d->maxMesh;
}

Texture Mesh::texture() const
{
    return d->texture;
}

void Mesh::render(const QPointF &pt)
{
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, d->vbos[0]);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, d->vbos[1]);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, d->vbos[2]);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, d->texture.id());
    glPushMatrix();
    glTranslatef(pt.x(), pt.y(), 0);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, d->vertices.size());
    glPopMatrix();
    glDisable(GL_TEXTURE_2D);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}


Mesh::~Mesh()
{
    glDeleteBuffers(3, d->vbos);
}

}

