/*
    KUserFace - advanced user properties
    Copyright (C) 2005 Friedrich W. H. Kossebau <Friedrich.W.H@Kossebau.de>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License version 2 as published by the Free Software Foundation.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301  USA
*/

#include <kdebug.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <unistd.h>
#include <stdlib.h>
#include <grp.h>

#include <qimage.h>
#include <qpixmap.h>
#include <qiconset.h>
#include <qvaluelist.h>
#include <qmap.h>
#include <qstringlist.h>
#include <qdragobject.h>

#include "kstaticdeleter.h"
#include "kstringhandler.h"
#include "kstandarddirs.h"
#include "kiconloader.h"
#include "kiconeffect.h"

#include "kuserface.h"



class KUserFaceLoaderPrivate {
  public:
    KUserFaceLoaderPrivate() {}

  public:
    QImage defaultFace;
    QString facesDir;
};

static KUserFaceLoader* loader = 0;
static KStaticDeleter<KUserFaceLoader> userFaceLoaderDeleter;

KUserFaceLoader::KUserFaceLoader() : d( new KUserFaceLoaderPrivate() ) {
  d->facesDir = KGlobal::dirs()->resourceDirs("data").last() + "kdm/faces/";
  d->defaultFace = QImage( d->facesDir + ".default.face.icon" );
}

KUserFaceLoader::~KUserFaceLoader() {
  delete d;
}

KUserFaceLoader *KUserFaceLoader::self() {
  if( loader == 0 )
    userFaceLoaderDeleter.setObject( loader, new KUserFaceLoader() );
  return loader;
}

QImage KUserFaceLoader::loadFace(const KUser &user, int size,
				 KUserFaceLoader::Policy policy) {
  QImage face = QImage( d->facesDir + user.loginName() + ".face.icon" );

  // not found?
  if( face.isNull() )
      face = d->defaultFace;

  // resize?
  if( size > 0 )
    face = face.smoothScale( size, size );

  return face;
}

QPixmap KUserFaceLoader::loadFaceIcon(const KUser &user, KIcon::Group group,
				      int size, int state, KUserFaceLoader::Policy policy) {
  KIconLoader *iconLoader = KGlobal::iconLoader();
  // adapt size
  if( size == 0 )
    size = iconLoader->currentSize(group);

  // get face
  QImage face = loadFace( user, size, policy );

  // apply effect
  KIconEffect *effect = iconLoader->iconEffect();
  if( effect )
    face = effect->apply( face, group, state );

  // create icon
  QPixmap icon;
  icon.convertFromImage(face);

  return icon;
}


class KUserFaceFactory : public QIconFactory
{
  public:
    KUserFaceFactory( const KUser &user, KIcon::Group g, int s, KUserFaceLoader::Policy p );
  public: // QIconFactory API
    virtual QPixmap* createPixmap( const QIconSet&, QIconSet::Size, QIconSet::Mode, QIconSet::State );
  private:
    KUser user;
    KIcon::Group group;
    int size;
    KUserFaceLoader::Policy policy;
};

QIconSet KUserFaceLoader::loadFaceIconSet(const KUser &user, KIcon::Group group, int size, KUserFaceLoader::Policy policy)
{
    QIconSet iconSet;
    iconSet.installIconFactory( new KUserFaceFactory(user,group,size,policy) );
    return iconSet;
}


KUserFaceFactory::KUserFaceFactory( const KUser &u, KIcon::Group g, int s, KUserFaceLoader::Policy p )
 : user( u ), group( g ), size( s ), policy( p ) {
    setAutoDelete( true );
}

QPixmap* KUserFaceFactory::createPixmap( const QIconSet&, QIconSet::Size, QIconSet::Mode mode, QIconSet::State ) {
    // QIconSet::Mode to KIcon::State conversion
    static const KIcon::States iconStates[] = { KIcon::DefaultState, KIcon::DisabledState, KIcon::ActiveState };
    int state = KIcon::DefaultState;
    if( mode <= QIconSet::Active )
        state = iconStates[ mode ];

    // return icon with effect
    return new QPixmap( KUserFaceLoader::self()->loadFaceIcon(user, group, size, state, policy) );
}



static const QString UserFaceProtocol = QString::fromLatin1( "kuser-face" );

class KUserMimeSourceFactory : public QMimeSourceFactory
{
public:
  const QMimeSource *data( const QString &AbsName ) const;
};

const QMimeSource *KUserMimeSourceFactory::data( const QString &AbsName ) const
{
  // extract and decode arguments
  QStringList Arguments = QStringList::split( QChar(':'), AbsName );
  QImageDrag *Source = 0;
  if ( Arguments[0] == UserFaceProtocol )
  {
    if ( Arguments.size() >= 3 )
    {
      KUser User( Arguments[1] );
      if( User.isValid() )
      {
        Source = new QImageDrag( KUserFaceLoader::self()->loadFace(User,Arguments[2].toInt()) );
      }
    }
    else
      kdDebug() << UserFaceProtocol << ": insufficient information in AbsName: " << Arguments << endl;
  }

  return Source;
}

class KUserRichTexter::KPrivate { };

static KUserMimeSourceFactory *userMimeSourceFactory = 0;

static KUserRichTexter* texter = 0;
static KStaticDeleter<KUserRichTexter> userRichTexterDeleter;

KUserRichTexter *KUserRichTexter::self() {
  if( texter == 0 )
    userRichTexterDeleter.setObject( texter, new KUserRichTexter() );
  return texter;
}

KUserRichTexter::KUserRichTexter() : d( new KPrivate() ) {}
KUserRichTexter::~KUserRichTexter() { delete d; }

QString KUserRichTexter::createFaceURL( const KUser &User, int Size ) const
{
  // ensure facefactory
  if( userMimeSourceFactory == 0 )
  {
    userMimeSourceFactory = new KUserMimeSourceFactory;
    QMimeSourceFactory::addFactory( userMimeSourceFactory );
  }

  return UserFaceProtocol + QChar(':') + User.loginName() + QChar(':') + QString::number(Size);
}

static const int DefaultStyle = ShowFullName | ShowFaceBehind /*| TwoRows*/;



static QString createName( const KUser &User, int NameStyle, bool UseTwoRows )
{
  QString S;

  QString Name = NameStyle&ShowFullName ? User.fullName() : User.loginName();
  QString Other = NameStyle==FirstLoginName ? User.fullName() : NameStyle==FirstFullName ? User.loginName()
                  : QString::null;
  if( UseTwoRows )
    S += QString::fromLatin1( "<td><nobr>" );

  S += Name;

  if( !Other.isNull() )
  {
    if( UseTwoRows )
      S += QString::fromLatin1("<br>(");
    else
      S += QString::fromLatin1(" (" );
    S += Other;
    S += QChar( ')' );
  }

  if( UseTwoRows )
    S += QString::fromLatin1( "</nobr></td>" );
  return S;
}

QString KUserRichTexter::createUserDesc( const KUser &User, int Style, bool Complete ) const
{
  if( Style < 0 )
    Style = DefaultStyle;


  bool UseTwoRows = Style & TwoRows;

  QString ToolTip;
  if( Complete )
    ToolTip += QString::fromLatin1( "<qt>" );
  if( UseTwoRows )
    ToolTip += QString::fromLatin1( "<table cellpadding=\"0\" cellspacing=\"1\"><tr>" );
  else 
    ToolTip += QString::fromLatin1("<nobr>");

  int FaceStyle = Style & ShowFaceMask;
  QString Face;
  if( FaceStyle & ShowFace )
  {
    if( UseTwoRows )
      Face += QString::fromLatin1( "<td>" );
    int FaceSize = UseTwoRows ? KIcon::SizeLarge  : KIcon::SizeSmall; // TODO: should depend on font size
    Face += QString::fromLatin1( "<img src=\"" );
    Face += createFaceURL( User, FaceSize );
    Face += QString::fromLatin1( "\">" );
    if( UseTwoRows )
      Face += QString::fromLatin1( "</td>" );
  }

  if( FaceStyle == ShowFaceBefore )
  {
    ToolTip += Face;
    ToolTip += QChar(' ');
  }
  ToolTip += createName( User, Style & NameMask, UseTwoRows );
  if( FaceStyle == ShowFaceBehind )
  {
    ToolTip += QChar(' ');
    ToolTip += Face;
  }

  if( UseTwoRows )
    ToolTip += QString::fromLatin1("</tr></table>");
  else 
    ToolTip += QString::fromLatin1("</nobr>");
  if( Complete )
    ToolTip += QString::fromLatin1( "</qt>" );

kdDebug() << ToolTip << endl;
  return ToolTip;
}
